//################################################################################
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.
//
//################################################################################


/* -------------------------------------------------------------------------------

	File:				MacBinary.cp
	
	Authors:		Josef W. Wankerl
	
	Description:
	
		This file contains code to encoding/decode MacBinary I/II/III files.
		 
	Modification History:

		Apr. 10, 1999
		
		Integrated into main source.
										
-------------------------------------------------------------------------------- */


#include <string.h>

#include "CVSEngine.h"
#include "CVSEngineUtils.h"
#include "UCVSRepositoryUtils.h"

#include "ThrowOSErr.h"
#include "StFileAccess.h"

#include "MoreDesktopMgr.h"
#include "MoreFilesExtras.h"

#include "MacBinary.h"

// JWW - The following routines were based on some functions from More Files
// The original routines were BumpDate and FSpBumpDate (thanks Jim!)

/*****************************************************************************/

static pascal OSErr	SetOtherInfo(short vRefNum,
						 long dirID,
						 ConstStr255Param name,
						 UInt32 createDate,
						 UInt32 modifiedDate,
						 Boolean useExtendedInfo,
						 UInt8 theScript,
						 UInt8 theExtendedFlags)
/* Given a file or directory, set its creation and modification dates. */
/* Optionally also set its extended finder flags and script fields. */
{
	CInfoPBRec pb;
	Str31 tempName;
	OSErr error;

	/* Protection against File Sharing problem */
	if ( (name == NULL) || (name[0] == 0) )
	{
		tempName[0] = 0;
		pb.hFileInfo.ioNamePtr = tempName;
		pb.hFileInfo.ioFDirIndex = -1;	/* use ioDirID */
	}
	else
	{
		pb.hFileInfo.ioNamePtr = (StringPtr)name;
		pb.hFileInfo.ioFDirIndex = 0;	/* use ioNamePtr and ioDirID */
	}
	pb.hFileInfo.ioVRefNum = vRefNum;
	pb.hFileInfo.ioDirID = dirID;
	error = PBGetCatInfoSync(&pb);
	if ( error == noErr )
	{
		pb.hFileInfo.ioFlCrDat = createDate;
		pb.hFileInfo.ioFlMdDat = modifiedDate;
		
		if (useExtendedInfo)
		{
			pb.hFileInfo.ioFlXFndrInfo.fdXFlags = theExtendedFlags;
			pb.hFileInfo.ioFlXFndrInfo.fdScript = theScript;
		}
		
		if ( pb.dirInfo.ioNamePtr == tempName )
		{
			pb.hFileInfo.ioDirID = pb.hFileInfo.ioFlParID;
		}
		else
		{
			pb.hFileInfo.ioDirID = dirID;
		}
		error = PBSetCatInfoSync(&pb);
	}
	
	return ( error );
}

/*****************************************************************************/

static pascal OSErr FSpSetOtherInfo(const FSSpec *spec, UInt32 createDate, UInt32 modifiedDate,
	Boolean useExtendedInfo, UInt8 theScript, UInt8 theExtendedFlags)
{
	return ( SetOtherInfo(spec->vRefNum, spec->parID, spec->name, createDate, modifiedDate,
		useExtendedInfo, theScript, theExtendedFlags) );
}

/*****************************************************************************/

static pascal OSErr	GetOtherInfo(short vRefNum,
						 long dirID,
						 ConstStr255Param name,
						 UInt32 *createDate,
						 UInt32 *modifiedDate,
						 UInt8 *theScript,
						 UInt8 *theExtendedFlags)
/* Given a file or directory, get its creation and modification dates and extended Finder info. */
{
	CInfoPBRec pb;
	Str31 tempName;
	OSErr error;

	/* Protection against File Sharing problem */
	if ( (name == NULL) || (name[0] == 0) )
	{
		tempName[0] = 0;
		pb.hFileInfo.ioNamePtr = tempName;
		pb.hFileInfo.ioFDirIndex = -1;	/* use ioDirID */
	}
	else
	{
		pb.hFileInfo.ioNamePtr = (StringPtr)name;
		pb.hFileInfo.ioFDirIndex = 0;	/* use ioNamePtr and ioDirID */
	}
	pb.hFileInfo.ioVRefNum = vRefNum;
	pb.hFileInfo.ioDirID = dirID;
	error = PBGetCatInfoSync(&pb);
	if ( error == noErr )
	{
		*createDate = pb.hFileInfo.ioFlCrDat;
		*modifiedDate = pb.hFileInfo.ioFlMdDat;
		
		*theExtendedFlags = pb.hFileInfo.ioFlXFndrInfo.fdXFlags;
		*theScript = pb.hFileInfo.ioFlXFndrInfo.fdScript;
	}
	
	return ( error );
}

/*****************************************************************************/

static pascal OSErr FSpGetOtherInfo(const FSSpec *spec, UInt32 *createDate, UInt32 *modifiedDate,
	UInt8 *theScript, UInt8 *theExtendedFlags)
{
	return ( GetOtherInfo(spec->vRefNum, spec->parID, spec->name, createDate, modifiedDate,
		theScript, theExtendedFlags) );
}

// CRC 16 table lookup array
UInt16 CRC16Table[256] =
	{0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
	0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
	0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
	0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
	0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
	0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
	0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
	0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
	0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
	0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
	0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
	0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
	0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
	0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
	0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
	0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
	0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
	0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
	0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
	0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
	0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
	0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
	0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
	0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
	0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
	0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
	0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
	0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
	0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
	0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
	0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
	0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0};

// CalculateCRC -- Excerpted from my Shrink II source code from years back...
static UInt16 CalculateCRC(UInt16 CRC, Ptr dataBlock, UInt32 dataSize)
{	
	while (dataSize)
	{
		CRC = (CRC << 8) ^ CRC16Table[((*dataBlock) ^ (CRC >> 8)) & 0x00FF];
		dataBlock++;
		dataSize--;
	}

	return CRC;
}

/*
	IsValidMacBinary
	
	This tests for a valid MacBinary, MacBinary II, or MacBinary III header
*/

Boolean IsValidMacBinary(char *theHeader, MacHeaderVersion *theVersion)
{
	Boolean isValid;
	MacHeaderVersion internalVersion;
	int i;
	
	isValid = false;
	internalVersion = kMacBinary;
	
	/*
		MacBinary, MacBinary II, and MacBinary III have the following minimum requirements:
			MBOldVersion must be zero
			MBZeroFill1 must be zero
			MBZeroFill2 must be zero
			MBComputerOS must be zero
			MBFilenameLength must be between 1 and 63
			
			MacBinary III restricts MBFilenameLength to 31, but that's checked for later
	*/
	
	if ((MBOldVersion(theHeader) == 0) && (MBZeroFill1(theHeader) == 0) &&
		(MBZeroFill2(theHeader) == 0) && (MBComputerOS(theHeader) == 0) &&
		(MBFilenameLength(theHeader) >= 1) && (MBFilenameLength(theHeader) <= 63))
	{
		/*
			MacBinary II and MacBinary III have the following requirements:
				MBSecondaryHeader must be zero
				MBMinVersion must be kVersion2 (129)
				MBChecksum must match the CRC of bytes 0 to kChecksumEnd (124)
		*/
		
		if ((MBSecondaryHeader(theHeader) == 0) && (MBMinVersion(theHeader) == kVersion2) &&
			(MBChecksum(theHeader) == CalculateCRC(0, theHeader, kChecksumEnd)))
		{
			// Check for the specific MacBinary II version number
			if (MBVersion2(theHeader) == kVersion2)
			{
				isValid = true;
				internalVersion = kMacBinaryII;
			}
			
			/*
				MacBinary III has the following requirements:
				MBVersion2 must be kVersion3 (130)
				MBSignature must be kMacBinarySignature ('mBIN')
				MBFilenameLength must be between 1 and 31
			*/
			
			// Check for the specific MacBinary III version and tags
			else if ((MBVersion2(theHeader) == kVersion3) &&
				(MBSignature(theHeader) == kMacBinarySignature) &&
				(MBFilenameLength(theHeader) <= 31))
			{
				isValid = true;
				internalVersion = kMacBinaryIII;
			}
		}
		
		if (internalVersion == kMacBinary)
		{
			/*
				MacBinary has the following additional requirements:
					MBDataFork must not be bigger than 0x007FFFFF
					MBResourceFork must not be bigger than 0x007FFFFF
					
					Unused bytes in the header must also be zero
			*/
			
			// Test for valid ranges in the data fork and resource fork sizes
			if (!((MBDataFork(theHeader) & ~0x007FFFFF) && (MBResourceFork(theHeader) & ~0x007FFFFF)))
				isValid = true;
			
			// Test for the large field of zeros
			i = 0;
			while (isValid && (i < kVersion1ZeroFillLength))
			{
				if (MBFindByte(theHeader, kVersion1ZeroFill + i) != 0)
					isValid = false;
				
				i++;
			}
		}
	}
	
	if (theVersion)
		*theVersion = internalVersion;
	
	return isValid;
}

OSErr FSpEncodeMacBinaryIIIHeader(const FSSpec *theSpec, char *theHeader, UInt32 *dataForkSize,
	UInt32 *resourceForkSize, UInt32 *totalSize, Str255 theComment)
{
	OSErr theErr;
	UInt32 createDate;
	UInt32 modifiedDate;
	UInt8 theScript;
	UInt8 theExtendedFlags;
	FileInfo theFileInfo;
	
	theErr = FSpGetFileSize(theSpec, (long *) dataForkSize, (long *) resourceForkSize);
	
	if (theErr == noErr)
		theErr = ::FSpGetFInfo(theSpec, (FInfo *) &theFileInfo);
	
	if (theErr == noErr)
		theErr = FSpGetOtherInfo(theSpec, &createDate, &modifiedDate, &theScript, &theExtendedFlags);
	
	if (theErr == noErr)
		{
			theErr = ::FSpDTGetComment(theSpec, theComment);
		
			if ( theErr != noErr )
				{
					theErr = noErr;
					theComment[0] = 0;
				}
		}
		
	if (theErr == noErr)
	{
		*totalSize = kMacBinaryHeaderSize + ((*dataForkSize + 0x7F) & ~0x7F) +
			((*resourceForkSize + 0x7F) & ~0x7F) + theComment[0];
		
		memset(theHeader, 0, kMacBinaryHeaderSize);
		
		// Yes, I know setting fields to zero in the header is redundant, but I do it anyway
		MBOldVersion(theHeader) = 0;
		memcpy(&MBFilenameLength(theHeader), theSpec->name, theSpec->name[0] + 1);
		MBFiletype(theHeader) = theFileInfo.fileType;
		MBCreator(theHeader) = theFileInfo.fileCreator;
		MBFinderFlagsHigh(theHeader) = (theFileInfo.finderFlags >> 8) & 0xFF;
		MBZeroFill1(theHeader) = 0;
		MBVerticalPos(theHeader) = theFileInfo.location.v;
		MBHorizontalPos(theHeader) = theFileInfo.location.h;
		MBFolderID(theHeader) = theFileInfo.reserved;
		MBProtected(theHeader) = 0;
		MBZeroFill2(theHeader) = 0;
		MBDataFork(theHeader) = *dataForkSize;
		MBResourceFork(theHeader) = *resourceForkSize;
		MBCreatedDate(theHeader) = createDate;
		MBModifiedDate(theHeader) = modifiedDate;
		MBComment(theHeader) = theComment[0];
		MBFinderFlagsLow(theHeader) = theFileInfo.finderFlags & 0xFF;
		MBSignature(theHeader) = kMacBinarySignature;
		MBScript(theHeader) = theScript;
		MBExtendedFlags(theHeader) = theExtendedFlags;
		MBUnpacked(theHeader) = 0;
		MBSecondaryHeader(theHeader) = 0;
		MBVersion2(theHeader) = kVersion3;
		MBMinVersion(theHeader) = kVersion2;
		
		MBChecksum(theHeader) = CalculateCRC(0, theHeader, kChecksumEnd);
	}
	
	return theErr;
}

OSErr FSpDecodeMacBinary(const FSSpec *inTargetSpec, const OSType inCreator, const OSType inType,
	const ScriptCode inScript, CEncoderYielder inYielder, void *pInContext)
{
	OSErr theErr;
	Boolean isCreated;
	Boolean isRenamed;
	UInt32 theBufferSize;
	FSSpec sourceSpec;
	FSSpec targetSpec;
	char *theBuffer;
	char theHeader[kMacBinaryHeaderSize];
	
	isCreated = false;
	isRenamed = false;
	theBuffer = NULL;
	
	try
	{
		if (inTargetSpec == NULL)
			ThrowOSErr(kCVSEngineBadParamErr);
		
		theBuffer = ::NewPtr(kDefaultBufferSize);
		
		if (theBuffer == NULL)
			theErr = memFullErr;
		else
			theErr = ::MemError();
		
		ThrowOSErr(theErr);
		
		theBufferSize = kDefaultBufferSize;
		
		//  Rename the file we're trying to decode to something squirley...
		sourceSpec = *inTargetSpec;
		theErr = FSpSafeRename(&sourceSpec);
		ThrowOSErr(theErr);
		isRenamed = true;
		
		targetSpec = *inTargetSpec;
		theErr = ::FSpCreate(&targetSpec, inCreator, inType, inScript);
		ThrowOSErr(theErr);
		isCreated = true;
		
		//  Open both file forks for locking from other threads. Scope
		//  is important here because StFileAccess will close files when
		//  destroyed.
		{
			StFileAccess targetItemAccess(targetSpec);
			StFileAccess sourceItemAccess(sourceSpec);
			
			targetItemAccess.SetDelete(true);
			
			SInt16 sourceDataRef = sourceItemAccess.GetFSDataForkRef();
			SInt16 sourceResRef = sourceItemAccess.GetFSResForkRef();
			
			SInt16 targetDataRef = targetItemAccess.GetFSDataForkRef();
			SInt16 targetResRef = targetItemAccess.GetFSResForkRef();
			
			UInt32 maxOffset;
			theErr = ::GetEOF(sourceDataRef, (SInt32 *) &maxOffset);
			ThrowOSErr(theErr);
			
			if (maxOffset < kMacBinaryHeaderSize)
			{
				theErr = kCVSEngineNotValidEncodedDataErr;
				throw theErr;
			}
			
			UInt32 requestCount = kMacBinaryHeaderSize;
			theErr = FSReadOffset(sourceDataRef, &requestCount, &theHeader, 0);
			ThrowOSErr(theErr);
			
			MacHeaderVersion theVersion;
			if (!IsValidMacBinary(theHeader, &theVersion))
			{
				theErr = kCVSEngineNotValidEncodedDataErr;
				throw theErr;
			}
			
			UInt32 remaining = MBDataFork(theHeader);
			UInt32 currentSourcePos =
				(kMacBinaryHeaderSize + MBSecondaryHeader(theHeader) + 0x7F) & ~0x7F;
			UInt32 currentTargetPos = 0;
			
			while (remaining)
			{
				if (remaining > kDefaultBufferSize)
					requestCount = kDefaultBufferSize;
				else
					requestCount = remaining;
				
				theErr = FSReadOffset(sourceDataRef, &requestCount, theBuffer, currentSourcePos);
				ThrowOSErr(theErr);
				
				theErr = FSWriteOffset(targetDataRef, &requestCount, theBuffer, currentTargetPos);
				ThrowOSErr(theErr);
				
				remaining -= requestCount;
				currentSourcePos += requestCount;
				currentTargetPos += requestCount;
				
				if (inYielder != NULL)
				{
					theErr = (inYielder)(pInContext, maxOffset, currentSourcePos);
					ThrowOSErr(theErr);
				}
			}
			
			remaining = MBResourceFork(theHeader);
			currentSourcePos = (currentSourcePos + 0x7F) & ~0x7F;
			currentTargetPos = 0;
			
			while (remaining)
			{
				if (remaining > kDefaultBufferSize)
					requestCount = kDefaultBufferSize;
				else
					requestCount = remaining;
				
				theErr = FSReadOffset(sourceDataRef, &requestCount, theBuffer, currentSourcePos);
				ThrowOSErr(theErr);
				
				theErr = FSWriteOffset(targetResRef, &requestCount, theBuffer, currentTargetPos);
				ThrowOSErr(theErr);
				
				remaining -= requestCount;
				currentSourcePos += requestCount;
				currentTargetPos += requestCount;
				
				if (inYielder != NULL)
				{
					theErr = (inYielder)(pInContext, maxOffset, currentSourcePos);
					ThrowOSErr(theErr);
				}
			}
			
			remaining = MBComment(theHeader);
			currentSourcePos = (currentSourcePos + 0x7F) & ~0x7F;
			
			// The comment field did not exist until MacBinary II
			if (remaining && (theVersion != kMacBinary))
			{
				// This works because comments cannot be longer than 200 characters and the
				// buffer size should be much (much much) larger than that
				if (remaining > kDefaultBufferSize)
					requestCount = kDefaultBufferSize;
				else
					requestCount = remaining;
				
				// Read the comment to theBuffer[1] so we can set theBuffer[0] to the length
				theErr = FSReadOffset(sourceDataRef, &requestCount, theBuffer + 1, currentSourcePos);
				ThrowOSErr(theErr);
				
				// FSpDTSetComment expects a pascal string for the comment, so set the length byte
				*theBuffer = requestCount;
				theErr = ::FSpDTSetComment(&targetSpec, (ConstStr255Param) theBuffer);
				ThrowOSErr(theErr);
				
				currentSourcePos += requestCount;
				
				if (inYielder != NULL)
				{
					theErr = (inYielder)(pInContext, maxOffset, currentSourcePos);
					ThrowOSErr(theErr);
				}
			}
			
			FileInfo theFileInfo;
			theFileInfo.fileType = MBFiletype(theHeader);
			theFileInfo.fileCreator = MBCreator(theHeader);
			
			// The low byte of the finderFlags did not exist until MacBinary II
			if (theVersion == kMacBinary)
				theFileInfo.finderFlags = MBFinderFlagsHigh(theHeader) << 8;
			else
				theFileInfo.finderFlags = MBFinderFlags(theHeader);
			
			// These bits should always be clear when creating a new file
			theFileInfo.finderFlags &= ~(kIsOnDesk + kHasBeenInited);
			
			theFileInfo.location.v = MBVerticalPos(theHeader);
			theFileInfo.location.h = MBHorizontalPos(theHeader);
			theFileInfo.reserved = 0;
			theErr = ::FSpSetFInfo(&targetSpec, (FInfo *) &theFileInfo);
			ThrowOSErr(theErr);
			
			// The script and extended Finder flags did not exist until MacBinary III
			Boolean useExtendedFlags = (theVersion == kMacBinaryIII);
			
			// Do this so the modification date isn't set by closing the file
			targetItemAccess.CloseAllFileRefs();
			
			theErr = ::FSpSetOtherInfo(&targetSpec, MBCreatedDate(theHeader),
				MBModifiedDate(theHeader), useExtendedFlags, MBScript(theHeader),
				MBExtendedFlags(theHeader));
			ThrowOSErr(theErr);
			
			targetItemAccess.SetDelete(false);
		}
		
		::FSpDelete(&sourceSpec);
	}
	catch (OSErr theException)
	{
		theErr = theException;
		
		if (isCreated)
			::FSpDelete(&targetSpec);
		
		if (isRenamed)
			::FSpRename(&sourceSpec, targetSpec.name);
	}
	catch (...)
	{
		theErr = kCVSEngineUnknownExceptionErr;
		
		if (isCreated)
			::FSpDelete(&targetSpec);
		
		if (isRenamed)
			::FSpRename(&sourceSpec, targetSpec.name);
	}
	
	try
	{
		if (theBuffer != NULL)
			::DisposePtr(theBuffer);
	}
	catch (...)
	{
	}
	
	return theErr;
}