//################################################################################
//
//  Copyright (C) 1997-1998 Netscape Communications Corporation.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.
//
//################################################################################


/* -------------------------------------------------------------------------------

	File:				CCVSEntries.cp
	
	Authors:		Ram Rajadhyaksha
	
	Description:
	
		This file contains code to read/load/write the "Entries" file in
		each working directory.
		 
	Modification History:

		16  Feb		1999		Ram Rajadhyaksha
		
				Disabled yielding when writing a modified entry file. This prevents
				problems where the front end is reading an entries file with fsRdPerm
				while the CVS Engine is writing a new version to disk with fsRdWrPerm.
				
				One way to fix this is to make sure that the between yields, the writing
				operation left the file in a "decent" state. (ie, valid but maybe not
				all the data is there yet)
				
		15 	Dec		1998		Ram Rajadhyaksha
		
				Created.
										
-------------------------------------------------------------------------------- */

#include <string.h>

#include "ThrowOSErr.h"
#include "StringUtils.h"
#include "CVSEngine.h"
#include "CVSEngineUtils.h"
#include "UCVSRepositoryUtils.h"

#include "CCVSEntries.h"


const unsigned char kCVSDirectoryName[] 		= "\pCVS";
const unsigned char	kCVSEntriesName[] 			= "\pEntries";
const unsigned char kCVSEntriesStaticName[] = "\pEntries.Static";
const unsigned char kCVSEntriesLogName[] 		= "\pEntries.Log";
const unsigned char kCVSEntriesTempName[] 	= "\pEntries.Tmp";


#ifdef CVSENGINE_TERMINATE_WITH_LF

unsigned char kEntriesTerminatorByte	= 0x0A;

#else

unsigned char kEntriesTerminatorByte	= 0x0D;

#endif


CCVSEntries::CCVSEntries()
	{	
		mFileSpec.name[0] = 0;
		mFileSpec.vRefNum = 0;
		mFileSpec.parID = 0;

		mWorkingDir.name[0] = 0;
		mWorkingDir.vRefNum = 0;
		mWorkingDir.parID = 0;
		
		mFileRef = 0;
		mPerm = 0;
		mDirty = false;		
	}
	
	
CCVSEntries::~CCVSEntries()
	{	
		if ( FSpIsValid(mFileSpec) && (mFileRef != 0) )
			{			
				CloseEntriesFile();			
			}		
	}
	
	
#pragma mark -


OSErr
CCVSEntries::LoadWorkingDirectory( const FSSpec *inWorkingDirectory, const SInt8 inPerm )
	{
			
		OSErr				result 					= noErr;
		Str255			pstrPartialPath = "\p";
		FSSpec			fSpec						= { 0, 0, "\p" };
		CInfoPBRec	pb;

		Assert_( inWorkingDirectory != NULL );
		
		if ( inWorkingDirectory != NULL )
			{		
				pstrPartialPath[0] = 1;
				pstrPartialPath[1] = ':';
				PStrConcat( pstrPartialPath, inWorkingDirectory->name, sizeof(pstrPartialPath) );
				PStrConcat( pstrPartialPath, "\p:", sizeof(pstrPartialPath) );
				PStrConcat( pstrPartialPath, kCVSDirectoryName, sizeof(pstrPartialPath) );
				PStrConcat( pstrPartialPath, "\p:", sizeof(pstrPartialPath) );
				PStrConcat( pstrPartialPath, kCVSEntriesName, sizeof(pstrPartialPath) );
				
				pb.hFileInfo.ioCompletion 	= NULL;
				pb.hFileInfo.ioFDirIndex 		= 0;
				pb.hFileInfo.ioNamePtr 			= pstrPartialPath;
				pb.hFileInfo.ioVRefNum 			= inWorkingDirectory->vRefNum;
				pb.hFileInfo.ioDirID 				= inWorkingDirectory->parID;
				
				result = ::PBGetCatInfoSync( &pb );
				
				if ( result == noErr ) 
					{				
						PStrConcat( fSpec.name, kCVSEntriesName, sizeof(fSpec.name) );
						
						fSpec.parID = pb.hFileInfo.ioFlParID;
						fSpec.vRefNum = pb.hFileInfo.ioVRefNum;
						
						result = LoadEntriesFile( &fSpec, inWorkingDirectory, inPerm );					
					}		
			}	
				
		return result;
		
	}
	
	
OSErr
CCVSEntries::LoadEntriesFile( const FSSpec *inEntriesFile, const FSSpec *inWorkingDirectory, const SInt8 inPerm )
	{
	
		OSErr 		intResult			= noErr;
		
		SInt32		intFileLen		= 0;
		SInt32		intBufSize		= 0;

		UInt32		intFileOffset	= 0;		
		UInt32		intBufIndex 	= 0;
		UInt32		intBufCount		= 0;
		
		UInt32		intLineCount 	= 0;
		UInt32		intCopyLen 		= 0;
		
		Str255		pstrEntry			= "\p";
		
		Boolean		bDone					= false;
		Boolean		bTooBig				= false;
		
		char			*pBuffer			= NULL;
		char			*pLineTerm		= NULL;
		char			strEntryLine[kMaxLineSize]	= "";
		
		Assert_( inEntriesFile != NULL );
		
		try
			{		
				if ( inEntriesFile == NULL )
					throw( (OSErr)kCVSEngineBadParamErr );

				if ( mFileRef != 0 )
					throw( (OSErr)kCVSEngineBadParamErr );
				
				if ( mRootNode == NULL )
					CreateTree();
						
				intResult = ::FSpOpenDF( inEntriesFile, inPerm, &mFileRef );
			
				if ( intResult != noErr )
					{				
						mFileRef = 0;					
						throw intResult;					
					}	
	
				mFileSpec = *inEntriesFile;
				mPerm = inPerm;
						
				intResult = ::GetEOF( mFileRef, &intFileLen );
				ThrowOSErr( intResult );
							
				if ( intFileLen > kMaxBufSize )
					{
						intBufSize = kMaxBufSize;
					}
				else
					{
						intBufSize = intFileLen;
					}
				
				//  +1 for a terminator byte. This ensures that
				//  the ANSI functions like strpbrk() don't 
				//  "run away" into oblivion.
				
				pBuffer = ::NewPtr( intBufSize + 1 );
				intResult = ::MemError();
				
				if ( (intResult == noErr) && (pBuffer == NULL) )
					{
						intResult = memFullErr;
					}
						
				Assert_( intResult == noErr );
				ThrowOSErr( intResult );
				
				//  Terminate the last byte
				pBuffer[intBufSize] = 0;
				
				while ( intFileOffset < intFileLen )
					{					
						intBufIndex = 0;				
						intBufCount = intFileLen - intFileOffset;
						
						if ( intBufCount > intBufSize )
							{
								intBufCount = intBufSize;		
							}
							
						intResult = FSReadOffset( mFileRef, &intBufCount, pBuffer, intFileOffset );
						ThrowOSErr( intResult );
						
						//  This should be ok, even if we fill the
						//  buffer completely because we allocated
						//  an extra byte in the ::NewPtr() call.	
						pBuffer[intBufCount] = 0;
						
						bDone = false;
						
						while ( intBufIndex < intBufCount )
							{		
								//  I hope this ANSI function is ok to use.
								//  The terminator string is both '\r' and '\n'
								//  chars so we work with CVS Entries generated
								//  by Rhapsody's CVS.
								
								pLineTerm = strpbrk( pBuffer + intBufIndex, "\r\n" );
										
								if ( pLineTerm != NULL ) 
									{								
										intCopyLen = pLineTerm - (pBuffer + intBufIndex);
											
										if ( (!bTooBig) && (intCopyLen < (kMaxLineSize - intLineCount)) )
											{		
												if ( intCopyLen > 0 )
													{								
														::BlockMoveData( pBuffer + intBufIndex, strEntryLine + intLineCount, intCopyLen );
														intLineCount += intCopyLen;
													}	
													
												strEntryLine[intLineCount] = 0;
														
												//  Ok now we got the line, stuff it into the Radix Tree
												
												if ( intLineCount > 0 )
													{
														if ( (strEntryLine[0] == '/') || (strEntryLine[0] == 'D') )
															{											
																intResult = AddModifyEntryStr( strEntryLine );
																Assert_( intResult == noErr );	
																
																//  Ok, this will fail if we get an invalid line
																//  format, or a line with a filename that has
																//  a \r or \n char in it. So, for now my attitude
																//  is "screw it" and just ignore the error.
																
																if ( intResult != noErr )
																	{
																		if ( (intResult != kCVSEngineBadParamErr) && (intResult != kCVSEngineInvalidFileNameErr) )
																			{
																				throw intResult;
																			}
																	}													
															}													
														else
															{												
																//  Uh oh, this is an unsupported line format. 
																//  The CVS 1.10 spec says we should preserve
																//  unknown lines. Um, someone want to write
																//  the code to do that?
															}	
													}
											}
										
										intLineCount = 0;	
										bTooBig = false;									
									}								
								else 
									{									
										intCopyLen = intBufCount - intBufIndex;
									
										if ( bTooBig == false )
											{												
												if ( intCopyLen < (kMaxLineSize - intLineCount) )
													{													
														::BlockMoveData( pBuffer + intBufIndex, strEntryLine + intLineCount, intCopyLen );
														intLineCount += intCopyLen;												
													}													
												else 
													{
														bTooBig = true;
													}
											}																						
									}
									
								intBufIndex += intCopyLen + 1;							
							}
							
						intFileOffset += intBufCount;
						
						//  Call our yield function here
						
						intResult = Yield();
						ThrowOSErr( intResult );									
					}
					
				mDirty = false;
						
				if ( inWorkingDirectory != NULL )
					{
						mWorkingDir = *inWorkingDirectory;
					}							
			}
		
		catch ( ... )
			{			
				if ( intResult == noErr )
					{
						intResult = kCVSEngineUnknownErr;
					}
						
			 	if ( mFileRef != 0 )
			 		{			
						OSErr	intTempErr = noErr;
			 		
			 			intTempErr = ::FSClose( mFileRef );
			 			mFileRef = 0;
			 			
			 			Assert_( intTempErr == noErr );		 			
			 		}
			 		
			 	mFileSpec.name[0] = 0;
			 	mFileSpec.parID = 0;
			 	mFileSpec.vRefNum = 0;

			 	mWorkingDir.name[0] = 0;
			 	mWorkingDir.parID = 0;
			 	mWorkingDir.vRefNum = 0;
			 	
			 	mPerm = 0;
			 	mDirty = false;		 		
			}

		if ( pBuffer != NULL )
			{
				::DisposePtr( (Ptr)pBuffer );
				pBuffer = NULL;
			}
				
		return intResult;
		
	}


OSErr
CCVSEntries::CloseEntriesFile()
	{
	
		OSErr intResult = noErr;
		
		try
			{
			
				if ( mFileRef != 0 )
					{
					
						if ( (mPerm == fsRdWrPerm) && (mDirty) )
							{
								intResult = FlushEntriesFile();
								Assert_( intResult == noErr );
							}
							
						//  Don't throw.
						
						intResult = ::FSClose( mFileRef );
						mFileRef = 0;
						Assert_( intResult == noErr );
						
						mPerm = 0;
						mDirty = false;
						
						mFileSpec.name[0] = 0;
						mFileSpec.parID = 0;
						mFileSpec.vRefNum = 0;

						mWorkingDir.name[0] = 0;
						mWorkingDir.parID = 0;
						mWorkingDir.vRefNum = 0;
						
						try
							{
								DestroyTree();
							}
						catch ( ... )
							{
								// we're just generally hosed if we get an exception here
							}
						
						mRootNode = NULL;
							
					}
					
			}
			
		catch ( OSErr intException )
			{
			
				intResult = intException;
				
			}
			
		catch ( ... )
			{
			
				intResult = kCVSEngineUnknownExceptionErr;
				
			}
			
		return intResult;
		
	}
	
	
OSErr
CCVSEntries::FlushEntriesFile( const Boolean inTouchModDate )
	{
	
		OSErr 	intResult 			= 0;
		
		SInt32	intFileLen			= 0;
		SInt32	intModDate			= 0;
		UInt32	intFileOffset 	= 0;
		
		UInt32	intBufSize 			= 0;
		UInt32	intBufCount			= 0;
		UInt32	intCount				= 0;
		UInt32	intDirCount			= 0;
		
		char		*pBuffer 				= NULL;
		
		try
			{			
				intBufSize = kMaxBufSize;				
				pBuffer = (char *)::NewPtr( intBufSize );		
				Assert_( pBuffer != NULL );
				
				if ( pBuffer == NULL )
					throw ::MemError();
				
				intResult = ::GetEOF( mFileRef, &intFileLen );
				ThrowOSErr( intResult );
						
				//  Recursively walk the radix tree, stuffing data into the file
				
				WalkEntryList( mRootNode, intFileLen, pBuffer, intBufSize, &intBufCount, &intFileOffset, &intDirCount );

				intResult = ::SetEOF( mFileRef, (intFileOffset + intBufCount) );
				ThrowOSErr( intResult );
													
				//  If there is data left, blow it out.
				
				if ( intBufCount > 0 )
					{												
						intCount = intBufCount;
						intResult = FSWriteOffset( mFileRef, &intCount, pBuffer, intFileOffset );
						ThrowOSErr( intResult );
						intFileOffset += intCount;
					}
				
				//  If we didn't write any directory entries, we need to
				//  set the last line to 'D/r'. This should be optimized
				//  so ::SetEOF is not called twice.
				
				if ( intDirCount == 0 )
					{
						intResult = ::SetEOF( mFileRef, (intFileOffset + 2) );
						ThrowOSErr( intResult );

						intCount = 2;
						intResult = FSWriteOffset( mFileRef, &intCount, "D\r", intFileOffset );
						ThrowOSErr( intResult );
						intFileOffset += intCount;					
					}
						
				if ( inTouchModDate )
					{
						intModDate = FSpGetModDate( &mFileSpec );
						FSpSetModDate( &mFileSpec, intModDate + 1 );		
					}	
			}
			
		catch ( OSErr intException )
			{		
				intResult = intException;				
			}			
		catch ( ... )
			{		
				intResult = kCVSEngineUnknownExceptionErr;				
			}
			
		if ( pBuffer != NULL )
			{
				::DisposePtr( (Ptr)pBuffer );
				pBuffer = NULL;
			}

		return intResult;
		
	}
	
	
#pragma mark -


OSErr
CCVSEntries::AddModifyEntryStr( const char *inEntryStr )
	{
	
		OSErr		intResult 			= 0;
		
		UInt32	intLineLen			= 0;
		UInt32	intCopyLen			= 0;
		Str255	pstrKey 				= "\p";
		
		char		*strEntryLine 	= NULL;
		char		*strOffset 			= 0;
		char		*strPrevOffset	= 0;
		
		try
			{		
				Assert_( inEntryStr != NULL );
				
				if ( inEntryStr == NULL )
					{
						ThrowOSErr( kCVSEngineBadParamErr );
					}
	
				intLineLen = strlen( inEntryStr );

				if ( intLineLen == 0 )
					{
						ThrowOSErr( kCVSEngineBadParamErr );
					}
				
				Assert_( (inEntryStr[0] == '/') || (inEntryStr[0] == 'D') );
				
				if ( (inEntryStr[0] != '/') && (inEntryStr[0] != 'D') )
					{
						ThrowOSErr( kCVSEngineBadParamErr );
					}				
						
				if ( (inEntryStr[0] == 'D') && (intLineLen == 1) )
					{				
						//  ignore this line				
					}
				else
					{
						strOffset = (char *)inEntryStr;
						strOffset++;
						
						if ( inEntryStr[0] == 'D' )
							{
								strOffset++;
							}
						
						strPrevOffset = strOffset;
						
						strOffset = strchr( strOffset, '/' );
						
						if ( (strOffset == NULL) || ((strOffset - inEntryStr) >= intLineLen) )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}
							
						intCopyLen = strOffset - strPrevOffset;
						
						if ( intCopyLen > 255 )
							{
								ThrowOSErr( kCVSEngineInvalidFileNameErr );
							}
							
						::BlockMoveData( strPrevOffset, &pstrKey[1], intCopyLen );
						pstrKey[0] = intCopyLen;

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
						{
							Str255 pstrTranslated = "\p";
							UInt32 intLen = pstrKey[0];
							
							CUTranslateToMacName( (char *)&pstrKey[1], intLen );
							pstrKey[0] = intLen;
							
							PStrLowercase( pstrKey );
							
							CUTranslateFromMacNamePStr( pstrKey, pstrTranslated );
							PStrCopy( pstrTranslated, pstrKey, sizeof(pstrKey) );						
						}
#else						
						PStrLowercase( pstrKey );
#endif
						
						strOffset++;
						strPrevOffset = strOffset;
				
//  Revision
				
						strOffset = strchr( strOffset, '/' );
						
						if ( (strOffset == NULL) || ((strOffset - inEntryStr) >= intLineLen) )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}
					
//  Time Stamp
	
						strOffset++;
						strPrevOffset = strOffset;
					
						strOffset = strchr( strOffset, '/' );
						
						if ( (strOffset == NULL) || ((strOffset - inEntryStr) >= intLineLen) )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}

//  Options
	
						strOffset++;
						strPrevOffset = strOffset;
					
						strOffset = strchr( strOffset, '/' );
						
						if ( (strOffset == NULL) || ((strOffset - inEntryStr) >= intLineLen) )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}
					
//  Ok, the entry line seems to be somewhat sane. Put it in the Radix tree.
				
						AddKeyPStrToTree( pstrKey, inEntryStr, intLineLen );
					
						mDirty = true;
					}				
			}
			
		catch ( OSErr intException )
			{		
				intResult = intException;			
			}			
		catch ( ... )
			{			
				Assert_( false );
				intResult = kCVSEngineUnknownExceptionErr;				
			}
			
		Assert_( intResult == noErr );
		
		return intResult;
		
	}
	
	
OSErr
CCVSEntries::AddModifyEntry( const Str255 inNamePStr, const Str255 inRevisionPStr, const Str255 inTimeStampPStr,
													const Str255 inOptionsPStr,	const Str255 inTagPStr, const Boolean inIsDirectory)
	{
	
		OSErr 	intResult										= noErr;
		UInt32	intLength										= 0;
		UInt32	intIndex										= 0;
		Str255	pstrKey											= "\p";

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
		Str255	pstrTranslated							= "\p";
#endif

		char		strEntryLine[kMaxLineSize] 	= "";
		
		try
			{	

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
				CUTranslateFromMacNamePStr( inNamePStr, pstrTranslated );
				intLength = 5 + pstrTranslated[0] + inRevisionPStr[0] + inTimeStampPStr[0] + inOptionsPStr[0] + inTagPStr[0];
#else
				intLength = 5 + inNamePStr[0] + inRevisionPStr[0] + inTimeStampPStr[0] + inOptionsPStr[0] + inTagPStr[0];
#endif			
	
				if ( inIsDirectory )
					intLength++;
					
				if ( intLength >= kMaxLineSize )
					throw( (OSErr) kCVSEngineBadParamErr );
					
				if ( inIsDirectory )
					{			
						strEntryLine[0] = 'D';
						intIndex++;					
					}
				
				strEntryLine[intIndex++] = '/';

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
				::BlockMoveData( &pstrTranslated[1], &strEntryLine[intIndex], pstrTranslated[0] );
				intIndex += pstrTranslated[0];
#else				
				::BlockMoveData( &inNamePStr[1], &strEntryLine[intIndex], inNamePStr[0] );
				intIndex += inNamePStr[0];
#endif
				
				strEntryLine[intIndex++] = '/';
				::BlockMoveData( &inRevisionPStr[1], &strEntryLine[intIndex], inRevisionPStr[0] );
				intIndex += inRevisionPStr[0];
			
				strEntryLine[intIndex++] = '/';
				::BlockMoveData( &inTimeStampPStr[1], &strEntryLine[intIndex], inTimeStampPStr[0] );
				intIndex += inTimeStampPStr[0];

				strEntryLine[intIndex++] = '/';
				::BlockMoveData( &inOptionsPStr[1], &strEntryLine[intIndex], inOptionsPStr[0] );
				intIndex += inOptionsPStr[0];

				strEntryLine[intIndex++] = '/';
				::BlockMoveData( &inTagPStr[1], &strEntryLine[intIndex], inTagPStr[0] );
				intIndex += inTagPStr[0];
								
				PStrCopy( inNamePStr, pstrKey, sizeof(pstrKey) );
				PStrLowercase( pstrKey );

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
				CUTranslateFromMacNamePStr( pstrKey, pstrTranslated );
				PStrCopy( pstrTranslated, pstrKey, sizeof(pstrKey) );
#endif
								
				AddKeyPStrToTree( pstrKey, strEntryLine, intIndex );
				
				mDirty = true;				
			}
			
		catch ( OSErr intException )
			{			
				intResult = intException;				
			}						
		catch ( ... )
			{			
				intResult = kCVSEngineUnknownExceptionErr;				
			}
			
		return intResult;
		
	}
	
	
OSErr
CCVSEntries::GetEntry( const Str255 pstrInKey, Str255 pstrOutRevision, Str255 pstrOutTimeStamp, Str255 pstrOutOptions, Str255 pstrOutTag )
	{
	
		OSErr		intResult 									= noErr;
		UInt32	intEntryLen									= kMaxLineSize - 1;
		UInt32	intCopyLen									= 0;
		
		Str255	pstrKey											= "\p";

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
		Str255	pstrTranslated							= "\p";
#endif

		char		strEntryLine[kMaxLineSize] 	= "";
		char		*strOffset									= NULL;
		char		*strLastOffset							= NULL;
		
		try	
			{		
				if ( pstrInKey == NULL )
					{
						ThrowOSErr( kCVSEngineBadParamErr );
					}
				else
					{			
						PStrCopy( pstrInKey, pstrKey, sizeof(pstrKey) );						
						PStrLowercase( pstrKey );

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
						CUTranslateFromMacNamePStr( pstrKey, pstrTranslated );
						PStrCopy( pstrTranslated, pstrKey, sizeof(pstrKey) );
#endif
							
						if ( !MatchKeyPStr(pstrKey, strEntryLine, intEntryLen, true) )
							{
								ThrowOSErr( kCVSEngineNoRevisionDataErr );
							}
							
						if ( intEntryLen == 0 )
							{
								ThrowOSErr( kCVSEngineNoRevisionDataErr );
							}
							
						//  paranoid termination
						
						if ( intEntryLen > 0 )
							{
								strEntryLine[intEntryLen] = 0;
							}
							
//  First slash
		
						strOffset = strchr( strEntryLine, '/' );
				
						if ( strOffset == NULL )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}

//  Name
		
						strOffset = strchr( strOffset + 1, '/' );
		
						if ( strOffset == NULL )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}

//  Revision

						strLastOffset = strOffset + 1;				
						strOffset = strchr( strLastOffset, '/' );
		
						if ( strOffset == NULL )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}

						if ( pstrOutRevision != NULL )
							{
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutRevision[1], intCopyLen );
								pstrOutRevision[0] = intCopyLen;
							}
						
//  Time Stamp

						strLastOffset = strOffset + 1;
						strOffset = strchr( strLastOffset, '/' );
		
						if ( strOffset == NULL )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}
							
						if ( pstrOutTimeStamp != NULL )
							{
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutTimeStamp[1], intCopyLen );
								pstrOutTimeStamp[0] = intCopyLen;
							}

//  Options

						strLastOffset = strOffset + 1;
						strOffset = strchr( strLastOffset, '/' );
		
						if ( strOffset == NULL )
							{
								ThrowOSErr( kCVSEngineBadParamErr );
							}
							
						if ( pstrOutOptions != NULL )
							{
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutOptions[1], intCopyLen );
								pstrOutOptions[0] = intCopyLen;
							}

//  Tag
						
						if ( pstrOutTag != NULL )
							{
								if ( (strEntryLine + intEntryLen) > (strOffset + 1) )
									{
									
										intCopyLen = (strEntryLine + intEntryLen) - (strOffset + 1);		

										if ( intCopyLen > 255 )
											intCopyLen = 255;
											
										::BlockMoveData( strOffset + 1, &pstrOutTag[1], intCopyLen );
										pstrOutTag[0] = intCopyLen;
									
									}
								else
									{
										pstrOutTag[0] = 0;
									}
							}
						
					}
			
			}
		
		catch ( OSErr intException )
			{
				intResult = intException;
			}
		catch ( ... )
			{
				intResult = kCVSEngineUnknownExceptionErr;
			}
		
		Assert_( (intResult == noErr) || (intResult == kCVSEngineNoRevisionDataErr) );
			
		return intResult;
	
	}
	
	
OSErr
CCVSEntries::RemoveEntry( const Str255 inNamePStr )
	{
	
		OSErr		intResult = noErr;
		Str255	pstrKey 	= "\p";

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
		Str255	pstrTranslated	= "\p";
#endif
		
		Assert_( inNamePStr != NULL );
		
		try
			{				
				PStrCopy( inNamePStr, pstrKey, sizeof(pstrKey) );
				PStrLowercase( pstrKey );

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
				CUTranslateFromMacNamePStr( pstrKey, pstrTranslated );
				PStrCopy( pstrTranslated, pstrKey, sizeof(pstrKey) );
#endif
				
				//  Ok, due to the fact that I wrote the Radix Tree code a year ago
				//  and am completely clueless on how it works now, to delete a node
				//  I just re-add it with a zero length context ptr. This won't delete
				//  the node data, but it will free the entry string.
				
				AddKeyPStrToTree( pstrKey, NULL, 0 );
				mDirty = true;			
			}			
		catch ( OSErr intException )
			{			
				intResult = intException;				
			}
		catch ( ... )
			{			
				intResult = kCVSEngineUnknownExceptionErr;				
			}
			
		return intResult;
		
	}
	

OSErr
CCVSEntries::GetEntryItems( const char *strInEntry, const UInt32 intInLen, 
														Str255 pstrOutName, Str255 pstrOutRevision, Str255 pstrOutTimeStamp, 
														Str255 pstrOutOptions, Str255 pstrOutTag )
	{
	
		OSErr		intResult 									= noErr;
		UInt32	intEntryLen									= kMaxLineSize;
		UInt32	intCopyLen									= 0;
		
		char		*strOffset									= NULL;
		char		*strLastOffset							= NULL;
		
		try	
			{
			
				if ( (strInEntry == NULL) || (intInLen == 0) )
					ThrowOSErr( kCVSEngineBadParamErr );
				else
					{
					
//  First slash
		
						strOffset = strchr( strInEntry, '/' );
				
						if ( strOffset == NULL )
							ThrowOSErr( kCVSEngineBadParamErr );

						strLastOffset = strOffset + 1;

//  Name
		
						strOffset = strchr( strOffset + 1, '/' );
		
						if ( strOffset == NULL )
							ThrowOSErr( kCVSEngineBadParamErr );

						if ( pstrOutName != NULL )
							{							
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutName[1], intCopyLen );
								pstrOutName[0] = intCopyLen;
							}
							
//  Revision

						strLastOffset = strOffset + 1;				
						strOffset = strchr( strLastOffset, '/' );
		
						if ( strOffset == NULL )
							ThrowOSErr( kCVSEngineBadParamErr );

						if ( pstrOutRevision != NULL )
							{						
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutRevision[1], intCopyLen );
								pstrOutRevision[0] = intCopyLen;								
							}
						
//  Time Stamp

						strLastOffset = strOffset + 1;
						strOffset = strchr( strLastOffset, '/' );
		
						if ( strOffset == NULL )
							ThrowOSErr( kCVSEngineBadParamErr );

						if ( pstrOutTimeStamp != NULL )
							{
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutTimeStamp[1], intCopyLen );
								pstrOutTimeStamp[0] = intCopyLen;
							}
							
//  Options

						strLastOffset = strOffset + 1;
						strOffset = strchr( strLastOffset, '/' );
		
						if ( strOffset == NULL )
							ThrowOSErr( kCVSEngineBadParamErr );

						if ( pstrOutOptions != NULL )
							{
								intCopyLen = strOffset - strLastOffset;
								
								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strLastOffset, &pstrOutOptions[1], intCopyLen );
								pstrOutOptions[0] = intCopyLen;
							}

//  Tag
						
						if ( pstrOutTag != NULL )
							{
								intCopyLen = (strInEntry + intInLen) - (strOffset + 1);		

								if ( intCopyLen > 255 )
									intCopyLen = 255;
									
								::BlockMoveData( strOffset + 1, &pstrOutTag[1], intCopyLen );
								pstrOutTag[0] = intCopyLen;
							}
					}
			
			}
		
		catch ( OSErr intException )
			{
				intResult = intException;
			}
		catch ( ... )
			{
				intResult = kCVSEngineUnknownExceptionErr;
			}
			
		return intResult;
	
	}


void
CCVSEntries::WalkEntryList( CRadixTreeNodeRec 	*inCurNode, 
														const UInt32 				inFileLen, 
														char 								*inBuffer, 
														const UInt32 				inBufLen, 
														UInt32 							*inOutBufCount, 
														UInt32 							*inOutFileOffset, 
														UInt32 							*inOutDirCount )
	{
	
		OSErr	intResult = noErr;
		UInt32	intCount = 0;
		
		Assert_( ::StackSpace() > kMinStackSize );
		
		if ( inCurNode != NULL )
			{
		
				Assert_( inOutBufCount != NULL );
				Assert_( inOutFileOffset != NULL );
				
				Assert_( (*inOutBufCount) < inBufLen );
				
				if ( inCurNode->fSyntheticNode == false )
					{
					
						if ( inCurNode->fContextPtr != NULL )
							{
							
								if ( ((*inOutBufCount) + inCurNode->fContextDataLen + 1) > inBufLen )
									{
									
										if ( ((*inOutBufCount) + (*inOutFileOffset)) > inFileLen )
											{
											
												intResult = ::SetEOF( mFileRef, (*inOutBufCount) + (*inOutFileOffset) );
												
												if ( intResult != noErr )
													throw intResult;
													
											}
										
										intCount = *inOutBufCount;
										intResult = FSWriteOffset( mFileRef, &intCount, inBuffer, *inOutFileOffset );
										
										Assert_( intCount == (*inOutBufCount) );
										
										if ( intResult != noErr )
											throw intResult;
											
										(*inOutFileOffset) += (*inOutBufCount);
										
										*inOutBufCount = 0;
										
										//  Yielding is disabled for writing entries files because
										//  that should be an "atomic" operation. If we implemented
										//  decent read/write locking schemes, we wouldn't have this 
										//  problem but this is much simpler.
										
										//  Call our yield function here
										
										// intResult = Yield();
										
										//if ( intResult != noErr )
										//	throw intResult;
										
									}
									
								::BlockMoveData( inCurNode->fContextPtr, &inBuffer[*inOutBufCount], inCurNode->fContextDataLen );
								(*inOutBufCount) += inCurNode->fContextDataLen;
								
								inBuffer[(*inOutBufCount)++] = kEntriesTerminatorByte;
									
								//  Keep track of how many directory entries we write
								
								if ( (inCurNode->fContextDataLen > 2) && (((char *)inCurNode->fContextPtr)[0] == 'D') && (((char *)inCurNode->fContextPtr)[1] == '/') )
									{
										(*inOutDirCount)++;
									}
									
							}
				
					}
					
				WalkEntryList( inCurNode->fChildren, inFileLen, inBuffer, inBufLen, inOutBufCount, inOutFileOffset, inOutDirCount );
				WalkEntryList( inCurNode->fSiblings, inFileLen, inBuffer, inBufLen, inOutBufCount, inOutFileOffset, inOutDirCount );

				Assert_( ((inCurNode->fKeyData != NULL) && (inCurNode->fKeyDataLen > 0)) || ((inCurNode->fKeyData == NULL) && (inCurNode->fKeyDataLen == 0)) );
				Assert_( ((inCurNode->fContextPtr != NULL) && (inCurNode->fContextDataLen > 0)) || ((inCurNode->fContextPtr == NULL) && (inCurNode->fContextDataLen == 0)) );

			}
	
	}

	
Boolean
CCVSEntries::AreYouThisDirectory( const FSSpec *inWorkingDirectory )
	{
	
		Boolean bResult = false;
		
		Assert_( inWorkingDirectory != NULL );
		// Assert_( FSpIsValid(mWorkingDir) );
		
		if ( (inWorkingDirectory != NULL) && (FSpIsValid(mWorkingDir)) )
			{
				bResult = FSpIsEqual( inWorkingDirectory, &mWorkingDir );					
			}
			
		return bResult;
		
	}