//################################################################################
//
//  Copyright (C) 1997-1998 Netscape Communications Corporation.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.
//
//################################################################################


/* -------------------------------------------------------------------------------

	File:				UCVSRepositoryUtils.cp
	
	Authors:		Scott M. Silver and Ram Rajadhyaksha
	
	Description:
	
		This file contains all the internal utility functions for CCVSRepository.
		 
	Modification History:

		10 Jan 1998		Ram Rajadhyaksha
		
					Rewrote CreateItemFromCVSPath() to fix aliasing problems and correctly
					generate CVS Root files.
					
					Removed stupid UCVSRepositoryUtils class.
					
					Changed many function names from XX to CUXX.
					
		15 Dec 1998		Ram Rajadhyaksha
		
					Fixed bug in CreateItemFromCVSPath. First, it would improperly exit
					if the remote repository path was longer than the local one. Second,
					it wouldn't return an error even though it did exit because of the first
					problem.
					
-------------------------------------------------------------------------------- */


#include <ctype.h>
#include <string.h>

#include "MoreFiles.h"
#include "FileCopy.h"

#include "StringUtils.h"

#include "CVSEngineUtils.h"
#include "UCVSRepositoryUtils.h"


unsigned char kFakeRepositoryPathPStr[] = "\pCVSROOT/Emptydir";


UInt32
TruncateFileNameCStr(	const char* inSrcName, char* inDestName, const UInt32 inDesiredNameLen )
	{
		
		UInt32	srcNameLen;
		UInt32 	retVal;
			
		srcNameLen = strlen(inSrcName);
		
		retVal = TruncateFileName(inSrcName, srcNameLen, inDestName, inDesiredNameLen);
		
		*(inDestName + inDesiredNameLen) = '\0';
		
		return (retVal);
		
	}


UInt32								
TruncateFileNamePStr( const Str255 inSrcName, Str255 inDestName, const UInt32 inDesiredNameLen )
	{

		UInt32 retVal;
		
		retVal = TruncateFileName((char*) inSrcName + 1, *inSrcName, (char*) inDestName + 1, inDesiredNameLen);
		
		*inDestName = inDesiredNameLen;
		
		return (retVal); 

	}


UInt32
TruncateFileName( const char *inSrcName, const UInt32 inSrcNameLen, char *inDestName, const UInt32 inDesiredNameLen )
	{
	
		UInt32 	intCopyLen = 0;
		
		if ( inDesiredNameLen > 0 )
			{
				if ( inSrcNameLen > inDesiredNameLen )
					{
						intCopyLen = inDesiredNameLen / 2;
						
						::BlockMoveData( inSrcName, inDestName, intCopyLen );
						::BlockMoveData( inSrcName + (inSrcNameLen - intCopyLen), inDestName + (inDesiredNameLen - intCopyLen), intCopyLen );
						
						//  If the dest size is odd, this case will happen
						
						if ( (inDesiredNameLen - intCopyLen) > intCopyLen )
							{
								inDestName[intCopyLen] = '';
								Assert_( (intCopyLen + 1) == (inDesiredNameLen - intCopyLen) );
							}
						else if ( intCopyLen > 0 )
							{
								inDestName[intCopyLen] = '';
							}
					}
				else if ( inSrcNameLen == inDesiredNameLen )
					{
						::BlockMoveData( inSrcName, inDestName, inSrcNameLen );
					}
			}
			
		return inDesiredNameLen;
		
	}


char *
CUFilterErrorHeader( const char* inName )
	{
	
		const char* cp;
		
		for (cp = inName; *cp != ':' && *cp != '\0'; cp++) {;}
		
		if (*cp == '\0')
		
			return ((char*)inName);
		
		else
			{	
				// skip the colon then skip spaces after the colon
				for (cp = cp + 1; isspace(*cp); cp++) {;}

				return ((char*)cp);
			}
	
	}

	
//################################################################################
//################################################################################
#pragma mark -
#pragma mark CVS PATH ROUTINES
		

	
OSErr CUCreateItemFromCVSPath(	const Str255											inUserNamePStr,
																const Str255											inServerNamePStr,
																const UInt16											inPort,
																const AuthenticationMethod				inAuthMethod,
																const char												*inLocalPathStr,
																const UInt32											inLocalPathLen,
																const char												*inRemotePathStr,
																const UInt32											inRemotePathLen,
																const char												*inOffsetPathStr,
																const UInt32											inOffsetPathLen,
																const char												*inCVSRootStr,
																const UInt32											inCVSRootLen,
																const	FSSpec											*inProjectRoot,
																FSSpec														*outTargetSpec,
																const CCVSEngineFileCreateInfo		*inFileCreateInfo,																		
																Boolean														*outCreatedItem)
	{

/*
**  CreateItemFromCVSLocalPath() takes a CVS path name returned from the server
**  and creates the directory hierarchy (if necessary) and files.
**
**  Note that the CVS server returns both a local and remote path name. This
**  function correctly deals with this (well, I hope so at least) and such.
**
**  IMPORTANT: Do not mess with the paths that the CVS server returns before
**  passing them to CreateItemFromCVSLocalPath(). Just pass it the raw paths.
**  There is error checking here that deals with incorrectly formatted path names.
*/

/*
**  Setting the "Repository" file to the correct path is somewhat of a mystery.
**  I'm not exactly sure what magic the Unix client does but from what I've seen 
**  it follows these rules:
**
**  If the directory is not alias, put it's relative remote path name in the 
**  repository file. (relative to the root)
**
**  If the directory is aliased AND has no corresponding directory in the
**  repository, put "CVSROOT/EmptyDir" in the Repository file. Else if the 
**  directory does exist in the repository, put the relative (to the root)
**  path name.
**
**  Note that some versions of GNU CVS still use the full path name in the 
**  Repository file. We only put in the partial path name because there are 
**  various complications.
*/

		OSErr				result 																= noErr;
		
		char				strRepositoryPath[kMaxPathSize] 			= "";
		char				strLocalPath[kMaxPathSize] 						= "";
		
		Str255			pstrFileName													= "\p";
		Str255			pstrItemName													= "\p";
		Str255			pstrTemp															= "\p";
		Str255			pstrRoot															= "\p";
						
		UInt32			intIndex															= 0;
		UInt32			intMaxIndex														= 0;
		UInt32			intLen																= 0;
		UInt32			intMaxRemoteItems											= 0;
		UInt32			intRepositoryPathLen									= 0;
		UInt32			intLocalPathLen												= 0;
		
		Boolean			bIsAlias															= false;
		Boolean			bCreatedItem													= false;
		Boolean			bIsDirectory													= false;
		Boolean			bTruncated														= false;
		FSSpec			fsCurSpec															= { 0, 0, "\p" };
		
		CInfoPBRec	pb;
		
		if ( 	(inLocalPathStr != NULL) && (inLocalPathLen > 0) && 
					(inRemotePathStr != NULL) && (inRemotePathLen > 0) && 
					(inCVSRootStr != NULL) &&	// inCVSRootLen can be 0 because that signifies '/'
					
					/* (inRemotePathLen > inLocalPathLen) */ // Bug Fix
					
					(inRemotePathLen > inCVSRootLen) && 
					(inProjectRoot != NULL) && (outTargetSpec != NULL) &&
					(outCreatedItem != NULL) )
			{			
				outTargetSpec->name[0] = 0;
				outTargetSpec->vRefNum = 0;
				outTargetSpec->parID = 0;
															
//  Get the dirID of the root.
												
				fsCurSpec = *inProjectRoot;
				
				pb.hFileInfo.ioCompletion = NULL;
				pb.hFileInfo.ioNamePtr = fsCurSpec.name;
				pb.hFileInfo.ioVRefNum = fsCurSpec.vRefNum;
				pb.hFileInfo.ioDirID = fsCurSpec.parID;
				pb.hFileInfo.ioFDirIndex = 0;
				
				result = ::PBGetCatInfoSync( &pb );

//  Here we need to find out if the CVS local path has a one-to-one 
//  relationship with the remote path, or if it is an alias. 
//  Since the local path is relative, and has no file name, and the remote
//  path is fully qualified and has a file name, we chop off the CVSROOT and
//  file name for the remote path.

				if ( result == noErr )
					{
						if ( inLocalPathLen < kMaxPathSize )
							{							
								::BlockMoveData( inLocalPathStr, strLocalPath, inLocalPathLen );
								intLocalPathLen = inLocalPathLen;

//  Call StripCVSPath() to remove those './' path items.
										
								result = CUStripCVSPath( strLocalPath, &intLocalPathLen );
							}
						else 
							{
								result = kCVSEnginePathOverflow;
							}						
					}
					
//  If the CVS command was not run in the root directory, we need
//  to deal with this because it screws with the path names.
	
				if ( (result == noErr) && (inOffsetPathStr != NULL) && (inOffsetPathLen > 0) )
					{										
						if ( (kMaxPathSize - (intLocalPathLen + 1)) > inOffsetPathLen )
							{										
								if ( intLocalPathLen > 0 )
									{												
										::BlockMoveData( strLocalPath, &strLocalPath[inOffsetPathLen + 1], intLocalPathLen );
										strLocalPath[inOffsetPathLen] = '/';
										intLocalPathLen++;											
									}
								
								::BlockMoveData( inOffsetPathStr, strLocalPath, inOffsetPathLen );
								intLocalPathLen += inOffsetPathLen;												
							}											
						else 
							{
								result = kCVSEnginePathOverflow;	
							}						
					}
											
//  StripCVSPath() doesn't remove the trailing '/' 
//  for good reasons, so we remove it here.								
										
				if ( result == noErr )
					{					
						if ( intLocalPathLen > 0 )
							if ( strLocalPath[intLocalPathLen - 1] == '/' )
								intLocalPathLen--;							
					}		
										
							
				if ( result == noErr )
					{					
						if ( inRemotePathLen < kMaxPathSize )
							{

//  Strip off the CVSROOT string
					
								intRepositoryPathLen = inRemotePathLen - (inCVSRootLen + 1);
								::BlockMoveData( &inRemotePathStr[inCVSRootLen + 1], strRepositoryPath, intRepositoryPathLen );

								result = CUStripCVSPath( strRepositoryPath, &intRepositoryPathLen );
							}
						else 
							{
								result = kCVSEnginePathOverflow;
							}
					}

//  If the remote path ends with a '/', then the path
//  specifies a directory, not a file.
				
				if ( result == noErr )
					{				
						if ( intRepositoryPathLen > 0 )
							{
								if ( strRepositoryPath[intRepositoryPathLen - 1] == '/' )
									{								
										bIsDirectory = true;				
										intRepositoryPathLen--;	
									}		
							}
						else
							{
								/* bug fix */
								
								/* if the file name is 0 length, the path must be
								   referring to the root directory */
								   
								bIsDirectory = true;
							}	
					}
									
				if ( (result == noErr) && (!bIsDirectory) )
					{
							
//  Strip off the file name (it could also be a directory name)
				
						CUFastGetLastCVSPathItem( strRepositoryPath, intRepositoryPathLen, pstrFileName );
						
//  The CVS repository path should never be shorter than
//  the file name. (even it the CVSROOT is '/')
													
						Assert_( pstrFileName[0] <= intRepositoryPathLen );
								
						if ( pstrFileName[0] <= intRepositoryPathLen )
							intRepositoryPathLen -= pstrFileName[0];
	
						if ( intRepositoryPathLen > 0 )
							if ( strRepositoryPath[intRepositoryPathLen - 1] == '/' )
								intRepositoryPathLen--;							
					}
										
//  Compare the local and remote paths. If they are different
//  then we have an alias. I'm hoping that this is a reliable
//  way to do this.
						
				if ( result == noErr )
					{							
						if ( intLocalPathLen != intRepositoryPathLen )
							{
								bIsAlias = true;			
							}
						else
							{							
								if ( strncmp(strLocalPath, strRepositoryPath, intLocalPathLen) != 0 )
									{
										bIsAlias = true;			
									}						
							}
					}

//  This is the annoying part. In order to write the "Repository" file in the
//  CVS folder correctly, we need to analyze the local and remote paths. If the
//  local path is an alias, we determine which directories in it's path are
//  "imaginary" versus "real". "Imaginary" directories do not have an actual
//  directory in the repository. For those, we write a "CVSROOT/EmptyDir" repository
//  path. For the real directories, we write the real path.
				
				if ( result == noErr )
					{
						intMaxIndex = CUCountCVSPathItems( strLocalPath, intLocalPathLen );
						intMaxRemoteItems = CUCountCVSPathItems( strRepositoryPath, intRepositoryPathLen );
					}
					
				if ( result == noErr )
					for ( intIndex = 1; (intIndex <= intMaxIndex) && (result == noErr); intIndex++ )
						{
							bCreatedItem = false;
																								
//  GetCVSPathItem() extracts the next item name in a path string.
//  Ie: a from a/b/c. 
							
							result = CUGetCVSPathItem( strLocalPath, intLocalPathLen, intIndex, pstrItemName );

#ifdef CVSENGINE_USE_MIME_FILE_NAMES						
							if ( result == noErr )
								{
									UInt32 intItemNameLen = 0;
									
									intItemNameLen = pstrItemName[0];
									CUTranslateToMacName( (char *)&pstrItemName[1], intItemNameLen );
									pstrItemName[0] = intItemNameLen;
								}
#endif

//  Make sure the item we're looking for is a folder and that it has a valid Mac name.
									
							if ( result == noErr )
								{		
									result = CUIsValidMacName( pstrItemName );						
								}
								
							if ( result == noErr )	
								{								
									if ( pstrItemName[pstrItemName[0]] == '/' )
										{
											pstrItemName[0]--;
										}
										
									pb.hFileInfo.ioNamePtr = pstrItemName;
								
									result = ::PBGetCatInfoSync( &pb );

//  If it doesn't exist, create it.
											
									if ( (result == fnfErr) || (result == dirNFErr) )
										{
											fsCurSpec.parID = pb.hFileInfo.ioDirID;
											fsCurSpec.vRefNum = pb.hFileInfo.ioVRefNum;
											fsCurSpec.name[0] = 0;
											PStrConcat( fsCurSpec.name, pstrItemName, sizeof(fsCurSpec.name) );
						
											result = ::FSpDirCreate( &fsCurSpec, smSystemScript, &pb.hFileInfo.ioDirID );

											if ( result == noErr )
												{
													bCreatedItem = true;
													*outCreatedItem = true;
												}
												
//  Set the folder to display in list view

											if ( inFileCreateInfo->mFolderOptions.mFoldersGetListView )
												{
													if ( result == noErr )
														result = FSpSetFolderListView( &fsCurSpec, kViewFolderByName );
												}
										}
									else if ( result == noErr )
										{
											fsCurSpec.parID = pb.hFileInfo.ioFlParID;
											fsCurSpec.vRefNum = pb.hFileInfo.ioVRefNum;
											fsCurSpec.name[0] = 0;
											PStrConcat( fsCurSpec.name, pstrItemName, sizeof(fsCurSpec.name) );

//  If we found a file instead of a directory, then bail.
												
											if ( (pb.dirInfo.ioFlAttrib & ioDirMask) == 0 )
												{
													result = kCVSEngineAlreadyExists;
												}										
										}
										
									if ( result == noErr ) 
										{
											if ( bCreatedItem || (CUIsRepositoryDir(pb.hFileInfo.ioVRefNum, pb.hFileInfo.ioDirID) == false) )
												{

//  Create all the CVS internal information
															
													result = CUCreateRepositoryDir( pb.hFileInfo.ioVRefNum, pb.hFileInfo.ioDirID, 'CWIE', 'TEXT', smSystemScript );

//  Set the CVS Root

													if ( result == noErr )
														{
															if (inCVSRootLen == 0)
																{
																	pstrTemp[0] = 0;
																	PStrConcat( pstrTemp, "\p/", sizeof(pstrTemp) );
																} 
															else
																{
																	if ( inCVSRootLen < 255 )
																		{
																			pstrTemp[0] = inCVSRootLen;
																		}
																	else 
																		{
																			pstrTemp[0] = 255;
																		}
																		
																	::BlockMoveData( inCVSRootStr, &pstrTemp[1], pstrTemp[0] );														
																}
														
															result = CUCreateRootPStr( inAuthMethod, inUserNamePStr, inServerNamePStr, inPort, pstrTemp, pstrRoot );
															
															if ( result == noErr )
																{
																	result = FSpSetRoot( &fsCurSpec, pstrRoot );
																}
																
															Assert_( result == noErr );
														}

//  Set the Repository
																						
													if ( result == noErr )
														{																								
															if ( (intMaxIndex - intIndex) < intMaxRemoteItems )
																{
																	intLen = 255;
																	result = CUCopyCVSPathItems( strRepositoryPath, intRepositoryPathLen, (char *)&pstrTemp[1], &intLen, intMaxRemoteItems - (intMaxIndex - intIndex) );																																	

																	if ( result == noErr )
																		{
																			pstrTemp[0] = intLen;
																		}
																}																												 				 	
															else 
																{
																	PStrCopy( kFakeRepositoryPathPStr, pstrTemp, sizeof(pstrTemp) );
																}
														
//  We used to write the relative path name to
//  the repository. Now we include the whole pathname
//  for compatibility with GNU MacCVS.
													
															if ( result == noErr )
																{																
																	if ( (255 - pstrTemp[0]) > (inCVSRootLen + 1) )
																		{
																			::BlockMoveData( &pstrTemp[1], &pstrTemp[inCVSRootLen + 2], pstrTemp[0] );
																			::BlockMoveData( inCVSRootStr, &pstrTemp[1], inCVSRootLen );
																			pstrTemp[inCVSRootLen + 1] = '/';
																			pstrTemp[0] += inCVSRootLen + 1;
																		}																												
																}

															if ( result == noErr )
																{														
																	while ( (pstrTemp[0] > 0) && (pstrTemp[pstrTemp[0]] == '/') )
																		pstrTemp[0]--;
																		
																	Assert_( pstrTemp[0] > 0 );
																	
																	result = FSpSetRepository( &fsCurSpec, pstrTemp );
																}													
														}
												}												
										}
								}																											
						} /* for */												
	
//  Now create the file if we need to...

//  The pb should contain the information for the
//  last directory we scanned/created.
										
					if ( (result == noErr) && (!bIsDirectory) )
						{
							fsCurSpec.vRefNum = pb.hFileInfo.ioVRefNum;
							fsCurSpec.parID = pb.hFileInfo.ioDirID;
							fsCurSpec.name[0] = 0;

							if ( pstrFileName[0] > 0 )
								{
									if ( pstrFileName[pstrFileName[0]] == '/' )
										{
											pstrFileName[0]--;
										}
								}
								
							pstrItemName[0] = 0;
							PStrConcat( pstrItemName, pstrFileName, sizeof(pstrItemName) );						
							Assert_( pstrItemName[0] > 0 );

#ifdef CVSENGINE_USE_MIME_FILE_NAMES
							{
								UInt32 intItemNameLen = 0;
								
								intItemNameLen = pstrItemName[0];
								CUTranslateToMacName( (char*)&pstrItemName[1], intItemNameLen );
								pstrItemName[0] = intItemNameLen;
							}
#endif
							
							if ( pstrItemName[0] > kMaxFileNameSize )
								{
									TruncateFileNamePStr( pstrFileName, pstrItemName, kMaxFileNameSize );
									bTruncated = true;
								}
								
							result = CUIsValidMacName( pstrItemName );
							
							if ( result == noErr )
								{
									PStrConcat( fsCurSpec.name, pstrItemName, sizeof(fsCurSpec.name) );
							
									result = ::PBGetCatInfoSync( &pb );
								
									if ( (result == fnfErr) || (result == dirNFErr) )
										{																					
											//  Is it safe to use FSpCreateResFile() 
											//  since it follows the PMSP?
								
											::FSpCreateResFile( &fsCurSpec, inFileCreateInfo->mFileCreator, inFileCreateInfo->mFileType, inFileCreateInfo->mScriptType );
											result = ::ResError();
								
											if ( result == noErr )
												{
													*outCreatedItem = true;	
												}									
										}						
									else if ( result == noErr )
										{
											//  If we found a directory when we were 
											//  looking for a file, die with an error.
											
											if ( (pb.dirInfo.ioFlAttrib & ioDirMask) != 0 )
												{
													result = kCVSEngineAlreadyExists;
												}
										}
								}
																									
					}
		
				if ( result == noErr )
					{					
						*outTargetSpec = fsCurSpec;					
				
						if ( bTruncated )
							{
								result = kCVSEngineFileNameTruncatedErr;
							}
					}				
			}		
		else 
			{
				result = kCVSEngineBadParamErr;
			}
							
		return result;
	
	}


UInt32
CUCountCVSPathItems( const char *inPathStr, const UInt32 inPathLen )
	{
	
		UInt32		result 	= 0;
		UInt32		index		= 0;							
		Boolean		done 		= false;

		if ( (inPathStr != NULL) && (inPathLen > 0) )
			{			
				while ( !done )
					{					
						if ( index >= inPathLen )
							{							
								result++;
								done = true;							
							}
						else if ( inPathStr[index] == '/' )
							{							
								if ( index != (inPathLen - 1) )
									result++;
								
								index++;									
							}
						else
							{							
								while ( (index < inPathLen) && (inPathStr[index] != '/') )
									index++;							
							}
					
					}					
			}
				
		return result;		
	}
	

OSErr						
CUCopyCVSPathItems( const char *inPathStr, const UInt32 inPathLen, char *outItemStr, UInt32 *inOutItemLen, const UInt32 inItemCount ) 
	{

		OSErr			result 				= noErr;
		
		UInt32		curIndex			= 0;
		UInt32		maxLen				= 0;
		UInt32		copiedItems		= 0;
		UInt32		charIndex			= 0;
	
		if ( (inPathStr != NULL) && (inPathLen > 0) && (outItemStr != NULL) && (inOutItemLen != NULL) && ((*inOutItemLen) > 0) && (inItemCount > 0) )
			{			
				maxLen = *inOutItemLen;
				(*inOutItemLen) = 0;
				
				while ( (copiedItems < inItemCount) && (result == noErr) && (charIndex < inPathLen) )
					{					
						if ( inPathStr[charIndex] == '/' )
							copiedItems++;
							
						if ( curIndex < maxLen )
							{
							
								outItemStr[curIndex] = inPathStr[charIndex];
								curIndex++;
								
							}
						else result = kCVSEnginePathOverflow;
							
						charIndex++;												
					}
/*
**  Should we return an error if we didn't copy the requested
**  number of items?
*/			
				if ( result == noErr )
					*inOutItemLen = curIndex;							
			}
		else result = kCVSEngineBadParamErr;
						
		return result;
	}
	
	
OSErr
CUGetCVSPathItem( const char *inPathStr, const UInt32 inPathLen, const UInt32 inItemIndex, Str255 outItemPStr )
	{
	
/*
**  GetCVSPathItem returns the nth item in a path string. Ie. in the path "a/b/c",
**  "b/" would be the second item. If the item string return has a trailing '/'
**  character, it is obviously a directory. GetCVSPathItem() assumes that the path
**  string is in correct format, does not contain any multiple sequences of '/' 
**  characters in a row and does not begin with a '/' character.
*/
	
		OSErr			result 					= noErr;
		Boolean		done 						= false;
		Boolean		foundIt					= false;
							
		UInt32		curItemIndex 		= 0;
		UInt32		charIndex 			= 0;
			
		if ( (inPathStr != NULL) && (inPathLen > 0) && (inItemIndex > 0) && (outItemPStr != NULL) )
			{		
				outItemPStr[0] = 0;
					
				curItemIndex = 1;
				
				while ( (charIndex < inPathLen) && (!foundIt) )
					{

						if ( (curItemIndex == inItemIndex) && (inPathStr[charIndex] != '/') )
							{							
								while ( (charIndex < inPathLen) && (inPathStr[charIndex] != '/') )
									{								
										if ( outItemPStr[0] < 255 )
											{											
												outItemPStr[0]++;
												
												outItemPStr[outItemPStr[0]] = inPathStr[charIndex];											
											}
										
										charIndex++;								
									}
									
								if ( (charIndex < inPathLen) && (inPathStr[charIndex] == '/') )
									{									
										if ( outItemPStr[0] < 255 )
											{											
												outItemPStr[0]++;
												outItemPStr[outItemPStr[0]] = '/';												
											}																						
									}
									
								foundIt = true;							
							}						
						else 
							{							
								while ( (charIndex < inPathLen) && (inPathStr[charIndex] != '/') )
									charIndex++;
							
								curItemIndex++;
								
								if ( (charIndex < inPathLen) && (inPathStr[charIndex] == '/') )
									charIndex++;									
							}																		
					}
					
				if ( !foundIt )
					result = kCVSEngineItemNotFound;	
			
			}
			
		return result;			
	}
	

OSErr
CUGetLastCVSPathItem( const char *inPathStr, const UInt32 inPathLen, Str255 outItemPStr )
	{
	
		OSErr 	result 		= noErr;
		UInt32	itemCount = 0;
	
		if ( (inPathStr != NULL) && (inPathLen > 0) && (outItemPStr != NULL) )
			{
			
				outItemPStr[0] = 0;
				
				itemCount = CUCountCVSPathItems( inPathStr, inPathLen );
				
				if ( itemCount > 0 )
					{
					
						result = CUGetCVSPathItem( inPathStr, inPathLen, itemCount, outItemPStr );
						
					}
				else result = kCVSEngineBadParamErr;
					
			}
		else result = kCVSEngineBadParamErr;
		
		return result;
			
	}
	
	

OSErr
CUFastGetLastCVSPathItem( const char *inPathStr, const UInt32 inPathLen, Str255 outItemPStr )
	{
	
		OSErr 	result 		= noErr;
		UInt32	copyLen	 	= 0;
		UInt32	index			= 0;
		
		if ( (inPathStr != NULL) && (inPathLen > 0) && (outItemPStr != NULL) )
			{	
				outItemPStr[0] = 0;
				
				index = inPathLen - 1;
				
				if ( inPathStr[index] == '/' )
					index--;
					
				while ( (index > 0) && (inPathStr[index] != '/') )
					index--;

				if ( inPathStr[index] == '/' )
					index++;
										
				if ( index < inPathLen )
					{					
						copyLen = inPathLen - index;
						
						if ( copyLen > 255 )
							copyLen = 255;
						
						outItemPStr[0] = copyLen;
						
						if ( copyLen > 0 )
							::BlockMoveData( &inPathStr[index], &outItemPStr[1], outItemPStr[0] );						
					}										
			}
		else result = kCVSEngineBadParamErr;
		
		return result;			
	}

	
//################################################################################
//################################################################################
#pragma mark -
#pragma mark FILE SYSTEM ROUTINES


OSErr
FSWriteOffset( const SInt32 inFileRef, UInt32 *inOutCount, void *inData, const UInt32 inOffset, const UInt32 inCacheFlags )
	{
	
		OSErr						result 		= noErr;
		HParamBlockRec	pb;
		
		if ( (inFileRef > 0) && (inData != NULL) && (inOutCount != NULL) )
			{
			
				pb.ioParam.ioCompletion 	= NULL;
				pb.ioParam.ioVRefNum			= 0x00000000;
				pb.ioParam.ioRefNum				= inFileRef;
				pb.ioParam.ioBuffer 			= (char *)inData;
				pb.ioParam.ioReqCount 		= *inOutCount;
				
				pb.ioParam.ioPosMode 			= fsFromStart;
				
				/*
				**  If the user specified cache settings, use them
				*/
				
				if ( inCacheFlags & pleaseCacheMask )
					{
						pb.ioParam.ioPosMode |= pleaseCacheMask;
					}
				else if ( inCacheFlags & noCacheMask )
					{
						pb.ioParam.ioPosMode |= noCacheMask;
					}
					
				pb.ioParam.ioPosOffset 		= inOffset;
						
				result = ::PBWriteSync( (ParmBlkPtr)&pb );
				
				*inOutCount = pb.ioParam.ioActCount;
				
			}
					
		else result = kCVSEngineBadParamErr;
		
		return result;
	
	}


OSErr
FSReadOffset( const SInt32 inFileRef, UInt32 *inOutCount, void *inData, const UInt32 inOffset, const UInt32	inCacheFlags )
	{
	
		OSErr						result 		= noErr;
		HParamBlockRec	pb;
		
		if ( (inFileRef > 0) && (inData != NULL) && (inOutCount != NULL) )
			{
			
				pb.ioParam.ioCompletion 	= NULL;
				pb.ioParam.ioVRefNum			= 0x00000000;
				pb.ioParam.ioRefNum				= inFileRef;
				pb.ioParam.ioBuffer 			= (char *)inData;
				pb.ioParam.ioReqCount 		= *inOutCount;
	
				pb.ioParam.ioPosMode 			= fsFromStart;

				/*
				**  If the user specified cache settings, use them
				*/
				
				if ( inCacheFlags & pleaseCacheMask )
					{
						pb.ioParam.ioPosMode |= pleaseCacheMask;
					}
				else if ( inCacheFlags & noCacheMask )
					{
						pb.ioParam.ioPosMode |= noCacheMask;
					}

				pb.ioParam.ioPosOffset 		= inOffset;
						
				result = ::PBReadSync( (ParmBlkPtr)&pb );
				
				*inOutCount = pb.ioParam.ioActCount;
				
			}
					
		else result = kCVSEngineBadParamErr;
		
		return result;
	
	}


OSErr
FSAsyncCopy( FSAsyncCopyContextPtr inPB )
	{
	
		OSErr	result = noErr;
		
		static IOCompletionUPP	sCompletionProc;
				
		if ( inPB == NULL )
			{
				result = kCVSEngineBadParamErr;
			}
			
		if ( result == noErr )
			{			
				if ( (inPB->sSrcRef == 0) || (inPB->sDestRef == 0) || (inPB->sBuffer == NULL) || (inPB->sBufMax == 0) )
					{
						result = kCVSEngineBadParamErr;
					}
			}

		if ( result == noErr )
			{
				if ( inPB->sCopyLen == 0 )
					{
						/*  whhoops, don't forget to set this... */
						inPB->sFinished = true;
						return noErr;
					}
			}
			
		if ( result == noErr )
			{
				if ( sCompletionProc == NULL )
					{
				 		sCompletionProc = NewIOCompletionProc( FSAsyncCopyCompletion );
			
						if ( sCompletionProc == NULL )
							{
								result = memFullErr;
							}
					}
			}
			
		if ( result == noErr )
			{
				inPB->sBytesCopied									= 0;
				inPB->sResult												= noErr;
				inPB->sFinished											= false;
				inPB->sKillFlag											= false;
				
				inPB->sPB.ioParam.ioCompletion 			= sCompletionProc;
				inPB->sPB.ioParam.ioVRefNum					= 0;
				inPB->sPB.ioParam.ioRefNum					= inPB->sSrcRef;
				inPB->sPB.ioParam.ioBuffer 					= (inPB->sBuffer);
				inPB->sPB.ioParam.ioPosMode 				= fsFromStart;
				inPB->sPB.ioParam.ioPosOffset 			= inPB->sSrcOffset;

				inPB->sPB.ioParam.ioReqCount 			= inPB->sCopyLen;
				
				if ( inPB->sPB.ioParam.ioReqCount > inPB->sBufMax )
					{
						inPB->sPB.ioParam.ioReqCount = inPB->sBufMax;
					}						
						
				result = PBReadAsync( (ParmBlkPtr)inPB );
				Assert_( result == noErr );
			}			
			
		return result;
			
	}	



#ifdef powerc
void FSAsyncCopyCompletion( register ParmBlkPtr inPB )
#else
void FSAsyncCopyCompletion( register ParmBlkPtr inPB:__A0 )	/* get PB from register A0 */
#endif
	{	
		FSAsyncCopyContextPtr sContextPtr = (FSAsyncCopyContextPtr)inPB;

		#ifdef __IODEBUGSTR	
		DebugStr( "\pFSAsyncCopyCompletion: Enter" );
		#endif
			
		//Assert_( inPB != NULL );
		
		if ( sContextPtr == NULL )
			{
				return;
			}
			
		if ( sContextPtr->sPB.ioParam.ioResult != noErr )
			{
				sContextPtr->sResult		= sContextPtr->sPB.ioParam.ioResult;
				sContextPtr->sFinished 	= true;
				
				/* If we got an error, bail immediately. */
				return;
			}
			
		if ( sContextPtr->sKillFlag == true )
			{
				#ifdef __IODEBUGSTR
				DebugStr( "\pFSAsyncCopyCompletion: Kill flag activated." );
				#endif
						
				sContextPtr->sResult = kCVSEngineOperationAbortedErr;
				sContextPtr->sFinished = true;
				return;
			}
			
		if ( sContextPtr->sPB.ioParam.ioRefNum == sContextPtr->sSrcRef )
			{
				sContextPtr->sSrcOffset += sContextPtr->sPB.ioParam.ioActCount;

				sContextPtr->sPB.ioParam.ioRefNum			= sContextPtr->sDestRef;
				sContextPtr->sPB.ioParam.ioReqCount 	= sContextPtr->sPB.ioParam.ioActCount;
				sContextPtr->sPB.ioParam.ioPosOffset 	= sContextPtr->sDestOffset;
					
				#ifdef __IODEBUGSTR
				DebugStr("\pFSAsyncCopyCompletion: Writing");			
				#endif
				
				PBWriteAsync( inPB );
			}
		else if ( sContextPtr->sPB.ioParam.ioRefNum == sContextPtr->sDestRef )
			{
				sContextPtr->sDestOffset += sContextPtr->sPB.ioParam.ioActCount;
				sContextPtr->sBytesCopied += sContextPtr->sPB.ioParam.ioActCount;
									
				if ( sContextPtr->sBytesCopied < sContextPtr->sCopyLen )
					{
						sContextPtr->sPB.ioParam.ioRefNum			= sContextPtr->sSrcRef;
						sContextPtr->sPB.ioParam.ioPosOffset 	= sContextPtr->sSrcOffset;
								
						sContextPtr->sPB.ioParam.ioReqCount 	= sContextPtr->sCopyLen - sContextPtr->sBytesCopied;
						
						if ( sContextPtr->sPB.ioParam.ioReqCount > sContextPtr->sBufMax )
							{
								sContextPtr->sPB.ioParam.ioReqCount = sContextPtr->sBufMax;
							}

						#ifdef __IODEBUGSTR
						DebugStr("\pFSAsyncCopyCompletion: Reading");	
						#endif
										
						PBReadAsync( inPB );
					}
				else
					{
						#ifdef __IODEBUGSTR
						DebugStr("\pFSAsyncCopyCompletion: Finished...we survived alive!");
						#endif
						
						/* returning the result should tell the caller we're done */
						sContextPtr->sResult = sContextPtr->sPB.ioParam.ioResult;		
						sContextPtr->sFinished = true;			
					}
			}
		else
			{
				/*  Uh oh, we should never get here. */
				sContextPtr->sResult = kCVSEngineUnknownExceptionErr;
				sContextPtr->sFinished = true;
			}
	}