// ===========================================================================
//	CLocalTreeView.cp
//  by Josef W. Wankerl
// ===========================================================================

#ifdef PowerPlant_PCH
	#include PowerPlant_PCH
#endif

#include "CLocalTreeView.h"

PP_Begin_Namespace_PowerPlant


// ---------------------------------------------------------------------------
//	 CLocalTreeView						Stream Constructor		  [public]
// ---------------------------------------------------------------------------

CLocalTreeView::CLocalTreeView(
	LStream*	inStream,
	ClassIDT	inImpID)
	
	: LTextGroupBox(inStream, inImpID),
	LDropArea(UQDGlobals::GetCurrentPort())
{
	// Start out with nobody available to receive a drop
	fHandler = NULL;
}


// ---------------------------------------------------------------------------
//	 CLocalTreeView						Parameterized Constructor [public]
// ---------------------------------------------------------------------------

CLocalTreeView::CLocalTreeView(
	const SPaneInfo&	inPaneInfo,
	const SViewInfo&	inViewInfo,
	Boolean				inPrimary,
	ResIDT				inTextTraitsID,
	ConstStringPtr		inTitle,
	ClassIDT			inImpID)
	
	: LTextGroupBox(inPaneInfo, inViewInfo, inPrimary, inTextTraitsID, inTitle, inImpID),
	LDropArea(UQDGlobals::GetCurrentPort())
{
	// Start out with nobody available to receive a drop
	fHandler = NULL;
}


// ---------------------------------------------------------------------------
//	 CLocalTreeView						Destructor				  [public]
// ---------------------------------------------------------------------------

CLocalTreeView::~CLocalTreeView()
{
}


// ---------------------------------------------------------------------------
//	 PointInDropArea												  [public]
// ---------------------------------------------------------------------------
//	Return whether a Point, in Global coords, is inside a DropArea

Boolean
CLocalTreeView::PointInDropArea(
	Point	inGlobalPt)
{
	// Valid drop area is anywhere within the view
	GlobalToPortPoint(inGlobalPt);
	return IsHitBy(inGlobalPt.h, inGlobalPt.v);
}


// ---------------------------------------------------------------------------
//	 HiliteDropArea												  [public]
// ---------------------------------------------------------------------------
//	Hilite a DropArea to indicate that it can accept the current Drag

void
CLocalTreeView::HiliteDropArea(
	DragReference	inDragRef)
{
	// Setup the drawing environment
	FocusDraw(this);
	ApplyForeAndBackColors();
	
	// Get the boundary of the view
	Rect	dropRect;
	CalcLocalFrameRect(dropRect);
	
	// Move the top of the hilight rectangle so it's inside the box boundary
	FontInfo	fontInfo;
	::GetFontInfo(&fontInfo);
	dropRect.top += fontInfo.ascent - 1;
	
	// Adjust the hilight area so it's completely inside the box boundary
	::MacInsetRect(&dropRect, 1, 1);
	StRegion	dropRgn(dropRect);
	
	// Now show it
	::ShowDragHilite(inDragRef, dropRgn, true);
	
	mIsHilited = true;
}


// ---------------------------------------------------------------------------
//	 UnhiliteDropArea												  [public]
// ---------------------------------------------------------------------------
//	Unhilite a DropArea when a drag leaves the area

void
CLocalTreeView::UnhiliteDropArea(
	DragReference	inDragRef)
{
	FocusDraw(this);
	ApplyForeAndBackColors();
	::HideDragHilite(inDragRef);
}


// ---------------------------------------------------------------------------
//	 ItemIsAcceptable												  [public]
// ---------------------------------------------------------------------------
//	Return whether a Drag item is acceptable

Boolean
CLocalTreeView::ItemIsAcceptable(
	DragReference	inDragRef,
	ItemReference	inItemRef)
{
	HFSFlavor		theFlavor;
	FlavorFlags		theFlags;
	Size			theSize;
	OSErr			err;
	
	// If nobody has been set to receive the drag, don't accept one
	if (fHandler == NULL) return false;
	
	// See if there's there's a file or folder being dragged
	err = ::GetFlavorFlags(inDragRef, inItemRef, flavorTypeHFS, &theFlags);
	if (err != noErr) return false;
	
	// Get which file or folder is being dragged
	theSize = sizeof(theFlavor);
	err = ::GetFlavorData(inDragRef, inItemRef, flavorTypeHFS, &theFlavor, &theSize, 0);
	if (err != noErr) return false;
	
	// Resolve any aliases, and in the process, find out if the drag is a folder
	Boolean isDirectory;
	Boolean isAlias;
	err = ResolveAliasFile(&theFlavor.fileSpec, true, &isDirectory, &isAlias);
	if (err != noErr) return false;
	
	// Accept only folders
	return isDirectory;
}


// ---------------------------------------------------------------------------
//	 DoDragReceive													  [public]
// ---------------------------------------------------------------------------
//	Receive the items from a completed Drag and Drop
//
//	This function gets called after items are dropped into a DropArea.
//	The drag contains items that are acceptable, as defined by the
//	DragIsAcceptable() and ItemIsAcceptable() member functions.

void
CLocalTreeView::DoDragReceive(
	DragReference	inDragRef)
{
	// Get the drag attributes
	DragAttributes	dragAttrs;
	::GetDragAttributes(inDragRef, &dragAttrs);
	
	// Get the first item reference that was in the drag
	ItemReference	itemRef;
	::GetDragItemReferenceNumber(inDragRef, 1, &itemRef);
	
	// Get which folder was dropped
	Size			theSize;
	HFSFlavor		theFlavor;
	OSErr			err;
	theSize = sizeof(theFlavor);
	err = ::GetFlavorData(inDragRef, itemRef, flavorTypeHFS, &theFlavor, &theSize, 0);
	
	// Resolve any aliases, and in the process, find out if the drag is a folder
	Boolean isDirectory;
	if (err == noErr)
	{
		Boolean isAlias;
		err = ResolveAliasFile(&theFlavor.fileSpec, true, &isDirectory, &isAlias);
	}
	
	// Accept only folders
	if ((err == noErr) && isDirectory)
	{
		// Remember which folder was dropped, and tell the listener to get it
		fDropped = theFlavor.fileSpec;
		fHandler->ListenToMessage('GtDr', NULL);
	}
}


void CLocalTreeView::SetPreferenceHandler(StDialogHandler *stHandler)
{
	// Remember who to tell when a folder is dropped
	fHandler = stHandler;
}

FSSpec *CLocalTreeView::GetDroppedFolder(void)
{
	// Give the listener the folder that was dropped
	return &fDropped;
}

PP_End_Namespace_PowerPlant