//################################################################################
//
//  Copyright (C) 1997-1998 Netscape Communications Corporation.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.
//
//
//		File: 		CMacCVSProjectView.cp
//	
//	  Authors: 	Dan Clifford and Ram Rajadhyaksha
//
//################################################################################


/* -------------------------------------------------------------------------------

	File:				CMacCVSProjectView.cp
	
	Authors:		Dan Clifford and Ram Rajadhyaksha
	
	Description:
	
		This file contains all the code for the neato-keen Finder-like display.
		 
	Modification History:

		21 Apr 1999		Josef W. Wankerl
		
					Added Reveal In Finder support.
		
		10 Jan 1998
		
					New gray scale look.

-------------------------------------------------------------------------------- */


#include <string.h>
#include <stdio.h>

#include <Scrap.h>

#include <LArray.h>
#include <URegions.h>
#include <UException.h>
#include <UAttachments.h>
#include <UGAColorRamp.h>
#include <LTableMonoGeometry.h>
#include <LTableMultiSelector.h>
#include <LTableArrayStorage.h>
#include <LNodeArrayTree.h>
#include <LDropFlag.h>
#include <LTextButton.h>

#include "ThrowOSErr.h"
#include "MacCVSPPConstants.h"
#include "StringUtils.h"
#include "URegionUtilities.h"
#include "UErrorDialog.h"

#include "CSpinIndicator.h"
#include "UMacCVSUtils.h"
#include "CPathPopup.h"

#include "CVSEngineUtils.h"
#include "CMacCVSDoc.h"
#include "CMacCVSProjectViewItem.h"
#include "CMacCVSProjectViewFile.h"
#include "CMacCVSProjectViewFolder.h"
#include "CMacCVSProjectViewRootFolder.h"

#include "CMacCVSProjectViewSelector.h"
#include "CMacCVSProjectView.h"
#include "UKeyScrollAttachments.h"

#include "Find_icon.h"

UInt32	CMacCVSProjectView::sYieldCount = 0;
Handle	CMacCVSProjectView::sFileStatusIconSuites[kMacCVSProjectorTotalCount] = { 0, 0, 0, 0, 0, 0, 0 };


CMacCVSProjectView::CMacCVSProjectView( LStream *inStream ): 

	LHierarchyTable( inStream )

	{
			
		mViewThread		 						= NULL;
		mViewEventQueue 					= NULL;
		mWatchList 								= NULL;

		mAnchorItem 							= NULL;
		mCurrentSortMethod  			= kNameSorted;
		mCreateNewView						= true;			
		mUserDraggedSelection			= false;
		
		mIconOffset 							= kCellIconOffset;
		mTextOffset 							= kCellTextOffset;
		mLockIconOffset						= kFileLockIconOffset;
		mStatusIconOffset					= kFileStatusIconOffset;
		mVersionOffset 						= kCellVersionTextOffset;
		mTagOffset 								= kCellTagTextOffset;
		mDateOffset 							= kCellModificationTextOffset;

		mWatchTime								= kDefaultBackgroundUpdateTime;
	
		mBGColor.red							= 0;
		mBGColor.green 						= 0;
		mBGColor.blue 						= 0;
		
//  Offset the flag rect by 2 so the drop flag is properly centered.

		::OffsetRect( &mFlagRect, 0, 2 );
		
		try
			{

				mBGColor = UGAColorRamp::GetColor( colorRamp_Gray1 );
				SetForeAndBackColors( NULL, &mBGColor );
			
//  Load the status icons if they are not already available.	
			
				if (sFileStatusIconSuites[0] == NULL) 
					{
					
						for ( UInt32 currentIcon = 0; currentIcon < kMacCVSProjectorTotalCount; currentIcon++ )
						
							::GetIconSuite( sFileStatusIconSuites + currentIcon, 
											kFirstFileStatusIconSuiteID + currentIcon, 
											svAllAvailableData);
						
					}

				mViewEventQueue = new LArray( sizeof(CMacCVSProjectViewEventRec) );
				ThrowIfNil_( mViewEventQueue );
				
				mWatchList = new LArray( sizeof(CMacCVSProjectViewFolder *) );
				ThrowIfNil_( mWatchList );
				
			}
		
		catch ( ... )
			{
			
				if ( mViewEventQueue != NULL )
					delete mViewEventQueue;
				
				if ( mWatchList != NULL )
					delete mWatchList;
								
				throw;
							
			}

	}


CMacCVSProjectView::~CMacCVSProjectView()
	{

//  Wait for the thread to die. Please see important comments
//  in CMacCVSProjectViewWindow::~CMacCVSProjectViewWindow().	

		if ( mViewThread != NULL )
			{
			
				mViewThread->ShutdownThread();
				
				while ( mViewThread != NULL )
					CThreadedTask::Yield();
			
			}
				
		delete mViewEventQueue;
		delete mWatchList;

		if ( mAnchorItem != NULL )
			delete mAnchorItem;

	}

	
void
CMacCVSProjectView::FinishCreateSelf( void )
	{

		mTableGeometry = new LTableMonoGeometry( this, kDefaultRowWidth, kDefaultRowHeight );
		mTableStorage = new LTableArrayStorage( this, sizeof(CMacCVSProjectViewItem *) );
		mTableSelector = new CMacCVSProjectViewSelector( this );
		
		// Insert a single column.
		InsertCols( 1, 0, nil, nil, false );
		
		//	Attachment for keyboard navigation
		AddAttachment( new UProjectKeyScrollAttachment(this) );
			
		LHierarchyTable::FinishCreateSelf();
				
	}


#pragma mark -
#pragma mark ACCESSORS


void
CMacCVSProjectView::Shutdown()
	{
	
		if ( mViewThread != NULL )
			{
			
				mViewThread->ShutdownThread();
				
				while ( mViewThread != NULL )
					CThreadedTask::Yield();
					
			}
			
	}
	
	
void
CMacCVSProjectView::UpdateProjectItem( const FSSpec *inTargetItem )
	{

		Assert_( inTargetItem != NULL );
		
		if ( inTargetItem != NULL )
			{

				CMacCVSProjectViewEventRec theEvent = { kViewUpdateProjectItemFSSpecEvent, 0, 0 };
										
				FSSpec *theSpecPtr = (FSSpec *)::NewPtr( sizeof(FSSpec) );
				ThrowIfNil_( theSpecPtr );
				
				::BlockMoveData( inTargetItem, theSpecPtr, sizeof(FSSpec) );
				theEvent.fEventData = theSpecPtr;
				
				PostProjectViewEvent( theEvent );
				
			}

	}
	

CMacCVSProjectViewFolder *
CMacCVSProjectView::GetRootItem()
	{

		return mAnchorItem;

	}


void
CMacCVSProjectView::SetRootItem( FSSpec *inAnchorFileSpec, FSSpec *inProjectRootFileSpec )
	{
		
		OSErr									result						= noErr;
		TableIndexT						maxRows						= 0x00000000,
													maxColumns 				= 0x00000000,
													rowIndex					= 0x00000000;
						
		CMacCVSProjectViewEventRec		fakeUpdateEvent		= { kViewUpdateProjectItemEvent, NULL };

//  First delete threads...
						
		if ( mViewThread != NULL ) 
			{
			
				mViewThread->ShutdownThread();
				
				while ( mViewThread != NULL )
					CThreadedTask::Yield();
					
//  Make sure the spinning indicator is off.

				BroadcastMessage( CSpinIndicator::msg_SpinNoMoreWaiters );

			}

		mWatchList->RemoveItemsAt( mWatchList->GetCount(), LArray::index_First );
		mViewEventQueue->RemoveItemsAt( mViewEventQueue->GetCount(), LArray::index_First );
		
//  Then delete the table.
	
		RemoveAllRows( true );
				
		if ( mAnchorItem != NULL ) 
			{
		
				delete mAnchorItem;
				mAnchorItem = NULL;
			
			}
				
		if ( inAnchorFileSpec != NULL ) 
			{
			
				if ( inProjectRootFileSpec != NULL )
					{
						//  Make sure the anchor is a decendent of the project root
						
						if ( !FSpIsChildOf(inAnchorFileSpec, inProjectRootFileSpec) )
							{
								ThrowOSErr( dirNFErr );
							}

						mProjectRootFileSpec = *inProjectRootFileSpec;
					}
				else
					{
						mProjectRootFileSpec = *inAnchorFileSpec;
					}
					
				if ( (inAnchorFileSpec->vRefNum != 0) && (inAnchorFileSpec->parID != 0) && (inAnchorFileSpec->name[0] > 0) ) 
					{
							
						mViewThread = new CMacCVSProjectViewThread( *this );
						ThrowIfNil_( mViewThread );
							
						mAnchorItem = new CMacCVSProjectViewRootFolder( inAnchorFileSpec, &mProjectRootFileSpec );
						ThrowIfNil_( mAnchorItem );
						
						mAnchorItem->SetComparator( GetCurrentComparator() );
												
						WatchFolder( mAnchorItem );
												
						fakeUpdateEvent.fEventItem = mAnchorItem;
						PostProjectViewEvent( fakeUpdateEvent );

						mViewThread->Resume();
													
					}
				
			}
			
	}


UInt32
CMacCVSProjectView::CountSelectedItems( void ) const
	{
		
		Int32 			theItemCount 	= 0x00000000;
		STableCell 	theCell( 0, 0 );
		
		while( GetNextSelectedCell(theCell) )
			theItemCount++;
			
		return theItemCount;
		
	}


LArray *
CMacCVSProjectView::GetFlatSelection()
	{
	
		CMacCVSProjectViewItem		*theItem							= NULL;
		
		UInt32 										ioCount								= 0;
		LArray										*outSelectedFSSpecs		= NULL;
		FSSpec										theFileSpec						= { 0, 0, "\p" };

		STableCell 								theCell( 0, 0 );
		STableCell								theWideCell( 0, 0 );
				
		outSelectedFSSpecs = new LArray( sizeof(FSSpec) );
		ThrowIfNil_( outSelectedFSSpecs );
		
		while ( GetNextSelectedCell(theCell) )
			{
			
				theItem = NULL;
				
				theWideCell = theCell;
				theWideCell.row = GetWideOpenIndex( theWideCell.row );
				
				ioCount = sizeof( CMacCVSProjectViewItem * );
				GetCellData( theWideCell, &theItem, ioCount );

				Assert_( ioCount == sizeof(CMacCVSProjectViewItem *) );
				
				if ( (theItem != NULL) && (ioCount == sizeof(CMacCVSProjectViewItem *)) )
					{
					
						theItem->GetFSSpec( theFileSpec );
						outSelectedFSSpecs->InsertItemsAt( 1, LArray::index_Last, &theFileSpec );
							
					}
					
			}
			
		return outSelectedFSSpecs;
	
	}


// ---------------------------------------------------------------------------
//		 GetCollapsedSelection
// ---------------------------------------------------------------------------
//	GetCollapsedSelection() returns an LArray list of FSSpecs that represent 
//  the currently selected items. If a container and some of it's children are 
//  selected, only the container's FSSpec will be added to the list.


LArray *
CMacCVSProjectView::GetCollapsedSelection()
	{
	
		CMacCVSProjectViewItem		*theItem						= NULL,
															*curSelectedParent	= NULL;
															
		UInt32 										theItemCount 				= 0,
															ioCount							= 0;

		FSSpec										theFileSpec					= { 0, 0, "\p" };
		Boolean										isAlreadySelected		= false;
		LArray										*outSelectedFSSpecs	= NULL;
		
		STableCell 								theCell( 0, 0 );
		STableCell								theWideCell( 0, 0 );
		
		LArray										curSelectedItems( sizeof(CMacCVSProjectViewItem *) );
		LArrayIterator						curSelectedItemsIterator( curSelectedItems );
			
		outSelectedFSSpecs = new LArray( sizeof(FSSpec) );
		ThrowIfNil_( outSelectedFSSpecs );
		
		while ( GetNextSelectedCell(theCell) )
			{
			
				theItem = NULL;

				theWideCell = theCell;
				theWideCell.row = GetWideOpenIndex( theWideCell.row );
				
				ioCount = sizeof( CMacCVSProjectViewItem * );
				GetCellData( theWideCell, &theItem, ioCount );

				Assert_( ioCount == sizeof(CMacCVSProjectViewItem *) );
				
				if ( (theItem != NULL) && (ioCount == sizeof(CMacCVSProjectViewItem *)) )
					{

						isAlreadySelected = false;
											
						curSelectedItemsIterator.ResetTo( LArrayIterator::from_Start );

//  Check to make sure this item's parents are not already selected.

						while ( (curSelectedItemsIterator.Next(&curSelectedParent)) && (!isAlreadySelected) )
							{
							
								if ( theItem->IsDescendentOf(curSelectedParent) )
									isAlreadySelected = true;
							
							}
							
						if ( !isAlreadySelected )
							{
							
								if ( theItem->IsContainer() )
									curSelectedItems.InsertItemsAt( 1, LArray::index_Last, &theItem );
							
								theItem->GetFSSpec( theFileSpec );
								outSelectedFSSpecs->InsertItemsAt( 1, LArray::index_Last, &theFileSpec );
								
							}
							
					}
					
			}
			
		return outSelectedFSSpecs;
	
	}
	
	
void
CMacCVSProjectView::GetNthItemFSSpec( const UInt32 intInIndex, FSSpec &fsOutItem )
	{

		CMacCVSProjectViewItem		*pCurItem			= NULL;
	
		TableIndexT								intMaxRows		= 0;
		TableIndexT								intMaxCols		= 0;

		UInt32										intCellSize		= 0;
		FSSpec										fsCurItem			= { 0 };		
		Boolean										bFoundIt			= false;
		
		STableCell sCurCell( 0, 1 );

		fsOutItem.name[0] = 0;
		fsOutItem.vRefNum = 0;
		fsOutItem.parID = 0;

		GetTableSize( intMaxRows, intMaxCols );
				
		if ( intInIndex > 0 && intInIndex <= intMaxRows )
			{
				sCurCell.row = GetWideOpenIndex( intInIndex );
				intCellSize = sizeof(pCurItem);
		
				GetCellData( sCurCell, &pCurItem, intCellSize );
				
				Assert_( intCellSize == sizeof(pCurItem) );
				Assert_( pCurItem != NULL );
				
				if ( pCurItem != NULL ) 
					{					
						pCurItem->GetFSSpec( fsCurItem );
						fsOutItem = fsCurItem;
					}			
			}
			
	}
	
	
void
CMacCVSProjectView::GetNamedItemFSSpec( ConstStringPtr pstrInName, FSSpec &fsOutItem )
	{
		
		CMacCVSProjectViewItem		*pCurItem			= NULL;
	
		TableIndexT								intRow				= 0;
		TableIndexT								intMaxRows		= 0;
		TableIndexT								intMaxCols		= 0;

		UInt32										intCellSize		= 0;
		FSSpec										fsCurItem			= { 0 };		
		Boolean										bFoundIt			= false;
		
		STableCell sCurCell( 0, 1 );

		fsOutItem.name[0] = 0;
		fsOutItem.vRefNum = 0;
		fsOutItem.parID = 0;

		GetTableSize( intMaxRows, intMaxCols );
				
		for ( intRow = 1; (intRow <= intMaxRows) && (!bFoundIt); intRow++ ) 
			{			
				sCurCell.row = GetWideOpenIndex( intRow );
				intCellSize = sizeof(pCurItem);
				
				GetCellData( sCurCell, &pCurItem, intCellSize );
				Assert_( intCellSize == sizeof(pCurItem) );
				
				if ( pCurItem != NULL ) 
					{					
						pCurItem->GetFSSpec( fsCurItem );
							
						bFoundIt = ::EqualString( fsCurItem.name, pstrInName, false, true );
								
						if ( bFoundIt )
							{
								fsOutItem = fsCurItem;
							}
					}					
			}
			
	}
	
	
UInt32
CMacCVSProjectView::GetVisibleItemCount()
	{

		UInt32	intCols = 0;
		UInt32	intRows = 0;
		
		GetTableSize( intRows, intCols );
		
		return intRows;
	
	}
	
	
#pragma mark -

	
void 
CMacCVSProjectView::ExpandRow( TableIndexT inWideOpenRow )
	{	
		
		CMacCVSProjectViewItem		*theExpandedFolderUncasted 	= NULL;
		CMacCVSProjectViewFolder	*theExpandedFolder 					= NULL;
		Rect											cellRect 										= { 0, 0, 0, 0 };
		STableCell								theExpandedItemCell( 0, 1 );
		
		LHierarchyTable::ExpandRow( inWideOpenRow );
		
		theExpandedItemCell.row = GetExposedIndex( inWideOpenRow );

//  This is annoying: I want to redraw the cell immediately. But if I do this
//  without erasing the current cell rect, the hiliting won't draw correctly.
		
		if ( (theExpandedItemCell.row > 0) && (GetLocalCellRect(theExpandedItemCell, cellRect)) ) 
			{

				FocusDraw();

				if ( CellIsSelected(theExpandedItemCell) )
					::EraseRect( &cellRect );
			
				DrawCell( theExpandedItemCell, cellRect );
		
				if ( CellIsSelected(theExpandedItemCell) )
					HiliteCellActively( theExpandedItemCell, true );
					
			}
			
		theExpandedFolderUncasted = ConvertTableIndexToItem( inWideOpenRow );

		if ( theExpandedFolderUncasted->IsContainer() )
			theExpandedFolder = dynamic_cast<CMacCVSProjectViewFolder *>( theExpandedFolderUncasted );

		Assert_( theExpandedFolder != NULL );
		
		if ( theExpandedFolder != NULL ) 
			{
				
				CMacCVSProjectViewEventRec	openEvent = { kViewExpandProjectFolderEvent, theExpandedFolder };
				PostProjectViewEvent( openEvent );

			}
		
	}


void 
CMacCVSProjectView::DeepExpandRow( TableIndexT inWideOpenRow )
	{
		
		CMacCVSProjectViewItem		*theExpandedFolderUncasted 	= NULL;
		CMacCVSProjectViewFolder	*theExpandedFolder 					= NULL;
		Rect											cellRect 										= { 0, 0, 0, 0 };
		STableCell								theExpandedItemCell( 0, 1 );
		
		LHierarchyTable::ExpandRow( inWideOpenRow );

		theExpandedItemCell.row = GetExposedIndex( inWideOpenRow );

//  See comments in CMacCVSProjectView::ExpandRow().
		
		if ( (theExpandedItemCell.row > 0) && (GetLocalCellRect(theExpandedItemCell, cellRect)) ) 
			{

				FocusDraw();

				if ( CellIsSelected(theExpandedItemCell) )
					::EraseRect( &cellRect );
			
				DrawCell( theExpandedItemCell, cellRect );
		
				if ( CellIsSelected(theExpandedItemCell) )
					HiliteCellActively( theExpandedItemCell, true );
					
			}
			
		theExpandedFolderUncasted = ConvertTableIndexToItem(inWideOpenRow);

		if ( theExpandedFolderUncasted->IsContainer() )
			theExpandedFolder = dynamic_cast<CMacCVSProjectViewFolder *>( theExpandedFolderUncasted );
			
		Assert_( theExpandedFolder != NULL );
		
		if ( theExpandedFolder != NULL ) 
			{
			
				CMacCVSProjectViewEventRec	openDeepEvent = { kViewExpandProjectFolderDeepEvent, theExpandedFolder };
				PostProjectViewEvent( openDeepEvent );

			}
					
	}


void
CMacCVSProjectView::CollapseRow( TableIndexT inWideOpenRow )
	{

		CMacCVSProjectViewItem		*theCollapseFolderUncasted 	= NULL;
		CMacCVSProjectViewFolder	*theCollapseFolder 					= NULL;
		Rect											cellRect 										= { 0, 0, 0, 0 };
		STableCell								theCollapseItemCell( 0, 1 );
			
		LHierarchyTable::CollapseRow( inWideOpenRow );
	
		theCollapseItemCell.row = GetExposedIndex( inWideOpenRow );

//  See comments in CMacCVSProjectView::ExpandRow().
				
		if ( (theCollapseItemCell.row > 0) && (GetLocalCellRect(theCollapseItemCell, cellRect)) ) 
			{

				FocusDraw();

				if ( CellIsSelected(theCollapseItemCell) )
					::EraseRect( &cellRect );
			
				DrawCell( theCollapseItemCell, cellRect );
		
				if ( CellIsSelected(theCollapseItemCell) )
					HiliteCellActively( theCollapseItemCell, true );
					
			}
			
		theCollapseFolderUncasted = ConvertTableIndexToItem(inWideOpenRow);

		if ( theCollapseFolderUncasted->IsContainer() )
			theCollapseFolder = dynamic_cast<CMacCVSProjectViewFolder *>(theCollapseFolderUncasted);

		Assert_( theCollapseFolder != NULL );
		
//  Some might ask if posting a close event is necessary because all we need to
//  do is call CMacCVSProjectViewFolder::SetOpenStatus(false) on the object. Since
//  CollapseRow is called by the main thread, we can't call SetOpenStatus() because
//  only the view thread should do that. Secondly, this properly deals with the
//  case where a deep expand task is currently running when the user does a deep
//  collapse.
		
		if ( theCollapseFolder != NULL ) 
			{
			
				CMacCVSProjectViewEventRec	closeEvent = { kViewCollapseProjectFolderEvent, theCollapseFolder };
				PostProjectViewEvent( closeEvent );

			}
		
	}


void
CMacCVSProjectView::DeepCollapseRow( TableIndexT inWideOpenRow ) 
	{

		CMacCVSProjectViewItem		*theCollapseFolderUncasted 	= NULL;
		CMacCVSProjectViewFolder	*theCollapseFolder 					= NULL;
		Rect											cellRect 										= { 0, 0, 0, 0 };
		STableCell								theCollapseItemCell( 0, 1 );
		
		LHierarchyTable::CollapseRow( inWideOpenRow );

		theCollapseItemCell.row = GetExposedIndex( inWideOpenRow );
		
//  See comments in CMacCVSProjectView::ExpandRow().
				
		if ( (theCollapseItemCell.row > 0) && (GetLocalCellRect(theCollapseItemCell, cellRect)) ) 
			{

				FocusDraw();

				if ( CellIsSelected(theCollapseItemCell) )
					::EraseRect( &cellRect );
			
				DrawCell( theCollapseItemCell, cellRect );
		
				if ( CellIsSelected(theCollapseItemCell) )
					HiliteCellActively( theCollapseItemCell, true );
					
			}
			
		theCollapseFolderUncasted = ConvertTableIndexToItem( inWideOpenRow );

		if ( theCollapseFolderUncasted->IsContainer() )
			theCollapseFolder = dynamic_cast<CMacCVSProjectViewFolder *>( theCollapseFolderUncasted );
		
		Assert_( theCollapseFolder != NULL );

//  This is a bit tricky. If a user deep expanded a folder, and the view thread is still processing
//  that request, then a standard deep close will not work properly. Ie. the open deep event is calling
//  ValidateContents() which adds new folders, which expands those after the DeepCollapseRows has finished.
//  To work around this, deep closes now post a close event which is picked up by the view thread after 
//  it has finished it's open operation.
		
		if ( theCollapseFolder != NULL ) 
			{
				
				CMacCVSProjectViewEventRec	closeDeepEvent = { kViewCollapseProjectFolderDeepEvent, theCollapseFolder };
				PostProjectViewEvent( closeDeepEvent );

			}
		
	}


#pragma mark -
#pragma mark ITEM MANIPULATION


CMacCVSProjectViewItem *
CMacCVSProjectView::ConvertTableIndexToItem( const TableIndexT inRow ) const
	{
		
		if ( inRow == (TableIndexT)0 )

			return mAnchorItem;
				
		else 
			{
				
				STableCell							tableCell( inRow, 1 );
				CMacCVSProjectViewItem	*theItem	= NULL;
				Uint32									cellSize 	= sizeof( CMacCVSProjectViewItem * );
			
				GetCellData( tableCell, &theItem, cellSize );
				Assert_( cellSize == sizeof(CMacCVSProjectViewItem *) );
				
				if ( cellSize == sizeof(CMacCVSProjectViewItem *) )
					
					return theItem;
				
				else
					
					return NULL;
					
			}
			
	}


TableIndexT 
CMacCVSProjectView::ConvertItemToTableIndex( const CMacCVSProjectViewItem *inItem, const TableIndexT inStartIndex ) const
	{

//  This is a *really* slow function so I minimize the amount of times this
//  function is called. Also, you can pass in a "hint" index to expedite searching.
//  ( see HandleOpenDeepEvent() )
	
		if ( inItem == mAnchorItem )
		
			return ( (TableIndexT)0 );
		
		else 
			{
			
				TableIndexT							outRows			= 0, 
																outCols			= 0,
																currentRow	= 0;

				CMacCVSProjectViewItem	*theCurItem	= NULL;
				UInt32									cellSize 		= 0;

				STableCell 							currentCell( 0, 1 );
			
				GetWideOpenTableSize( outRows, outCols );

				for ( currentRow = inStartIndex; currentRow <= outRows; currentRow++ ) 
					{
					
						currentCell.row 	= currentRow;
						theCurItem				= NULL;
						cellSize 					= sizeof( CMacCVSProjectViewItem * );
				
						GetCellData( currentCell, &theCurItem, cellSize );
					
						if ( theCurItem == inItem )
							return currentCell.row;
							
					}
			
			}
		
		return kProjectNoLongerExistsRow;
		
	}


TableIndexT
CMacCVSProjectView::GetFirstChildIndex( const TableIndexT inParentWideIndex )
	{
		
		TableIndexT		childIndex 		= 0x00000000;
		UInt32				maxRows				= 0x00000000,
									maxColumns		= 0x00000000;
						
		GetWideOpenTableSize( maxRows, maxColumns );
		
		if ( (inParentWideIndex == 0) && (maxRows > 0) )
			
			childIndex = 1;
			
		else if ( (inParentWideIndex + 1) <= maxRows ) 
			{
				
				childIndex = inParentWideIndex + 1;
				
				if ( GetNestingLevel(childIndex) != (GetNestingLevel(inParentWideIndex) + 1) )
					childIndex = 0;
			
			}
		
		return childIndex;

	}


TableIndexT
CMacCVSProjectView::GetNextSiblingIndex( const TableIndexT inWideIndex )
	{

		TableIndexT		nextSiblingIndex 	= 0x00000000,
									maxRows						= 0x00000000,
									maxColumns				= 0x00000000;
						
		UInt32				nestingLevel 			= 0x00000000;

		GetWideOpenTableSize( maxRows, maxColumns );

//  PowerPlant bug: CountAllDescendents will crash if you call it with
//  a row that is not in the table.
			
		if ( (inWideIndex > 0) && (inWideIndex <= maxRows) ) 
			{
			
				nestingLevel = GetNestingLevel( inWideIndex );
				
				nextSiblingIndex = inWideIndex + CountAllDescendents( inWideIndex ) + 1;
				
				GetWideOpenTableSize( maxRows, maxColumns );
				
				if ( (nextSiblingIndex > maxRows) || GetNestingLevel(nextSiblingIndex) != nestingLevel )
					nextSiblingIndex = 0;
					
			}
		
		return nextSiblingIndex;

	}


void
CMacCVSProjectView::RemoveTableItems( TableIndexT inParentIndex, LArray *inDeletedItems, Boolean inRefresh )
	{

		TableIndexT							currentIndex 	= GetFirstChildIndex( inParentIndex ),
														lastIndex			= 0,
														delItemIndex 	= 0x00000000;

		UInt32									cellDataSize	= 0x00000000;

		CMacCVSProjectViewItem 	*currentItem	= NULL;					
		STableCell							theCell( 0, 1 );
		Boolean									bDeletedCurrent = false;
		
		while ( (currentIndex > 0) && (inDeletedItems->GetCount() > 0) ) 
			{
			
				bDeletedCurrent = false;
				
				theCell.row = currentIndex;
				
				cellDataSize = sizeof( CMacCVSProjectViewItem * );
				currentItem = NULL;
				
				GetCellData( theCell, &currentItem, cellDataSize );
			
				Assert_( (currentItem != NULL) && (cellDataSize == sizeof(CMacCVSProjectViewItem *)) );
				
				if ( (currentItem != NULL) && (cellDataSize == sizeof(CMacCVSProjectViewItem *)) ) 
					{
					
						delItemIndex = inDeletedItems->FetchIndexOf( &currentItem );
						
						if ( delItemIndex != LArray::index_Bad ) 
							{
							
								if ( currentItem->IsContainer() ) 
									{

										// remove from watching stuff
										
										CMacCVSProjectViewFolder *theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>( currentItem );
										Assert_( theItemCasted != NULL );
										
										if ( theItemCasted != NULL )
											if ( theItemCasted->GetOpenStatus() )
												StopWatchingFolderDeep( theItemCasted );
										
									}
								
								RemoveRows( 1, currentIndex, inRefresh );
								inDeletedItems->RemoveItemsAt( 1, delItemIndex );
								
								delete currentItem;
								
								bDeletedCurrent = true;
							
							}
						
					}
				
				if ( bDeletedCurrent )
					{
					
						if ( lastIndex == 0 )
							
							currentIndex = GetFirstChildIndex( inParentIndex );
							
						else currentIndex = GetNextSiblingIndex( lastIndex );
						
					}
					
				else
					{
					
						lastIndex = currentIndex;
						currentIndex = GetNextSiblingIndex( currentIndex );
				
					}
				
			}
		
	}


void
CMacCVSProjectView::UpdateFolderContents( TableIndexT inWideIndex, Boolean inRefresh )
	{

		CMacCVSProjectViewItem		*theOldItem					= NULL,
															*theNewItem					= NULL,
															*theFolderUncasted  = NULL;
		
		CMacCVSProjectViewFolder	*theFolder					= NULL;
		
		TableIndexT								parentIndex 				= inWideIndex,
															lastSiblingIndex 		= 0x00000000,
															currentIndex 				= 0x00000000,
															exposedIndex				= 0x00000000;
		
		UInt32										itemCount						= 0x00000000;
		
		STableCell								theCell( 0, 1 );
		UInt32										cellDataSize				= 0x00000000;
		
		LArray										*removedItemsList		= NULL,
															*contentsList				= NULL;
		
		theFolderUncasted = ConvertTableIndexToItem( inWideIndex );
		
		Assert_( theFolderUncasted != NULL );
		
		if ( theFolderUncasted->IsContainer() )
			theFolder = dynamic_cast<CMacCVSProjectViewFolder *>( theFolderUncasted );
		
		Assert_( theFolder != NULL );
				
		if ( theFolder != NULL ) 
			{

//  First validate the folder's contents to see
//  if something has changed.
		
				try
					{
					
						if ( theFolder->IsOutOfDate() ) 
							{
							
								removedItemsList = theFolder->ValidateContentsChanged( &mProjectRootFileSpec );

//  Redraw the folder cell after we've revalidated it.

								exposedIndex = GetExposedIndex( inWideIndex );
								
								if ( exposedIndex > 0 )
									{
									
										theCell.row = exposedIndex;
										RefreshCell( theCell );
									
									}
																	
//  Delete project items that no longer exist.
																
								if ( removedItemsList != NULL ) 
									{

										RemoveTableItems( parentIndex, removedItemsList, inRefresh );
										delete removedItemsList;
										removedItemsList = NULL;
									
									}

//  Rebuild the item list in the table, inserting
//  new items where they belong.
				
								contentsList = theFolder->GetContainedItemsPtr();
								
								if ( contentsList != NULL ) 
									{
								
										itemCount = contentsList->GetCount();

										lastSiblingIndex = 0;
										currentIndex = GetFirstChildIndex( parentIndex );
										
										for ( UInt32 loopIndex = 1; loopIndex <= itemCount; loopIndex++ ) 
											{
												
												theNewItem = NULL;
												
												contentsList->FetchItemAt( loopIndex, &theNewItem );
												Assert_( theNewItem != NULL );
												
												if ( currentIndex > 0 ) 
													{

//  If there is a valid sibling row, compare the new item to the item in that row.
//  If they are not equal, the new item must come before the old one because the
//  sorting method is the same. So, insert that baby.
							
														theCell.row = currentIndex;
														cellDataSize = sizeof(CMacCVSProjectViewItem *);
									
														theOldItem = NULL;
													
														GetCellData( theCell, &theOldItem, cellDataSize );

														Assert_( (theOldItem != NULL) && (cellDataSize == sizeof(CMacCVSProjectViewItem *)) );
														
														if ( !CMacCVSProjectViewItem::IsEqualName(theNewItem, theOldItem) ) 
															{
									
																if ( (lastSiblingIndex == 0) && (parentIndex > 0) ) 
																							
																	currentIndex = InsertChildRows( 1, parentIndex, &theNewItem, sizeof(CMacCVSProjectViewItem *), theNewItem->IsContainer(), inRefresh );
																
																else 
																	
																	currentIndex = InsertSiblingRows( 1, lastSiblingIndex, &theNewItem, sizeof(CMacCVSProjectViewItem *), theNewItem->IsContainer(), inRefresh );

																if ( theNewItem->IsContainer() ) 
																	{
																	
																		LHierarchyTable::CollapseRow( currentIndex );
																																		
																		CMacCVSProjectViewFolder *theNewItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>( theNewItem );
																		Assert_( theNewItemCasted != NULL );
																		
																		if ( theNewItemCasted != NULL )
																			theNewItemCasted->SetComparator( GetCurrentComparator() );
																		
																	}
																									
															}

														else
															{

//  Else if the item is already in the table and hasn't 
//  changed position, refresh it.
															
																theCell.row = GetExposedIndex( theCell.row );
																RefreshCell( theCell );
																
															}

														lastSiblingIndex = currentIndex;
														currentIndex = GetNextSiblingIndex( currentIndex );
									
													} 
												
												else 
													{
							
//  There are no more siblings so add last child if parent is a row,
//  or add sibling if parent is the root.
				
														if ( parentIndex > 0 )
														
															currentIndex = AddLastChildRow( parentIndex,  &theNewItem, sizeof(CMacCVSProjectViewItem *), theNewItem->IsContainer(), inRefresh );
													
														else
														
															currentIndex = InsertSiblingRows( 1, lastSiblingIndex, &theNewItem, sizeof(CMacCVSProjectViewItem *), theNewItem->IsContainer(), inRefresh );
													
														if ( theNewItem->IsContainer() ) 
															{
														
																LHierarchyTable::CollapseRow( currentIndex );														
															
																CMacCVSProjectViewFolder *theNewItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>(theNewItem);
																Assert_( theNewItemCasted != NULL );
																
																if ( theNewItemCasted != NULL )
																	theNewItemCasted->SetComparator( GetCurrentComparator() );
																	
															}
															
														lastSiblingIndex = currentIndex;
														currentIndex = GetNextSiblingIndex( currentIndex );
									 
													}
							
											}
						
									}
									
							}
				
					}
				
				catch ( OSErr inErr )
					{

//  If the view root disappeared, delete all the rows. This could happen
//  if the user deleted the anchor item.
												
						if ( theFolder == mAnchorItem )
							{

								RemoveAllRows( true );
								mWatchList->RemoveItemsAt( mWatchList->GetCount(), LArray::index_First );

								delete theFolder;
								mAnchorItem = NULL;

							}

//  Else invalidate it's parent so that we can get rid of it.
						
						else 
							{
							
								CMacCVSProjectViewItem 	*parentItem	= theFolder->GetParent();
								Assert_( parentItem != NULL );
								
								if ( parentItem != NULL )
									parentItem->Invalidate();
									
							}

						if ( removedItemsList != NULL )
							{
							
								delete removedItemsList;
								removedItemsList = NULL;
								
							}
							
						if ( (inErr != fnfErr) && (inErr != dirNFErr) )
							throw inErr;
							
					}
					
				catch ( ... )
					{
					
						if ( removedItemsList != NULL )
							{
							
								delete removedItemsList;
								removedItemsList = NULL;
								
							}
							
						throw;
											
					}
										
			}
			
	}


void
CMacCVSProjectView::RefreshEntryData( CMacCVSProjectViewItem *inItem )
	{

//  This is a bit of a hack. Since entry information is no longer
//  stored with the file, we have to open the Entries file everytime
//  we need to validate an item. This sucks.

//  It remains to be seen how bad the performance is on this now that
//  we need to open the whole entries file everytime we want to update
//  one file.

//  You might as why I just don't leave mEntries open to some directory
//  and call CCVSEntries::AreYouThisDirectory on it. Well, I don't want
//  to leave a file open for every lurkers window and plus it would
//  make the view thread compete with the lurkers thread for the entries
//  object.

		OSErr		intResult				= noErr;
		Str255 	pstrRevision 		= "\p";
		Str255 	pstrTimeStamp 	= "\p";
		Str255 	pstrOptions 		= "\p";
		Str255 	pstrTag 				= "\p";

		FSSpec 	fsItem					= { 0, 0, "\p" };		
		FSSpec 	fsParent				= { 0, 0, "\p" };
		
		Assert_( inItem != NULL );
		ThrowIfNil_( inItem );

		CCVSEntries *theEntries = NULL;
		CMacCVSProjectViewFile *theItemCasted = NULL;
		
		try
			{
				theItemCasted = dynamic_cast<CMacCVSProjectViewFile *>(inItem);
				ThrowIfNil_( theItemCasted );

				inItem->GetFSSpec( fsItem );
				
				intResult = FSpGetParentFSSpec( &fsItem, &fsParent );
				ThrowOSErr( intResult );
				
				theEntries = new CCVSEntries;
				ThrowIfNil_( theEntries );
				
				intResult = theEntries->LoadWorkingDirectory( &fsParent, fsRdPerm );
				ThrowOSErr( intResult );

				intResult = theEntries->GetEntry( fsItem.name, pstrRevision, pstrTimeStamp, pstrOptions, pstrTag );
				
				Assert_( (intResult == noErr) || (intResult == kCVSEngineNoRevisionDataErr) );
		
				if ( intResult == noErr )
					{
						theItemCasted->SetEntry( pstrRevision, pstrTimeStamp, pstrOptions, pstrTag );
					}
				else
					{
						theItemCasted->SetEntry( NULL, NULL, NULL, NULL );
					}
			}
		catch ( ... )
			{
				if ( theItemCasted != NULL )
					{
						theItemCasted->SetEntry( NULL, NULL, NULL, NULL );
					}
			}
			
		try
			{
				if ( theEntries != NULL )
					{
						delete theEntries;
						theEntries = NULL;
					}
			}
		catch ( ... )
			{
			
			}
			
	}
	

void
CMacCVSProjectView::ResizeViewColumns( SInt16 inWindowWidth )
	{

		mIconOffset 							= kCellIconOffset;
		mTextOffset 							= kCellTextOffset;
		mVersionOffset 						= kCellVersionTextOffset;
		mTagOffset 								= kCellTagTextOffset;
		mDateOffset 							= kCellModificationTextOffset;
	
	}
	

#pragma mark -
#pragma mark DRAWING

	
void
CMacCVSProjectView::DrawSelf()
	{
	
		Rect 			theFrame 	= { 0, 0, 0, 0 };
		
		CalcLocalFrameRect( theFrame );
		
		::EraseRect( &theFrame );
	
		LHierarchyTable::DrawSelf();
		
	}
	

void
CMacCVSProjectView::DrawCell(const STableCell &inCell,const Rect &inLocalRect)
	{
		
		CMacCVSProjectViewItem 		*cellFile			= NULL;
		UInt32										relativeDate	= 0x00000000;
		TableIndexT								theWideIndex 	= GetWideOpenIndex( inCell.row );
		Rect											rTemp					= inLocalRect;
		
		cellFile = ConvertTableIndexToItem( theWideIndex );
		
		if ( cellFile != NULL ) 
			{
				
				Str255							cellName				= "\p";
				Rect 								statusIconRect	= { 0, 0, 0, 0 };
				ProjectItemStatus		itemStatus			= kMacCVSProjectorStatusUnknown;

				rTemp.top = rTemp.bottom - 1;
				
				::ForeColor( whiteColor );
				::PaintRect( &rTemp );

				::ForeColor( blackColor );
							
				if ( mCollapsableTree->IsCollapsable(theWideIndex) ) 
					{
					
						Rect theFlagRect = { 0, 0, 0, 0 };
						
						CalcCellFlagRect( inCell, theFlagRect );
						
						LDropFlag::Draw( theFlagRect, mCollapsableTree->IsExpanded(theWideIndex) );
					
					}

				UInt32 theNestingLevel = mCollapsableTree->GetNestingLevel(theWideIndex);
		
				DrawCellIcon( inCell, inLocalRect, CellIsSelected(inCell) );

//  Save and setup the text traits.				
				StTextState	theTextState;
				
				if ( cellFile->IsAlias() )
					UTextTraits::SetPortTextTraits(kCellTextTraitsItalicID);
				else
					UTextTraits::SetPortTextTraits(kCellTextTraitsID);

				::MoveTo( inLocalRect.left + theNestingLevel * kCellChildOffset + mTextOffset, 
									inLocalRect.bottom - kCellTextOffsetFromBottom );
						
				cellFile->GetNamePStr( cellName );
				::DrawString( cellName );

				if ( cellFile->IsAlias() )
					UTextTraits::SetPortTextTraits(kCellTextTraitsID);
				
//	Draw the status icon.
				
				itemStatus = cellFile->GetProjectItemStatus();

				if ( (itemStatus >= kMacCVSProjectorStatusUnowned) && (itemStatus < kMacCVSProjectorTotalCount) ) 
					{
					
					 	statusIconRect.top = inLocalRect.top + 1;
					 	statusIconRect.left = inLocalRect.left + mStatusIconOffset;
					 	statusIconRect.right = statusIconRect.left + kFileStatusIconWidth;
					 	statusIconRect.bottom = statusIconRect.top + kFileStatusIconHeight;

						::PlotIconSuite(&statusIconRect, kAlignNone, kTransformNone, sFileStatusIconSuites[itemStatus]);
					
					}
				
				if ( cellFile->IsLocked() )
					{
					 	statusIconRect.top = inLocalRect.top + 1;
					 	statusIconRect.left = inLocalRect.left + mLockIconOffset;
					 	statusIconRect.right = statusIconRect.left + kFileLockIconWidth;
					 	statusIconRect.bottom = statusIconRect.top + kFileLockIconHeight;
		
						::PlotIconID( &statusIconRect, kAlignNone, kTransformNone, kCellLockIconID );
					}
					
//	Draw the version string for the file.

				if ( itemStatus > kMacCVSProjectorStatusUnowned ) 
					{
					
						CMacCVSProjectViewFile	*projectFile 	= (CMacCVSProjectViewFile *)cellFile;
						Str255									versionString	= "\pN/A";
						
						::MoveTo(	inLocalRect.left + mVersionOffset, 
											inLocalRect.bottom - kCellTextOffsetFromBottom );
							
						projectFile->GetVersionPStr( versionString );
						
						if ( itemStatus >= kMacCVSProjectorStatusBroken ) 
							{						
								RGBColor grey = { 0x8000, 0x8000, 0x8000 };
								RGBForeColor( &grey );								
							}
						else if ( CUIsAddedOrRemovedRevision(versionString) )
							{
								RGBColor red = { 0xFFFF, 0x0000, 0x0000 };
								RGBForeColor( &red );
							}
						
						::DrawString( versionString );
						::ForeColor( blackColor );

						::MoveTo(	inLocalRect.left + mTagOffset, 
									inLocalRect.bottom - kCellTextOffsetFromBottom );

						if ( !cellFile->IsContainer() )
							((CMacCVSProjectViewFile *)cellFile)->GetStickyTagPStr(versionString);
						
						Assert_( versionString[0] != 1 );
						
						if ( versionString[0] > 1 )
							{
							
								versionString[0]--;
								::BlockMoveData( &versionString[2], &versionString[1], versionString[0] );
								
								::DrawString( versionString );
						
							}	
						
					}
					
				else if ( cellFile->IsCVSContainer() )
					{

						Str255 tagString	= "\p";
					
						::MoveTo(	inLocalRect.left + mTagOffset, 
										inLocalRect.bottom - kCellTextOffsetFromBottom );

						cellFile->GetStickyTagPStr( tagString );
	
						if ( tagString[0] > 1 )
							{
							
								tagString[0]--;
								::BlockMoveData( &tagString[2], &tagString[1], tagString[0] );
								
								::DrawString( tagString );
						
							}
					
					}
				
				::MoveTo(	inLocalRect.left + mDateOffset, 
							inLocalRect.bottom - kCellTextOffsetFromBottom );
				
				TString<Str255> theRelDate;
				
				UMacCVSUtils::RelativeDateString( cellFile->GetCachedModificationDate(), theRelDate );
				::DrawString( theRelDate );
				
			}

		LHierarchyTable::DrawCell( inCell, inLocalRect );
					
	}


void
CMacCVSProjectView::DrawCellIcon( const STableCell &inCell, const Rect &inLocalRect, Boolean inHilite )
	{

		TableIndexT								theWideIndex 	= GetWideOpenIndex( inCell.row );
		CMacCVSProjectViewItem		*theItem 			= ConvertTableIndexToItem( theWideIndex );
			
		if ( theItem != NULL ) 
			{
			
				UInt32 	theNestingLevel = mCollapsableTree->GetNestingLevel( theWideIndex );
				
				Rect 		theIconFrame 		= { 0, 0, 16, 16 };
				
				::OffsetRect(	&theIconFrame, inLocalRect.left + mIconOffset + theNestingLevel * kCellChildOffset, inLocalRect.top + 1 );
				
				IconTransformType theTransform = kTransformNone;
				
				if ( inHilite )
					theTransform += kTransformSelected;
				
				if ( mCollapsableTree->IsCollapsable(theWideIndex) && mCollapsableTree->IsExpanded(theWideIndex) )
					theTransform += kTransformOpen;
				
				Handle	theIconSuite 	= NULL;
				FSSpec	theItemSpec;
				
				theItem->GetFSSpec(theItemSpec);
					
#ifdef FASTER_FIND_ICON
				if ( Find_icon( &theItemSpec, theItem->GetFileInfoRecPtr(), kSelectorSmallThenLarge, &theIconSuite) == noErr)
#else
				if ( Find_icon( &theItemSpec, NULL, kSelectorSmallThenLarge, &theIconSuite) == noErr)
#endif
					::PlotIconSuite(  &theIconFrame, kAlignNone, theTransform, theIconSuite );
				else
					{
						ResIDT 	theIconID 			= theItem->IsContainer() ? kCellFolderIconID : kCellFileIconID;
						::PlotIconID( &theIconFrame, kAlignNone, theTransform, theIconID );
						
					}
					
				if ( theIconSuite != NULL) {
					::DisposeIconSuite (theIconSuite, true);
					theIconSuite = NULL;
				}
				
			}	
					
	}


void
CMacCVSProjectView::HiliteCellActively( const STableCell &inCell, Boolean inHilite )
	{
		
		Rect 	cellFrame = { 0, 0, 0, 0 };
				
		if ( GetLocalCellRect(inCell, cellFrame) && FocusExposed() ) 
			{
								
				DrawCellIcon( inCell, cellFrame, inHilite );

				Rect theTextFrame = { 0, 0, 0, 0 };
				
				CalcCellTextFrame( inCell, theTextFrame );
				::InsetRect( &theTextFrame, 0, 1 );
									
				::RGBBackColor( &mBGColor );
				
				StRegion theRegion( theTextFrame );
								
				URegionUtils::Hilite( theRegion );
				
			}
			
	}


void
CMacCVSProjectView::HiliteCellInactively( const STableCell &inCell, Boolean inHilite )
	{

		HiliteCellActively( inCell, inHilite );
					
	}


void 
CMacCVSProjectView::CalcCellTextFrame( const STableCell &inCell, Rect &outLocalFrame )
	{
	
// Start with the whole cell
	
		GetLocalCellRect( inCell, outLocalFrame );
		outLocalFrame.top += 2;
		outLocalFrame.bottom -= 2;
		
// Get the wide open index for the row.
	
		TableIndexT	theWideIndex = GetWideOpenIndex( inCell.row );

// Get the nesting level.
	
		UInt32	theNestingLevel = mCollapsableTree->GetNestingLevel( theWideIndex );

		CMacCVSProjectViewItem* theCellFile = ConvertTableIndexToItem( theWideIndex );

		TString<Str255> theName;
		theCellFile->GetNamePStr( theName );
		
		StTextState	theTextState;
		UTextTraits::SetPortTextTraits( kCellTextTraitsID );
		
		outLocalFrame.left = outLocalFrame.left + theNestingLevel * kCellChildOffset + mTextOffset - 1;
		outLocalFrame.right = outLocalFrame.left + ::StringWidth(theName) + 1;

	}


void
CMacCVSProjectView::CalcCellHotSpots( const STableCell &inCell, Rect &outIconFrame, Rect &outTextFrame )
	{

//  This function is almost exactly like CalcCellTextFrame() except that it returns
//  the icon frame also. I didn't want to have to call GetWideOpenIndex() twice, so
//  I just duplicated the code from CalcCellTextFrame() here.

		Rect		localRect			= { 0, 0, 0, 0 },
						iconFrame			= { 0, 0, 16, 16 };
				
// Start with the whole cell

		GetLocalCellRectAnywhere( inCell, localRect );
		
		outTextFrame = localRect;
		
		outTextFrame.top += 2;
		
// Get the wide open index for the row.

		TableIndexT	theWideIndex = GetWideOpenIndex( inCell.row );

// Get the nesting level.

		UInt32	theNestingLevel = mCollapsableTree->GetNestingLevel( theWideIndex );

		CMacCVSProjectViewItem* theCellFile = ConvertTableIndexToItem( theWideIndex );

		TString<Str255> theName;
		theCellFile->GetNamePStr( theName );
		
		StTextState	theTextState;
		UTextTraits::SetPortTextTraits( kCellTextTraitsID );
		
		outTextFrame.left = outTextFrame.left + theNestingLevel * kCellChildOffset + mTextOffset - 1;
		outTextFrame.right = outTextFrame.left + ::StringWidth(theName) + 1;

		::OffsetRect(	&iconFrame, localRect.left + mIconOffset + theNestingLevel * kCellChildOffset, localRect.top + 1 );

		outIconFrame = iconFrame;
						
	}


void
CMacCVSProjectView::HideSelf()
	{
	
		if ( IsOnDuty() ) 
			{				// Shouldn't be on duty when invisible
			
				SwitchTarget( GetSuperCommander() );
		
			}
	
	}
	

void
CMacCVSProjectView::AdaptToNewSurroundings()
	{
	
		Rect		rFrameRect 	= { 0, 0, 0, 0 };
		UInt32	intRowWidth = 0;
		
		LHierarchyTable::AdaptToNewSurroundings();
		
		CalcLocalFrameRect( rFrameRect );
		
		if ( (rFrameRect.right - rFrameRect.left) > kDefaultRowWidth )
			intRowWidth = rFrameRect.right - rFrameRect.left;
		else
			intRowWidth = kDefaultRowWidth;
		
		Assert_( mTableGeometry != NULL );
		
		if ( mTableGeometry != NULL )	
			mTableGeometry->SetColWidth( intRowWidth, 0, 0 );
		
	}
	
	
#pragma mark -
#pragma mark SELECTION


void
CMacCVSProjectView::ClickSelf( const SMouseDownEvent	&inMouseDown )
	{
	
		if ( !IsTarget() ) 
			{	
			
				SwitchTarget( this );
			
			}
		
		if ( IsTarget() ) 
			{
				
				LHierarchyTable::ClickSelf( inMouseDown );

//  This code was copied from LHierarchyTable::ClickSelf(). If the user
//  didn't click in any cell, do the drag selector.

				SPoint32		imagePt = { 0, 0 };
				STableCell	hitCell( 0, 0 );
				
				LocalToImagePoint( inMouseDown.whereLocal, imagePt );
				
				if ( !GetCellHitBy(imagePt, hitCell) )
					{
					
						mTableSelector->DragSelect( hitCell, inMouseDown );
					
					}
				
			}
	
	}
	
	
void
CMacCVSProjectView::ClickCell( const STableCell &inCell, const SMouseDownEvent &inMouseDown )
	{
		
		Boolean isOptionKeyDown 	= (inMouseDown.macEvent.modifiers & optionKey) != 0,
						isCommandKeyDown	= (inMouseDown.macEvent.modifiers & cmdKeyBit) != 0;

		LHierarchyTable::ClickCell( inCell, inMouseDown );
						
//  If the option key is down, display the cool path popup.
		
		if ( isOptionKeyDown  ) 
			{

				CMacCVSProjectViewItem		*theClickItem 		= NULL;
				TableIndexT								theWideIndex 			= 0x00000000;
				Rect 											theCellFrame 			= { 0, 0, 0, 0 };
				FSSpec 										theSpec						= { 0, 0, "\p" };
				Point 										theGlobalPoint		= { 0, 0 };

				theWideIndex = GetWideOpenIndex( inCell.row );
				theClickItem = ConvertTableIndexToItem( theWideIndex );
				Assert_( theClickItem != NULL );
				
				if ( theClickItem != NULL )
					{
					
						CalcCellTextFrame( inCell, theCellFrame );
				
						theClickItem->GetFSSpec( theSpec );

// Yep, they're magic constants.  These are used to align the text in the first item
// in the popup with the item's name text.  We could do a better calculation based on
// font info.
				
						theGlobalPoint.h = theCellFrame.left - 11;
						theGlobalPoint.v = theCellFrame.top + 2;
				
						LocalToPortPoint(theGlobalPoint);
						PortToGlobalPoint(theGlobalPoint);

						CPathPopup thePathPopup( theSpec, kCellTextTraitsID, true, true, true );
						thePathPopup.Track( theGlobalPoint );
						
					}
				
			}
			
		else if ( GetClickCount() > 1 )
			{
					
				CMacCVSProjectViewItem		*theCurItem 		= NULL;
				TableIndexT								theWideIndex 		= 0x00000000;
	
				STableCell								theCurCell( 0, 0 );
				
				while ( GetNextSelectedCell(theCurCell) )
					{
					
						theWideIndex = GetWideOpenIndex( theCurCell.row );
						theCurItem = ConvertTableIndexToItem( theWideIndex );
						Assert_( theCurItem != NULL );
						
						if ( theCurItem != NULL )
							{
							
								if ( !theCurItem->IsContainer() ) 
									theCurItem->OpenWithFinder();
											
								else if ( mCreateNewView == true )
									{
										MakeNewViewForCell( theCurCell );
									}

								else
									{

//  else just turn down the item

										TableIndexT		theWideIndex 	= GetWideOpenIndex( inCell.row );
										Boolean 			isExpanded 		= mCollapsableTree->IsExpanded( theWideIndex );
													
										if ( isExpanded )
										
											CollapseRow( theWideIndex );
										
										else ExpandRow( theWideIndex );

										FocusDraw();
										DrawDropFlag( inCell, theWideIndex );
										
									}
									
							}
							
					}
					
			}
						
	}
	

Boolean
CMacCVSProjectView::ClickSelect( const STableCell &inCell, const SMouseDownEvent &inMouseDown )
	{

		Boolean 	result					= false;
		Boolean 	isOptionKeyDown	= ( (inMouseDown.macEvent.modifiers & optionKey) != 0 );
		
		Rect 	theTextFrame = { 0, 0, 0, 0 },
					theIconFrame = { 0, 0, 0, 0 };
		
		CalcCellHotSpots( inCell, theIconFrame, theTextFrame );
		
		theIconFrame.top--;
		theIconFrame.bottom++;
		theIconFrame.left -= 2;
		theIconFrame.right += 2;
					
		if ( ::PtInRect(inMouseDown.whereLocal, &theIconFrame) || ::PtInRect(inMouseDown.whereLocal, &theTextFrame) ) 
			{
	
				if ( !isOptionKeyDown )
					{
						
						mUserDraggedSelection = false;
						
						LHierarchyTable::ClickSelect( inCell, inMouseDown );
						
						result = !mUserDraggedSelection;
						
					}
				else result = true;
				
			}
		
		else if ( (inMouseDown.macEvent.modifiers & cmdKey) || (inMouseDown.macEvent.modifiers & shiftKey) ) 
		{
			
				mTableSelector->DragSelect( inCell, inMouseDown );
		
		}
		else if ( !(inMouseDown.macEvent.modifiers & shiftKey) ) 
			{

//  Is this the right place for this? - rvr
					
				UnselectAllCells();
				mTableSelector->DragSelect( inCell, inMouseDown );
		
			}
			
		return result;
		
	}


void
CMacCVSProjectView::GetCellIconRgn( const STableCell &inCell, Boolean inHilite, RgnHandle ioRgn )
	{

		Rect										localCellRect = { 0, 0, 0, 0 };
		TableIndexT							theWideIndex 	= GetWideOpenIndex( inCell.row );
		CMacCVSProjectViewItem	*theItem 			= ConvertTableIndexToItem( theWideIndex );
				
		if ( theItem != NULL )
			{

				GetLocalCellRect( inCell, localCellRect );
				
				UInt32	theNestingLevel = mCollapsableTree->GetNestingLevel(theWideIndex);
			
				ResIDT 	theIconID 			= theItem->IsContainer() ? kCellFolderIconID : kCellFileIconID;
				Rect 		theIconFrame 		= { 0, 0, 16, 16 };
			
				::OffsetRect( &theIconFrame, localCellRect.left + mIconOffset + theNestingLevel * kCellChildOffset, localCellRect.top + 1 );
			
				IconTransformType theTransform = ttNone;
				
				if ( inHilite )
					theTransform += ttSelected;
					
				if ( mCollapsableTree->IsCollapsable(theWideIndex) && mCollapsableTree->IsExpanded(theWideIndex) )
					theTransform += ttOpen;
				
				::IconIDToRgn( ioRgn, &theIconFrame, atNone, theIconID );
			}	
			
	}


void
CMacCVSProjectView::GetHiliteRgn( RgnHandle ioHiliteRgn )
	{

		RgnHandle		cellRgn				= NULL,
								tempRgn				= NULL;
								
		Rect				visRect 			= { 0, 0, 0, 0 },
								cellTextRect 	= { 0, 0, 0, 0 };
		
		try
			{
			
				::EmptyRgn( ioHiliteRgn );
				
				GetRevealedRect( visRect );
				
				if ( !::EmptyRect(&visRect) ) 
					{
					
						PortToLocalPoint( topLeft(visRect) );
						PortToLocalPoint( botRight(visRect) );
						
						STableCell	topLeftCell, botRightCell;
						FetchIntersectingCells( visRect, topLeftCell, botRightCell );
						
						cellRgn = ::NewRgn();
						ThrowIfNil_( cellRgn );
						
						tempRgn = ::NewRgn();
						ThrowIfNil_( tempRgn );
						
						STableCell	cell;
						
						for ( cell.row = topLeftCell.row; cell.row <= botRightCell.row; cell.row++ ) 
							{
							
								for ( cell.col = topLeftCell.col; cell.col <= botRightCell.col; cell.col++ ) 
									{
									
										if ( CellIsSelected(cell) ) 
											{
												::EmptyRgn( cellRgn );
																			
												GetCellIconRgn( cell, CellIsSelected(cell), cellRgn );
												CalcCellTextFrame( cell, cellTextRect );
															
												::RectRgn( tempRgn, &cellTextRect );
												
												::UnionRgn( tempRgn, cellRgn, tempRgn );
												URegionUtils::MakeRegionOutline( tempRgn, &tempRgn );
												::UnionRgn( tempRgn, ioHiliteRgn, ioHiliteRgn);
											}
											
									}
									
							}
						
					}

				if ( cellRgn != NULL )
					{
						::DisposeRgn( cellRgn );
						cellRgn = NULL;
					}

				if ( tempRgn != NULL )
					{
						::DisposeRgn( tempRgn );
						tempRgn = NULL;
					}
					
			}
			
		catch ( ... )
			{
			
				if ( cellRgn != NULL )
					{
					
						::DisposeRgn( cellRgn );
						cellRgn = NULL;
						
					}

				if ( tempRgn != NULL )
					{
					
						::DisposeRgn( tempRgn );
						tempRgn = NULL;
						
					}
			
				throw;
				
			}
			
	}

// Get the cell rect whether it is visible or not
void
CMacCVSProjectView::GetLocalCellRectAnywhere( const STableCell	&inCell, Rect &outCellRect ) const
	{
		Int32		cellLeft, cellTop, cellRight, cellBottom;
		
		mTableGeometry->GetImageCellBounds(inCell, cellLeft, cellTop, cellRight, cellBottom);
		
		SPoint32	imagePt;
		imagePt.h = cellLeft;
		imagePt.v = cellTop;
		ImageToLocalPoint(imagePt, topLeft(outCellRect));
		outCellRect.right = outCellRect.left + (cellRight - cellLeft);
		outCellRect.bottom = outCellRect.top + (cellBottom - cellTop);

	}

// so that we update commands when keying through lists
void
CMacCVSProjectView::SelectionChanged()
	{
		LCommander::SetUpdateCommandStatus(true);
	}
		
		
void
CMacCVSProjectView::SelectItem( const FSSpec *inFileSpec)
	{
		CMacCVSProjectViewEventRec		selectEventRec = { CMacCVSProjectView::kViewSelectProjectItemFSSpecEvent, 0, 0 };
			
			FSSpec *theSpecPtr = (FSSpec *)::NewPtr( sizeof(FSSpec) );
			ThrowIfNil_( theSpecPtr );
			
			::BlockMoveData( inFileSpec, theSpecPtr, sizeof(FSSpec) );
			selectEventRec.fEventData = theSpecPtr;
			
			PostProjectViewEvent( selectEventRec );
	}

void
CMacCVSProjectView::MakeNewViewForCell( const STableCell &theCell )
	{
		UInt32									theWideIndex = GetWideOpenIndex( theCell.row );
		CMacCVSProjectViewItem	*theCurItem = ConvertTableIndexToItem( theWideIndex );
		
		
		Assert_( theCurItem != NULL );
		
		if ( theCurItem != NULL )
			{
			
				if ( mCreateNewView == true )
					{

						if ( theCurItem->IsContainer() ) 
							{

								Rect		textFrameWeDontCareAbout 	= { 0, 0, 0, 0 };
								CMacCVSProjectViewNewViewEventRec		broadcastData;
														
								theCurItem->GetFSSpec( broadcastData.fFileSpec );
								CalcCellHotSpots( theCell, broadcastData.fFromCellRect, textFrameWeDontCareAbout );

								LocalToPortPoint( topLeft(broadcastData.fFromCellRect) );
								PortToGlobalPoint( topLeft(broadcastData.fFromCellRect) );
								
								LocalToPortPoint( botRight(broadcastData.fFromCellRect) );
								PortToGlobalPoint( botRight(broadcastData.fFromCellRect) );
								
								BroadcastMessage( kNewViewMessage, &broadcastData );
							}	
					
					}
					
			}
	}
	
	
#pragma mark -
#pragma mark COMMAND HANDLING


Boolean
CMacCVSProjectView::CheckIfAnyItemsQualify( CommandT inCommand )
	{
	
		Boolean										result				= false;
		CMacCVSProjectViewItem		*pCurItem			= NULL;
		UInt32										intSize			 	= sizeof( CMacCVSProjectViewItem * );
		
		STableCell cCurCell;
		STableCell cWideOpenCell;
		
		while ( GetNextSelectedCell(cCurCell) && (!result) )
			{
				cWideOpenCell = cCurCell;
				cWideOpenCell.row = GetWideOpenIndex( cCurCell.row );
			
				if ( cWideOpenCell.row > 0 )
					{					
						pCurItem = NULL;
						
						GetCellData( cWideOpenCell, &pCurItem, intSize );
						Assert_( intSize == sizeof(CMacCVSProjectViewItem *) );
						ThrowIfNil_( pCurItem );
						
						switch ( inCommand )
							{
								// JWW - Added Reveal In Finder case
								case cmd_RevealInFinder:							
									result = true;									
								break;
								
								case cmd_Update:
								case cmd_UpdateTo:								
									result = pCurItem->CanUpdate();				
								break;
							
								case cmd_Checkout:	
								case cmd_CheckoutTo:									
									result = pCurItem->CanCheckout();									
								break;
								
								case cmd_AddFile:								
									result = pCurItem->CanBeAdded();									
								break;
								
								case cmd_Commit:							
									result = pCurItem->CanCheckin();									
								break;								

								case cmd_Editors:
									//  Because "Show Editors" works on the same
									//  criteria of files/folder as Watchers, we 
									//  have this OR statement.
									
									result = (pCurItem->CanWatch() || (pCurItem->CanEdit()));
								break;
								
								case cmd_Edit:
								case cmd_UnEdit:
									result = pCurItem->CanEdit();
								break;
								
								case cmd_ShowEditors:
								case cmd_Watchers:
								case cmd_WatchOn:
								case cmd_WatchOff:
								case cmd_WatchAdd:
								case cmd_WatchRemove:
								case cmd_ShowWatchers:
									result = pCurItem->CanWatch();
								break;
								
								case cmd_RevertFile:								
									result = pCurItem->CanRevert();									
								break;
																
								case cmd_ModifyReadOnly:								
									result = pCurItem->CanMRO();									
								break;
								
								case cmd_EditComment:								
									result = pCurItem->CanEditComment();									
								break;
								
								case cmd_CompareWithOriginal:								
									result = pCurItem->CanCompareWithOriginal();									
								break;
	
								case cmd_Diff:
								case cmd_AdvancedDiff:
									result = pCurItem->CanDiff();									
								break;
															
								case cmd_GetFileLog:								
									result = pCurItem->CanGetLog();									
								break;
								
								case cmd_GetFileStatus:								
									result = pCurItem->CanGetStatus();								
								break;

								case cmd_Tag:								
									result = pCurItem->CanTag();									
								break;
								
								case cmd_OrphanFiles:								
									result = pCurItem->CanOrphanFiles();
								break;
								
								case cmd_RemoveFiles:								
									result = pCurItem->CanRemove();
								break;								
							}
					}			
			}
			
		return result;
			
	}
	
	
void
CMacCVSProjectView::BeTarget()
	{
	
		mWatchTime = kDefaultForegroundUpdateTime;
	
	}
	
	
void
CMacCVSProjectView::DontBeTarget()
	{
	
		mWatchTime = kDefaultBackgroundUpdateTime;
		
	}
	
	
Boolean
CMacCVSProjectView::ObeyCommand(CommandT inCommand, void *ioParam)
	{
		EventRecord		*theEvent = (EventRecord *)ioParam;
		Boolean 			cmdHandled = true;
		
		switch (inCommand) 
			{
			
				case cmd_Copy:
				{
						Handle				dataHandle = NULL;
						STableCell 		theCurCell( 0, 0 );
						Boolean				clearData = true;
						
						dataHandle = ::NewHandle(0);
						if ( dataHandle == NULL)
							ThrowIfError_(MemError());
						
						try
							{
								while ( GetNextSelectedCell(theCurCell) )
									{
										CMacCVSProjectViewItem 		*cellFile			= NULL;
										TableIndexT								theWideIndex 	= GetWideOpenIndex( theCurCell.row );
										
										cellFile = ConvertTableIndexToItem( theWideIndex );
										
										if ( cellFile != NULL ) 
											{
													char		itemName[257];
			
													cellFile->GetNameCStr(itemName);
													
													strcat(itemName, "\r");
													ThrowIfError_( ::PtrAndHand( itemName, dataHandle, strlen( itemName) ) );
											}
									}
								
								Int32			length = ::GetHandleSize( dataHandle );
								
								::HLock( dataHandle );
								
								LClipboard::GetClipboard()->SetData('TEXT', *dataHandle, length - 1 );		//ignore trailing '\r'
						
								::DisposeHandle(dataHandle);		//doesn't matter that it's still locked
								dataHandle = NULL;

								
							}
						catch ( OSErr inErr )
							{
								if (dataHandle != NULL)
								{
									::DisposeHandle(dataHandle);
									dataHandle = NULL;
								}
								
								throw( inErr );
							}
								
					}
				break;
				
				case cmd_CopyPath:
					{
						Handle				dataHandle = NULL;
						STableCell 		theCurCell( 0, 0 );
						Boolean				clearData = true;
						
						dataHandle = ::NewHandle(0);
						if ( dataHandle == NULL)
							ThrowIfError_(MemError());
						
						try
							{
								while ( GetNextSelectedCell(theCurCell) )
									{
										CMacCVSProjectViewItem 		*cellFile			= NULL;
										TableIndexT								theWideIndex 	= GetWideOpenIndex( theCurCell.row );
										
										cellFile = ConvertTableIndexToItem( theWideIndex );
										
										if ( cellFile != NULL ) 
											{
													FSSpec			itemSpec;
													char				filePath[512];
													UInt32			pathLen = 510;
													
													cellFile->GetFSSpec(itemSpec);
													
													ThrowIfError_ ( FSpGetLocalCVSPath( &itemSpec, &mProjectRootFileSpec, filePath, &pathLen ) );
													filePath[pathLen] = '\0';
													strcat(filePath, "\r");
													ThrowIfError_( ::PtrAndHand( filePath, dataHandle, strlen( filePath) ) );
											}
									}
									
									
								Int32			length = ::GetHandleSize( dataHandle );
								
								::HLock( dataHandle );
								
								LClipboard::GetClipboard()->SetData('TEXT', *dataHandle, length - 1 );		//ignore trailing '\r'
						
								::DisposeHandle(dataHandle);		//doesn't matter that it's still locked
								dataHandle = NULL;
								
							}
						catch ( OSErr inErr )
							{
								if (dataHandle != NULL)
								{
									::DisposeHandle(dataHandle);
									dataHandle = NULL;
								}
								
								throw( inErr );
							}

					}
				break;
				
				case cmd_SelectAll:
				
					SelectAllCells();
					
				break;
				
				default:
				
					cmdHandled = LCommander::ObeyCommand( inCommand, ioParam );
				
				break;
			
			}
			
		return cmdHandled;
		
	}


void
CMacCVSProjectView::FindCommandStatus( CommandT inCommand, Boolean &outEnabled, Boolean &outUsesMark, Char16 &outMark, Str255 outName )
	{

//  Ok, if the world was perfect, the CMacCVSDoc would be responsible for doing this. However,
//  since CMacCVSProjectView knows first hand of item hierarchies and status, it is much, much
//  faster to do it here. If I were to do this in CMacCVSDoc::FindCommandStatus, I would have
//  to call GetCollapsedSelection() or whatever, and then read the ckid/mcvs information for
//  each FSSpec, and...etc...etc...

		outEnabled = false;	

		switch ( inCommand ) 
			{
			
				case cmd_Copy:
				case cmd_CopyPath:
				
					outEnabled = ( CountSelectedItems() > 0 ); // are any cells selected?
				
				break;
				
				case cmd_Cut:
				
					outEnabled = ( CountSelectedItems() > 0 );
					
				break;
					
				case cmd_SelectAll:
				
					outEnabled = true;
					
				break;
					
				case cmd_RevealInFinder:	// JWW - Added 04/21/99
				
				case cmd_Update:
				case cmd_UpdateTo:
				case cmd_Checkout:
				case cmd_CheckoutTo:
				
				case cmd_AddFile:
				case cmd_Commit:
				
				case cmd_Editors:
				case cmd_Edit:
				case cmd_UnEdit:
				case cmd_ShowEditors:
	
				case cmd_Watchers:
				case cmd_WatchOn:
				case cmd_WatchOff:
				case cmd_WatchAdd:
				case cmd_WatchRemove:
				case cmd_ShowWatchers:
			
				case cmd_RevertFile:
				case cmd_RemoveFiles:
				case cmd_Tag:				
	
				case cmd_ModifyReadOnly:
				case cmd_EditComment:
	
				case cmd_CompareWithOriginal:
				case cmd_Diff:
				case cmd_AdvancedDiff:
				case cmd_GetFileLog:
				case cmd_GetFileStatus:

				
					outEnabled = CheckIfAnyItemsQualify( inCommand );
					
				break;
										
				default:
				
					LCommander::FindCommandStatus( inCommand, outEnabled, outUsesMark, outMark, outName );

				break;
			
			}
			
	}


void
CMacCVSProjectView::ListenToMessage( MessageT inMessage, void *ioParam )
	{

		CMacCVSProjectViewEventRec	theNewEvent = { 0, 0 };
		
		switch ( inMessage ) 
			{
		
				case kNameSortMessage:
						
					if ( *(UInt32 *)ioParam != 0 )  
						{
						
							theNewEvent.fEventType = kViewSortByNameEvent;
							
							PostProjectViewEvent( theNewEvent );
						
						}
					
				break;
					
				case kStatusSortMessage:

					if ( *(UInt32 *)ioParam != 0 )  
						{
						
							theNewEvent.fEventType = kViewSortByStatusEvent;
							
							PostProjectViewEvent( theNewEvent );
						
						}
				
				break;
				
				case kVersionSortMessage:
				
					if ( *(UInt32 *)ioParam != 0 )  
						{
						
							theNewEvent.fEventType = kViewSortByVersionEvent;
							
							PostProjectViewEvent( theNewEvent );
						
						}
				
				break;
				
				case kDateSortMessage:
				
					if ( *(UInt32 *)ioParam != 0 )  
						{
						
							theNewEvent.fEventType = kViewSortByDateEvent;
							
							PostProjectViewEvent( theNewEvent );
						
						}

				break;
				
				case kTagSortMessage:
				
					if ( *(UInt32 *)ioParam != 0 )  
						{
						
							theNewEvent.fEventType = kViewSortByTagEvent;
							
							PostProjectViewEvent( theNewEvent );
						
						}
				
				break;
				
			}

	}
	

#pragma mark -
#pragma mark PROJECT VIEW THREADING


Boolean 
CMacCVSProjectView::GetNextProjectViewEvent(CMacCVSProjectViewEventRec &inEventRecord)
	{
		
		Boolean	result = false;
		
		if ( mViewEventQueue->GetCount() > 0 )
			{
			
				mViewEventQueue->FetchItemAt( 1, &inEventRecord );
		
				mViewEventQueue->RemoveItemsAt( 1, 1 );

				result = true;
							
			}
		
		return result;
			
	}


void 
CMacCVSProjectView::PostProjectViewEvent(const CMacCVSProjectViewEventRec &inEventRecord)
	{

		mViewEventQueue->InsertItemsAt( 1, 0x7FFFFFFF, &inEventRecord );
			
	}


void
CMacCVSProjectView::WatchFolder( CMacCVSProjectViewFolder *inFolder )
	{

		if ( inFolder->GetOpenStatus() == 0 )
			{
			
				mWatchList->InsertItemsAt( 1, LArray::index_Last, &inFolder );
	
				inFolder->SetOpenStatus( true );
				
			}
			
	}


void
CMacCVSProjectView::StopWatchingFolder( CMacCVSProjectViewFolder	*inFolder )
	{
	
		UInt32	index = 0x00000000;

		Assert_( inFolder->GetOpenStatus() );
		
		if ( inFolder->GetOpenStatus() )
			{
			
				index = mWatchList->FetchIndexOf( &inFolder );
	
				Assert_( index != LArray::index_Bad );
			
				if ( index != LArray::index_Bad )
					mWatchList->RemoveItemsAt( 1, index );
					
				inFolder->SetOpenStatus( false );

			}	
	
	}
	

// ---------------------------------------------------------------------------
//		 CMacCVSProjectView::StopWatchingFolderDeep
// ---------------------------------------------------------------------------
//  StopWatchingFolderDeep() is called for a folder we are about to delete from 
//  the table. It is important that you do actually remove the folder from the 
//  table because the open status is messed up by calling StopWatchingFolder() 
//  on folders without actually collapsing them in the table.
	
	
void
CMacCVSProjectView::StopWatchingFolderDeep( CMacCVSProjectViewFolder	*inFolder )
	{

		UInt32										maxItemIndex			= 0x00000000,
															curItemIndex			= 0x00000000;
						
		LArray										*childItems				= NULL;
		CMacCVSProjectViewFolder	*curItemPtr 			= NULL;
		
		if ( inFolder->GetOpenStatus() )
			{
			
				StopWatchingFolder( inFolder );
				
				childItems = inFolder->GetContainedItemsPtr();
				
				if ( childItems != NULL )
					{
					
						maxItemIndex = childItems->GetCount();
						
						for ( curItemIndex = 1; curItemIndex <= maxItemIndex; curItemIndex++ )
							{
								
								if ( childItems->FetchItemAt(curItemIndex, &curItemPtr) )
									{
									
										ThrowIfNil_( curItemPtr );
										
										if ( curItemPtr->IsContainer() )
											if ( curItemPtr->GetOpenStatus() )
												StopWatchingFolderDeep( curItemPtr );
												
									}
							
							}
					
					}
					
			}		

	}


void
CMacCVSProjectView::RefreshViewContents()
	{
	
		LArray										*itemList						= NULL;
		
		Boolean										bDone								= false;
		
		UInt32										maxIndex 						= 0x00000000,
															curIndex 						= 0x00000000,
															maxItemIndex				= 0x00000000,
															curItemIndex				= 0x00000000;

		CMacCVSProjectViewFolder	*theCurFolder				= NULL;
		CMacCVSProjectViewItem		*theCurItem					= NULL,
															*theParent					= NULL;
			
		CMacCVSProjectViewEventRec				updateEvent					= { kViewUpdateProjectItemEvent, NULL, (void *)1 };
											
		maxIndex = mWatchList->GetCount();
									
		for ( curIndex = 1; curIndex <= maxIndex; curIndex++ )
			{			
				mWatchList->FetchItemAt( curIndex, &theCurFolder );
				ThrowIfNil_( theCurFolder );
				
				if ( theCurFolder->IsOutOfDate() )
					{				
						updateEvent.fEventItem = theCurFolder;
						PostProjectViewEvent( updateEvent );						
					}					
				else
					{					
						itemList = theCurFolder->GetContainedItemsPtr();
						ThrowIfNil_( itemList );
																		
						maxItemIndex = itemList->GetCount();
						
						for ( curItemIndex = 1; (curItemIndex <= maxItemIndex) && (!bDone); curItemIndex++ )
							{
								if ( (sYieldCount % kDefaultYieldCount) == 0 )
									CThreadedTask::Yield();
							
								sYieldCount++;

								itemList->FetchItemAt( curItemIndex, &theCurItem );
								ThrowIfNil_( theCurItem );

								if ( theCurItem->IsContainer() == true )
									{									
										if ( theCurItem->ExistsLocal() == false )
											{
												theParent = NULL;
												theParent = theCurItem->GetParent();
														
												if ( theParent != NULL )
													{												
														theCurFolder = dynamic_cast<CMacCVSProjectViewFolder *>( theParent );
														ThrowIfNil_( theCurFolder );
														
														theCurFolder->Invalidate();
														
														updateEvent.fEventItem = theCurFolder;
														PostProjectViewEvent( updateEvent );
														
														bDone = true;						
													}													
												else
													{															
														theCurFolder = dynamic_cast<CMacCVSProjectViewFolder *>( theCurItem );
														ThrowIfNil_( theCurFolder );
														
														theCurFolder->Invalidate();

														updateEvent.fEventItem = theCurFolder;
														PostProjectViewEvent( updateEvent );
														
														bDone = true;													
													}											
											}
									}
								else 
									{									
										if ( theCurItem->IsOutOfDate() )
											{											
												theCurItem = theCurItem->GetParent();				
												theCurFolder = dynamic_cast<CMacCVSProjectViewFolder *>(theCurItem);
												Assert_( theCurFolder != NULL );
												ThrowIfNil_( theCurFolder );
												
												theCurFolder->Invalidate();
	
												updateEvent.fEventItem = theCurFolder;
												PostProjectViewEvent( updateEvent );

												bDone = true;												
											}							
									}							
							} //  for											
					}						
			}	
	}
	
	
void
CMacCVSProjectView::ThreadedTaskDied( const CThreadedTask *inTask )
	{
	
		Assert_( (inTask == mViewThread) && (mViewThread != NULL) );
		
		if ( inTask != NULL )
			if ( inTask == mViewThread )
				mViewThread = NULL;
	
	}
	
	
#pragma mark -
#pragma mark EVENT HANDLING


void
CMacCVSProjectView::ProcessProjectViewEvent( CMacCVSProjectViewEventRec &inEventRecord )
	{
				
		switch ( inEventRecord.fEventType ) 
			{
			
				case kViewExpandProjectFolderEvent:
				
					HandleExpandEvent( inEventRecord.fEventItem, true );
							
				break;
				
				case kViewExpandProjectFolderDeepEvent:
				
					HandleExpandDeepEvent( inEventRecord.fEventItem );
				
				break;

				case kViewCollapseProjectFolderEvent:
				
					HandleCollapseEvent( inEventRecord.fEventItem );
					
				break;
				
				case kViewCollapseProjectFolderDeepEvent:
				
					HandleCollapseDeepEvent( inEventRecord.fEventItem );
					
				break;
				
				case kViewUpdateProjectItemEvent:
				
					if ( inEventRecord.fEventData == 0 )
					
						HandleUpdateEvent( inEventRecord.fEventItem, true );
					
					else HandleUpdateEvent( inEventRecord.fEventItem, false );
					
				break;
				
				case kViewUpdateProjectItemFSSpecEvent:

					Assert_( inEventRecord.fEventData != NULL );
				
					if ( inEventRecord.fEventData != NULL )
						{
						
							FSSpec theUpdateSpec = *(FSSpec *)inEventRecord.fEventData;
							
							::DisposePtr( (Ptr)inEventRecord.fEventData );
							inEventRecord.fEventData = NULL;
							
							HandleUpdateFSSpecEvent( theUpdateSpec );
					
						}
						
				break;
				
				case kViewSelectProjectItemFSSpecEvent:
					Assert_( inEventRecord.fEventData != NULL );
				
					if ( inEventRecord.fEventData != NULL )
						{
						
							FSSpec theUpdateSpec = *(FSSpec *)inEventRecord.fEventData;
							
							::DisposePtr( (Ptr)inEventRecord.fEventData );
							inEventRecord.fEventData = NULL;
							
							HandleSelectFSSpecEvent( theUpdateSpec );
						}
				break;
				
				case kViewSortByNameEvent:
				
					SortProjectView( kNameSorted );
					
				break;
				
				case kViewSortByStatusEvent:

					SortProjectView( kStatusSorted );	
				
				break;
				
				case kViewSortByVersionEvent:
				
					SortProjectView( kVersionSorted );	
				
				break;
				
				case kViewSortByDateEvent:
				
					SortProjectView( kDateSorted );
					
				break;
				
				case kViewSortByTagEvent:

					SortProjectView( kTagSorted );	
				
				break;

			}

	}


void 
CMacCVSProjectView::HandleUpdateEvent( CMacCVSProjectViewItem *inUpdateItem, const Boolean inShowIndicator )
	{
		
		TableIndexT		itemIndex			= 0x00000000,
									exposedIndex	= 0x00000000;
		
		itemIndex = ConvertItemToTableIndex( inUpdateItem );
		
		if ( itemIndex != kProjectNoLongerExistsRow ) 
			{

//  If it's a folder, call UpdateFolderContents().
		
				if ( inUpdateItem->IsContainer() )
					{
					
						if ( inShowIndicator )
							{
							
								BroadcastMessage( CSpinIndicator::msg_SpinWaitBegin );
														
							}
				
						UpdateFolderContents( itemIndex, true );
						
						if ( inShowIndicator )
							{
							
								BroadcastMessage( CSpinIndicator::msg_SpinWaitEnd );

							}
							
					}

//  Else do it here.
					
				else
					{
					
						try
							{
							
								inUpdateItem->Invalidate();
								inUpdateItem->ValidateContents( &mProjectRootFileSpec );
							
							}
							
						catch ( ... )
							{

//  Uhoh, we caught an exception- invalid the parent container
//  so the view thread will revalidate it and nuke this item.

								CMacCVSProjectViewItem 	*parentItem	= inUpdateItem->GetParent();
							
								Assert_( parentItem != NULL );
								
								if ( parentItem != NULL )
									parentItem->Invalidate();
							
							}
					
						exposedIndex = GetExposedIndex( itemIndex );
						
						if ( exposedIndex != 0 )
							{
							
								STableCell	cell( exposedIndex, 1 );
								RefreshCell( cell );
							
							}
							
					}

			}
		
	}


void
CMacCVSProjectView::HandleUpdateFSSpecEvent( const FSSpec &inUpdateItem )
	{
	
		CMacCVSProjectViewItem		*theCurItem		= NULL;
	
		TableIndexT								currentRow		= 0,
															maxRows				= 0,
															maxColumns		= 0;

		UInt32										cellSize 			= 0;
		FSSpec										curFileSpec		= { 0, 0, "\p" };
		
		Boolean										bFoundIt			= false;
		STableCell 								currentCell( 0, 1 );

		GetWideOpenTableSize( maxRows, maxColumns );
		
		for ( currentRow = 1; (currentRow <= maxRows) && (!bFoundIt); currentRow++ ) 
			{
			
				currentCell.row = currentRow;
				cellSize = sizeof(CMacCVSProjectViewItem *);
				
				GetCellData( currentCell, &theCurItem, cellSize );
				Assert_( cellSize == sizeof(CMacCVSProjectViewItem *) );
				
				if ( (theCurItem != NULL) && (cellSize == sizeof(CMacCVSProjectViewItem *)) )
					{
					
						theCurItem->GetFSSpec( curFileSpec );
						
						if ( 	(curFileSpec.vRefNum == inUpdateItem.vRefNum) && 
									(curFileSpec.parID == inUpdateItem.parID) && 
									(curFileSpec.name[0] == inUpdateItem.name[0]) )
							{
							
								bFoundIt = ::EqualString( curFileSpec.name, inUpdateItem.name, false, true );
								
								if ( bFoundIt )
									{
									
//  If it's a folder, call UpdateFolderContents().
		
										if ( theCurItem->IsContainer() )
											UpdateFolderContents( currentRow, true );

//  Else do it here.
					
										else
											{
											
												try
													{													
														theCurItem->Invalidate();
														RefreshEntryData( theCurItem );
														theCurItem->ValidateContents( &mProjectRootFileSpec );
													}
													
												catch ( ... )
													{

//  Uhoh, we caught an exception- invalidate the parent container
//  so the view thread will revalidate it and nuke this item.

														CMacCVSProjectViewItem 	*parentItem	= theCurItem->GetParent();
													
														Assert_( parentItem != NULL );
														
														if ( parentItem != NULL )
															parentItem->Invalidate();
													
													}
					
												TableIndexT exposedIndex = GetExposedIndex( currentRow );
						
												if ( exposedIndex != 0 )
													{
													
														STableCell	cell( exposedIndex, 1 );
														RefreshCell( cell );
													
													}
							
											}
							
									}
									
							}
					
					}
					
			}
	
	}


void
CMacCVSProjectView::HandleSelectFSSpecEvent( const FSSpec &inSelectItem )
	{
	
		CMacCVSProjectViewItem		*theCurItem		= NULL;
	
		TableIndexT								currentRow		= 0,
															maxRows				= 0,
															maxColumns		= 0;

		UInt32										cellSize 			= 0;
		FSSpec										curFileSpec		= { 0, 0, "\p" };
		
		Boolean										bFoundIt			= false;
		STableCell 								currentCell( 0, 1 );

		GetWideOpenTableSize( maxRows, maxColumns );
		
		for ( currentRow = 1; (currentRow <= maxRows) && (!bFoundIt); currentRow++ ) 
			{
			
				currentCell.row = currentRow;
				cellSize = sizeof(CMacCVSProjectViewItem *);
				
				GetCellData( currentCell, &theCurItem, cellSize );
				Assert_( cellSize == sizeof(CMacCVSProjectViewItem *) );
				
				if ( (theCurItem != NULL) && (cellSize == sizeof(CMacCVSProjectViewItem *)) )
					{
					
						theCurItem->GetFSSpec( curFileSpec );
						
						if ( 	(curFileSpec.vRefNum == inSelectItem.vRefNum) && 
									(curFileSpec.parID == inSelectItem.parID) && 
									(curFileSpec.name[0] == inSelectItem.name[0]) )
							{
							
								bFoundIt = ::EqualString( curFileSpec.name, inSelectItem.name, false, true );
								
								if ( bFoundIt )
									{
										STableCell cExposedIndex( GetExposedIndex(currentCell.row), 1 );
										
										if ( cExposedIndex.row != 0 )
											{
												UnselectAllCells();
												SelectCell( cExposedIndex);
												ScrollCellIntoFrame( cExposedIndex );
											}
									}
									
							}
					
					}
					
			}
	
	}
	
		
TableIndexT
CMacCVSProjectView::HandleExpandEvent( CMacCVSProjectViewItem *inExpandItem, const Boolean inShowIndicator, const TableIndexT inParentIndex  )
	{
		
		TableIndexT								itemIndex				= 0x00000000;
		CMacCVSProjectViewFolder	*theItemCasted	= NULL;
				
		itemIndex = ConvertItemToTableIndex( inExpandItem, inParentIndex );
		
		if ( itemIndex != kProjectNoLongerExistsRow ) 
			{

				theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>( inExpandItem );
				ThrowIfNil_( theItemCasted );
				
				if ( inShowIndicator )
					{
					
						BroadcastMessage( CSpinIndicator::msg_SpinWaitBegin );
												
					}
					
//  add to watch list
				
				WatchFolder( theItemCasted );		
					
				UpdateFolderContents( itemIndex, true );

				if ( inShowIndicator )
					{
					
						BroadcastMessage( CSpinIndicator::msg_SpinWaitEnd );

					}
										
			}
		
		return itemIndex;
		
	}


void
CMacCVSProjectView::HandleExpandDeepFolder( CMacCVSProjectViewFolder *inExpandItem, TableIndexT inParentIndex )
	{
		
		LArray										*contents				= NULL;
		CMacCVSProjectViewItem		*theItem 				= NULL;
		CMacCVSProjectViewFolder 	*theItemCasted 	= NULL;

		UInt32										maxIndex				= 0x00000000,
															index						= 0x00000000;
								
		TableIndexT								myTableIndex		= 0x00000000;
		STableCell								exposedCell( 0, 1 );
		
//  Throw an exception if our stack gets too tight...
								
		if ( ::StackSpace() < kViewThreadMinimumStackSize )
			throw( (OSErr)memFullErr );				

		contents = inExpandItem->GetContainedItemsPtr();
		
		if ( contents != NULL ) 
			{
			
				maxIndex = contents->GetCount();
				
				for ( index = 1; index <= maxIndex; index++ ) 
					{
					
						if ( contents->FetchItemAt(index, &theItem) ) 
							if ( theItem->IsContainer() ) 
								{
								
									theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>(theItem);
									
									if ( theItemCasted != NULL ) 
										{
										
											if ( !theItemCasted->GetOpenStatus() ) 
												{
												
													myTableIndex = HandleExpandEvent( theItemCasted, false, inParentIndex + 1 );

													LHierarchyTable::ExpandRow(	myTableIndex );

													exposedCell.row = GetExposedIndex( myTableIndex );

//  Here we call RefreshCell() instead of redrawing the cell
//  immediately. This is to prevent weird update artifacts 
//  when deep expanding. (ie, the new folders being drawn
//  before the items)												

													RefreshCell( exposedCell );
																																						
												}
											else
												{
				
													myTableIndex = ConvertItemToTableIndex( theItemCasted, inParentIndex + 1 );
										
												}
												
											HandleExpandDeepFolder( theItemCasted, myTableIndex );
	
										}
									
								}
						
					}
				
			}
			
	}


void
CMacCVSProjectView::HandleExpandDeepEvent( CMacCVSProjectViewItem	*inExpandItem )
	{
		
		TableIndexT								itemIndex				= 0x00000000;
		CMacCVSProjectViewFolder	*theItemCasted	= NULL;
				
		itemIndex = ConvertItemToTableIndex( inExpandItem );
		
		if ( itemIndex != kProjectNoLongerExistsRow ) 
			{

				theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>( inExpandItem );
				ThrowIfNil_( theItemCasted );

				BroadcastMessage( CSpinIndicator::msg_SpinWaitBegin );

//  First expand the single item- since we already know the index, pass it
//  as a hint to HandleExpandEvent() so it doesn't have to spent 45 million
//  years looking for the item.

//  add to watch list
				
				WatchFolder( theItemCasted );		
					
				UpdateFolderContents( itemIndex, true );

//  Then expand the children, passing myIndex as a hint 
//  for the ConvertItemToTableIndex() in future HandleOpenEvent() calls.

				HandleExpandDeepFolder( theItemCasted, itemIndex );

				BroadcastMessage( CSpinIndicator::msg_SpinWaitEnd );	
				
			}	
	
	}


TableIndexT
CMacCVSProjectView::HandleCollapseEvent( CMacCVSProjectViewItem *inCollapseItem, const TableIndexT inParentIndex )
	{

		TableIndexT								myTableIndex		= 0x00000000;
		CMacCVSProjectViewFolder	*theItemCasted	= NULL;

		ThrowIfNil_( inCollapseItem );
				
		myTableIndex = ConvertItemToTableIndex( inCollapseItem, inParentIndex );
		
		if ( myTableIndex != kProjectNoLongerExistsRow )	
			{
			
				theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>( inCollapseItem );
				ThrowIfNil_( theItemCasted );
				
				StopWatchingFolder( theItemCasted );

			}
			
		return myTableIndex;
			
	}


void
CMacCVSProjectView::HandleCollapseDeepFolder( CMacCVSProjectViewFolder *inCollapseItem, TableIndexT inParentIndex )
	{
		
		CMacCVSProjectViewItem		*theItem				= NULL;
		CMacCVSProjectViewFolder 	*theItemCasted 	= NULL;
		LArray										*contents 			= NULL;
		
		TableIndexT								myTableIndex		= 0x00000000;
		UInt32										maxIndex				= 0x00000000,
															index						= 0x00000000;

//  Throw an exception if our stack gets too tight...
								
		if ( ::StackSpace() < kViewThreadMinimumStackSize )
			throw( (OSErr)memFullErr );
			
		ThrowIfNil_( inCollapseItem );
		
		contents = inCollapseItem->GetContainedItemsPtr();
		
		if ( contents != NULL ) 
			{
			
				maxIndex = contents->GetCount();
				
				for ( index = 1; index <= maxIndex; index++ ) 
					{
					
						if ( contents->FetchItemAt(index, &theItem) ) 
							if ( theItem->IsContainer() ) 
								{
								
									theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>(theItem);
									
									ThrowIfNil_( theItemCasted );
									
									if ( theItemCasted->GetOpenStatus() ) 
										{
							
											myTableIndex = HandleCollapseEvent( theItemCasted, inParentIndex + 1 );
											
											LHierarchyTable::CollapseRow( myTableIndex );
											
											HandleCollapseDeepFolder( theItemCasted, myTableIndex );
																						
										}
																			
								}
						
					}
				
			}


	}


void
CMacCVSProjectView::HandleCollapseDeepEvent( CMacCVSProjectViewItem	*inCollapseItem )
	{

//  See comment in CMacCVSProjectView::DeepCollapseRow().

		TableIndexT									myTableIndex 		= 0x00000000;
		CMacCVSProjectViewFolder		*theItemCasted	= NULL;
		
		myTableIndex = ConvertItemToTableIndex( inCollapseItem );
		
		if ( myTableIndex != kProjectNoLongerExistsRow )
			{
				
				theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>( inCollapseItem );
				ThrowIfNil_( theItemCasted );

				StopWatchingFolder( theItemCasted );
					
				HandleCollapseDeepFolder( theItemCasted, myTableIndex );

			}
			
	}


#pragma mark -
#pragma mark SORTING LIST VIEW


void 
CMacCVSProjectView::RebuildSortedProjectView( CMacCVSProjectViewFolder *inParentFolder, TableIndexT inParentIndex, Boolean inRefresh )
	{
		
		LArray 										*contents						= NULL;
		CMacCVSProjectViewItem		*theItem						= NULL;
		CMacCVSProjectViewFolder	*theItemCasted			= NULL;
		UInt32										maxIndex						= 0x00000000,
															curIndex						= 0x00000000,
															currentTableIndex		= 0x00000000;

//  Throw an exception if our stack gets too tight...
								
		if ( ::StackSpace() < kViewThreadMinimumStackSize )
			throw( (OSErr)memFullErr );
			
		ThrowIfNil_( inParentFolder );
			
		inParentFolder->SetComparator( GetCurrentComparator() );
			
		contents = inParentFolder->GetContainedItemsPtr();
			
		ThrowIfNil_( contents );
					
		maxIndex = contents->GetCount();
				
		for ( curIndex = 1; curIndex <= maxIndex; curIndex++ ) 
			{
					
				if ( (curIndex % 32) == 0 )
					CThreadedTask::Yield();
					
				contents->FetchItemAt( curIndex, &theItem );

				ThrowIfNil_( theItem );
						
				if ( inParentIndex == 0 ) 
				
					currentTableIndex = InsertSiblingRows( 1, currentTableIndex, &theItem, sizeof(CMacCVSProjectViewItem *), theItem->IsContainer(), inRefresh );
					
				else
				
					currentTableIndex = AddLastChildRow( inParentIndex, &theItem, sizeof(CMacCVSProjectViewItem *), theItem->IsContainer(), inRefresh );
					
				if ( theItem->IsContainer() ) 
					{
						
						theItemCasted = dynamic_cast<CMacCVSProjectViewFolder *>(theItem);
									
						ThrowIfNil_( theItemCasted );
						
						if ( theItemCasted->GetOpenStatus() )
						
							LHierarchyTable::ExpandRow( currentTableIndex );
						
						else
						
							LHierarchyTable::CollapseRow( currentTableIndex );
							
						RebuildSortedProjectView( theItemCasted, currentTableIndex, inRefresh );
					
					}
				
			}
		
	}


void 
CMacCVSProjectView::SortProjectView( ProjectViewSortMethod inSortMethod )
	{


		TableIndexT								rowCount				= 0x00000000,
															columnCount			= 0x00000000;
		CMacCVSProjectViewItem		*currentItem		= NULL;
		CMacCVSProjectViewFolder	*currentFolder	= NULL;
		
		if ( (inSortMethod != mCurrentSortMethod) && (mAnchorItem != NULL) )
			{
			
			 	BroadcastMessage( CSpinIndicator::msg_SpinWaitBegin );
				
				mCurrentSortMethod = inSortMethod;

				//	Remove all of the items in the table.
				
				RemoveAllRows( false );
				Draw( NULL );
				
				//	Recursively sort all of the project folders, re-adding
				//	their items as we go.

				RebuildSortedProjectView( mAnchorItem, 0, true );

				Refresh();
				
				BroadcastMessage( CSpinIndicator::msg_SpinWaitEnd );
			
			}
		
		else if ( mAnchorItem == NULL )
			mCurrentSortMethod = inSortMethod;
			
	}


LComparator *
CMacCVSProjectView::GetCurrentComparator()
	{

		LComparator	*result	= NULL;
		
		switch ( mCurrentSortMethod ) 
			{
			
				case kNameSorted:
				
					result = new LSortByNameComparator;
					
				break;
				
				case kStatusSorted:
				
					result = new LSortByStatusComparator;
				
				break;
				
				case kVersionSorted:

					result = new LSortByVersionComparator;
				
				break;

				case kTagSorted:

					result = new LSortByTagComparator;	
				
				break;
				
				case kDateSorted:

					result = new LSortByDateComparator;
				
				break;
				
				default:
				
					Assert_( false );
					
				break;
			
			}
		
		return result;

	}

		
//################################################################################
//################################################################################
#pragma mark -
#pragma mark PROJECT VIEW THREAD


CMacCVSProjectViewThread::CMacCVSProjectViewThread( CMacCVSProjectView &inOwnerView ):
	CThreadedTask( (const UInt32)CMacCVSProjectView::kDefaultViewThreadStackSize )
	{
	
		mOwnerView = &inOwnerView;
		mNextUpdateTime = 0;
		
	}
	
	
CMacCVSProjectViewThread::~CMacCVSProjectViewThread()
	{
	
		mOwnerView->ThreadedTaskDied( this );
	
	}
	
	
void
CMacCVSProjectViewThread::Run()
	{
	
		CMacCVSProjectViewEventRec	nextEvent 			= { 0, 0 };
			
		while ( true ) 
			{
			
				try 
					{

						CThreadedTask::Yield();
							
						if ( mOwnerView->GetNextProjectViewEvent(nextEvent) )
							
							mOwnerView->ProcessProjectViewEvent( nextEvent );
		
						else if ( ::LMGetTicks() > mNextUpdateTime )
							{

								mOwnerView->RefreshViewContents();
								mNextUpdateTime = ::LMGetTicks() + mOwnerView->mWatchTime;
								
							}
						
					}
						
				catch ( CThreadedTaskException inErr )
					{
					
						throw inErr;
					
					}
										
				catch ( ... )
					{
					
						mOwnerView->BroadcastMessage( CSpinIndicator::msg_SpinNoMoreWaiters );
		
					}
				
			} // while
		
	}

	
//################################################################################
//################################################################################
#pragma mark -
#pragma mark STATIC COMPARATORS FOR PROJECT ITEMS

Int32
LSortByNameComparator::Compare( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32												result 					= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemTwo != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewItem::CompareName(inItemOneCasted, inItemTwoCasted);
			
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareDate(inItemOneCasted, inItemTwoCasted);
						
			//	if (result == 0)
			//		result = CMacCVSProjectViewItem::CompareStatus(inItemOneCasted, inItemTwoCasted);

			//	if (result == 0)
			//		result = CMacCVSProjectViewFile::CompareVersion(inItemOneCasted, inItemTwoCasted);
					
			}
		
		return result;
		
	}

Int32
LSortByNameComparator::CompareToKey(const void* inItem, Uint32 inSize, const void* inKey) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItem;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= (CMacCVSProjectViewItem *)inKey;
		Int32						result 				= 0x00000000;

		Assert_(inItem != NULL);
		Assert_(inKey != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewItem::CompareName(inItemOneCasted, inItemTwoCasted);
			
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareDate(inItemOneCasted, inItemTwoCasted);
						
			//	if (result == 0)
			//		result = CMacCVSProjectViewItem::CompareStatus(inItemOneCasted, inItemTwoCasted);

			//	if (result == 0)
			//		result = CMacCVSProjectViewFile::CompareVersion(inItemOneCasted, inItemTwoCasted);
					
			}
		
		return result;
		
	}


Boolean
LSortByNameComparator::IsEqualTo( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32											result 						= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewItem::IsEqualName(inItemOneCasted, inItemTwoCasted);
				
			}
		
		return result;
		
	}


Int32
LSortByTagComparator::Compare( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted = *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted = *(CMacCVSProjectViewItem **)inItemTwo;
		
		Str255		tag1		= "\p";
		Str255 		tag2		= "\p";

		Int32			result	= 0x00000000;
		
		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{

				if ( !inItemOneCasted->IsContainer() && inItemTwoCasted->IsContainer() )

					result = -1;

				else if ( inItemOneCasted->IsContainer() && !inItemTwoCasted->IsContainer() )
			
					result = 1;
					
				else if ( inItemOneCasted->IsContainer() && inItemTwoCasted->IsContainer() )
				
					result = 0; // let the name comparison below take care of this

				else
				{
					// both are files
					((CMacCVSProjectViewFile*)inItemOneCasted)->GetStickyTagPStr(tag1);
					((CMacCVSProjectViewFile*)inItemTwoCasted)->GetStickyTagPStr(tag2);
					
					// take the strncmp with the min lengths of the two strings
					if (tag1[0] == 0 && tag2[0] > 0)
					
						result = 1;
					
					else if (tag1[0] > 0 && tag2[0] == 0)
					
						result = -1;
					
					else				
					
						result = strncmp((const char*)tag1 + 1, (const char*)tag2 + 1, tag1[0] < tag2[0] ? tag1[0] : tag2[0]);
				
				}
				
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareName(inItemOneCasted, inItemTwoCasted);
					
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareDate(inItemOneCasted, inItemTwoCasted);
				
				if (result == 0)
					result = CMacCVSProjectViewFile::CompareVersion(inItemOneCasted, inItemTwoCasted);

			}
		
		return result;
		
	}


Boolean
LSortByTagComparator::IsEqualTo( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		CMacCVSProjectViewFile		*inItemOneFile		= NULL;
		CMacCVSProjectViewFile		*inItemTwoFile		= NULL;
		
		Str255		inItemOneTag		= "\p",
							inItemTwoTag		= "\p";
								
		Int32			result 					= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				if ( !inItemOneCasted->IsContainer() && !inItemTwoCasted->IsContainer() ) 
					{
					
						inItemOneFile = dynamic_cast<CMacCVSProjectViewFile *>(inItemOneCasted);
						inItemTwoFile = dynamic_cast<CMacCVSProjectViewFile *>(inItemTwoCasted);
						
						if ( (inItemOneFile != NULL) && (inItemTwoFile != NULL) ) 
							{
							
								inItemOneFile->GetStickyTagPStr( inItemOneTag );
								inItemTwoFile->GetStickyTagPStr( inItemTwoTag );
								
								result = EqualString( inItemOneTag, inItemTwoTag, true, true );
								
							}
					
					}
				
			}
		
		return result;
		
	}


Int32
LSortByDateComparator::Compare( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32											result 						= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewItem::CompareDate( inItemOneCasted, inItemTwoCasted );
				
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareName( inItemOneCasted, inItemTwoCasted );
				
		//	if (result == 0)
		//		result = CMacCVSProjectViewItem::CompareStatus(inItemOneCasted, inItemTwoCasted);

		//	if (result == 0)
		//		result = CMacCVSProjectViewFile::CompareVersion(inItemOneCasted, inItemTwoCasted);
			
			}
		
		return result;
		
	}


Boolean
LSortByDateComparator::IsEqualTo( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32											result 						= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewItem::IsEqualDate(inItemOneCasted, inItemTwoCasted);
				
			}
		
		return result;
		
	}


Int32
LSortByStatusComparator::Compare( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo ) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32					result 				= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
				
				result = CMacCVSProjectViewItem::CompareStatus(inItemOneCasted, inItemTwoCasted);
			
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareName(inItemOneCasted, inItemTwoCasted);
					
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareDate(inItemOneCasted, inItemTwoCasted);
				
				if (result == 0)
					result = CMacCVSProjectViewFile::CompareVersion(inItemOneCasted, inItemTwoCasted);

			}
		
		return result;
		
	}


Boolean
LSortByStatusComparator::IsEqualTo( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32											result 						= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 

			result =  CMacCVSProjectViewFile::IsEqualStatus(inItemOneCasted, inItemTwoCasted);
			
		return result;
		
	}


Int32
LSortByVersionComparator::Compare( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32											result 				= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewFile::CompareVersion(inItemOneCasted, inItemTwoCasted);
			
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareName(inItemOneCasted, inItemTwoCasted);
					
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareDate(inItemOneCasted, inItemTwoCasted);
					
				if (result == 0)
					result = CMacCVSProjectViewItem::CompareStatus(inItemOneCasted, inItemTwoCasted);
			
			}
		
		return (result);

	}


Boolean
LSortByVersionComparator::IsEqualTo( const void* inItemOne, const void* inItemTwo, Uint32 inSizeOne, Uint32 inSizeTwo) const
	{

		CMacCVSProjectViewItem		*inItemOneCasted 	= *(CMacCVSProjectViewItem **)inItemOne;
		CMacCVSProjectViewItem		*inItemTwoCasted 	= *(CMacCVSProjectViewItem **)inItemTwo;
		Int32					result 				= 0x00000000;

		Assert_(inItemOne != NULL);
		Assert_(inItemOne != NULL);

		if ( inItemOneCasted != inItemTwoCasted ) 
			{
			
				result = CMacCVSProjectViewFile::IsEqualVersion(inItemOneCasted, inItemTwoCasted);
			
			}
		
		return result;
		
	}

