// 
//	CPathPopup.cp
// 
//
//  Copyright (C) 1997-1998 Netscape Communications Corporation.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.
//
//
//	Description:
//
//		Provides the Finder like path popup for cmd-clicks in the title area
//		of a window.
//
//	Notes:
//
//		Users of this class will probably want to do so by overriding the
//		LWindow::ClickInDrag().  When a cmd-click is detected in the title
//		area of the window, do the following.
//		
//			CPathPopup thePathPop(this, theSpec);
//			thePathPop.Track();
//
//		This object relies on the following resources:
//
//			STR#	STRx_Popup	-	strings for menu building
//
// 

#include "UMenuUtilities.h"
#include "UProcessUtilities.h"
#include "UFileUtilities.h"
#include "CPathPopup.h"

#include "Find_icon.h"
#include "FullPath.h"		// MoreFiles

// MenuID we use when creating the new path menu (not a resource, just an ID)
const ResIDT	MENU_PathPopup	= 10004;

// 
//		CPathPopup
//
//	The only (valid) constructor.  It retrieves the necessary resource based
//	control and builds the path popup.
//
// 

CPathPopup::CPathPopup(
						const FSSpec&	inSpec,
						ResIDT 			inTextTraitsID,
						Boolean			inAllowOpen,
						Boolean			inSwitchToFinder,
						Boolean			inShowIcons)
	{

		OSErr				error = noErr;
		CInfoPBRec	pb = {0};
		
		mSpec = inSpec;
		mTraitsID = inTextTraitsID;
		mAllowOpen = inAllowOpen;
		mFinderSwitch = inSwitchToFinder;
		mShowIcons = inShowIcons;
		mIconSuites = new LArray();
		
		mMenu = ::NewMenu(MENU_PathPopup, "\p");
		
		mTargetIsDirectory = UFileUtils::IsFolder(inSpec);
	}
	

// 
//		~CPathPopup
// 


CPathPopup::~CPathPopup()
	{

		::DisposeMenu(mMenu);
		
		if ( mIconSuites != NULL )
			{
				LArrayIterator			itemIterator( *mIconSuites );
				Handle							theCurItem = NULL;
				
				while ( itemIterator.Next(&theCurItem) )
					{
						::DisposeIconSuite(theCurItem, true);
					}
						
				delete mIconSuites;
				mIconSuites = NULL;
			}

	}


// 
//		Track
//
//	Position the menu control so that the top item lies exactly on top
//	of the current window title.  Then track the click.  The incoming
//	click point parameter is in global coordinates.
//
// 

void CPathPopup::Track(Point inGlobalPoint)
	{

		Int16 	theOldFlash	= 0x0000;
	
		if (mMenu == NULL)
			return;
			
		BuildPathMenu();
			
		if (!mAllowOpen)// We turn of menu flashing
			{
				theOldFlash = LMGetMenuFlash();
				SetMenuFlash(0);
			}

		::InsertMenu(mMenu, hierMenu);
		Int32 theResult = UMenuUtils::PopupWithTraits( mMenu, inGlobalPoint, 1, mTraitsID );
		::DeleteMenu(MENU_PathPopup);

		if (!mAllowOpen)
			SetMenuFlash(theOldFlash);
		
		// opening is allowed, the menu was left on a valid item, and the file is specified
		if (mAllowOpen && (HiWord(theResult) != 0) && (mSpec.parID != 0))
			DoAEOpenFolder(LoWord(theResult));
			
	}

// 
//		SetMenuItemIcon
// 

void CPathPopup::SetMenuItemIcon(Int16 inPathLength, const char *inPartialPath, Int16 inItemIndex)
	{
	
		FSSpec			theCurItemSpec;
		OSErr				result = FSpLocationFromFullPath(inPathLength, inPartialPath, &theCurItemSpec);

		if (noErr == result)
		{
		if ( UEnvironment::IsAppearanceRunning() )
			{
				Handle			theIconSuite = nil;
				
				if ( Find_icon( &theCurItemSpec, NULL, kSelectorSmallThenLarge, &theIconSuite) == noErr)
					{
							result = ::SetMenuItemIconHandle( mMenu, inItemIndex, kMenuIconSuiteType, theIconSuite);

							ThrowIfNil_( mIconSuites );
							// don't forget to dispose the icon suites when done. We do this in the destructor
							mIconSuites->AddItem( &theIconSuite, sizeof(Handle) );
					}
			}
		else
			{
					Int16				iconID = UFileUtils::IsFolder(theCurItemSpec) ? kFolderIconID : kFileIconID;
				
					if (theCurItemSpec.parID == fsRtParID)
						iconID = kDiskIconID;
						
					UMenuUtils::SetItemIconFromResID( mMenu, inItemIndex, iconID);
			}
		}
		
	}
	

// 
//		BuildPathMenu
// 


void CPathPopup::BuildPathMenu(void)
	{

		OSErr		result					= noErr;
		Int16 	thePathLength		= 0;
		UInt32	index						= 0;
		Handle 	thePathHandle		= NULL;
	
		if ( (mSpec.parID != 0) && (mSpec.vRefNum != 0) )
			{
				result = ::FSpGetFullPath( &mSpec, &thePathLength, &thePathHandle );
				
				if ( (result == noErr) && (thePathLength > 0) )
					{
						::HLock( thePathHandle );
			
						while ( index < thePathLength )
							{
							
								TString<Str255> thePathStr;
							
								while ( (index < thePathLength) && ((*thePathHandle)[index] != ':') )
									{
										thePathStr += (*thePathHandle)[index];
										index++;
									}
								
								if ( index < thePathLength )
									index++;
							
								if ( thePathStr[0] > 0 )
									{
										Int16		itemIndex = UMenuUtils::InsertMenuItem( mMenu, thePathStr, 0, mAllowOpen );
										
										if (mShowIcons)
												SetMenuItemIcon(index, *thePathHandle, itemIndex);
											
									}
									
							}

						::HUnlock(thePathHandle);
						::DisposeHandle(thePathHandle);
					}
			}
		else
			{
		
				Str255 errorStr = "\pAn error has occurred.";
				AppendMenu( mMenu, errorStr );
				
			}
				
	}
		

// 
//		BuildPathMenu
// 

void CPathPopup::DoAEOpenFolder( Int16 inMenuValue )
	{

		OSErr									error					= noErr;
		AliasHandle						alias 				= NULL;
		AppleEvent						event 				= { typeNull, NULL };
		ProcessSerialNumber 	theFinderPSN	= { 0, 0 };
													
		if ( inMenuValue < 1 )
			return;

		try
			{
			
				// only do this if the finder is running
				
				if ( UProcessUtils::IsApplicationRunning('MACS', theFinderPSN) )
					{
					
						// reconstruct the path from the menu
						TString<Str255> thePath, theMenuItemText;
																	
						for ( UInt32 index = ::CountMenuItems(mMenu); index >= inMenuValue; index-- )
							{
							
								GetMenuItemText( mMenu, index, theMenuItemText );
								thePath += theMenuItemText;
								
								if ( index > inMenuValue )
									thePath += ':';
								
							}
	
//  If the top item is a directory, we definately need to append the ':'.
			
						if ( (inMenuValue > 1) || (mTargetIsDirectory) )
							thePath += ':';
							
						FSSpec theSpec = { 0, 0, "\p" };
						OSErr theErr = ::FSpLocationFromFullPath(thePath.Length(), &thePath[1], &theSpec);
						ThrowIfOSErr_(theErr);

						StAEDescriptor	editorAddress(	typeProcessSerialNumber, 
														&theFinderPSN, 
														sizeof(theFinderPSN));

						//	Create the AppleEvent.
						error = ::AECreateAppleEvent(	kCoreEventClass, 
														kAEOpen, 
														(const AEAddressDesc *) &editorAddress, 
														0, 
														0, 
														&event);
						
						if ( error != noErr )
							throw( error );
									
						//	Create an alias for this item								
						error = ::NewAlias(NULL, &theSpec, &alias);

						if ( error != noErr )
							throw( error );

						//	Stuff the alias in the AppleEvent as the 
						//	direct parameter.
						::HLock((Handle)alias);
						StAEDescriptor	aliasObject(typeAlias, *alias, ::GetHandleSize((Handle)alias));
						::AEPutParamDesc(&event, keyDirectObject, aliasObject);
						::HUnlock((Handle)alias);

						//	Send the open AppleEvent
						UAppleEventsMgr::SendAppleEvent(event, true);
						
						
						if (mFinderSwitch)
							UProcessUtils::SendAppToFront(theFinderPSN);
							
					}
					
			}
		
		catch (ExceptionCode inErr)
			{
			
				::SysBeep(0);		
				
			}

		if ( alias != NULL )
			::DisposeHandle( (Handle)alias );
				
		::AEDisposeDesc( &event );

	}






