//################################################################################
//
//  Copyright (C) 1997-1998 Netscape Communications Corporation.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//  02111-1307, USA.
//
//################################################################################

/* -------------------------------------------------------------------------------

	File:				UProjectActions.cp
	
	Authors:		Ram Rajadhyaksha and others (unknown)
	
	Description:
	
		This file contains code to implement operations from the
		menu commands.
		 
	Modification History:

		21 Apr 1999		Josef W. Wankerl
		
					Added Reveal In Finder support.
		
		10 Apr	1999		Josef W. Wankerl
		
					Made Navigation Services aware.
		
		10 Jan 1998			Ram Rajadhyaksha
		
					Made several changes for compatiblity with new Entries file scheme.
					
		07 Jul 1998			Ram Rajadhyaksha
		
					Made various changes to support new appearance manager classes. Added
				  #include <LPopupButton.h> for new grayscale pop-up class.
							
-------------------------------------------------------------------------------- */


#include <string.h>
#include <UStandardFile.h>
#include <LTextEditView.h>
#include <UModalDialogs.h>
#include <LAction.h>
#include <LCheckbox.h>
#include <LRadioButton.h>
#include <LPopupButton.h>
#include <LEditText.h>
#include <LCheckBoxGroupBox.h>
#include <LPopupGroupBox.h>
#include <UConditionalDialogs.h>		// JWW - For Navigation Services

#include "MacCVSPPConstants.h"

#include "ThrowOSErr.h"
#include "UErrorDialog.h"
#include "CTagPopup.h"
#include "CMacCVSDoc.h"

#include "CVSEngineUtils.h"
#include "StringUtils.h"
#include "CCVSMappings.h"

#include "CThreadedTask.h"
#include "CMacCVSAddTask.h"
#include "CMacCVSCommitTask.h"
#include "CMacCVSCheckoutTask.h"
#include "CMacCVSEditTask.h"
#include "CMacCVSImportTask.h"
#include "CMacCVSLogWindow.h"
#include "CMacCVSCompareWindow.h"
#include "CMacCVSStatusWindow.h"
#include "CMacCVSTagTask.h"
#include "CMacCVSUpdateTask.h"
#include "CMacCVSWatchTask.h"
#include "CMacCVSRemoveTask.h"

#include "UProjectActions.h"


//  this is used to adjust the initial delay on threads 
//  spawned for diffs and logs, to reduce the impacts of 
//  server locking

static const kThreadSpawnIntervalTicks = 25;


CProjectAction::CProjectAction(	CMacCVSDoc *inOwnerDoc, LArray *inItemList, ResIDT inStringResID, Int16 inStringIndex ):	

	LAction( inStringResID, inStringIndex, false )

	{
			
		ThrowIfNil_( inOwnerDoc );
		mOwnerDoc = inOwnerDoc;

		if ( inItemList != NULL )
			if ( inItemList->GetCount() > 0 )
				mItemList = *inItemList;		//makes a copy of the array

		mCurrentEvent.descriptorType = typeNull;
		mReplyEvent.descriptorType = typeNull;
	
	}


CProjectAction::~CProjectAction()
	{
		
	}


Boolean 
CProjectAction::IsPostable() const
	{
		
		return false;		// project actions do not affect undo state
										// and are not undoable.
	}


void 
CProjectAction::SetCompletionEvent( AppleEvent currentEvent, AppleEvent replyEvent )
	{
	
		mCurrentEvent = currentEvent;
		mReplyEvent = replyEvent;
	
	}


Boolean CProjectAction::CheckAllItemsQualify( Boolean inWarnOfOpenFiles )
	{

		Boolean bHaveAsked 					= false;
		Boolean	bShouldContinue 		= true;
		Boolean	bSomethingRemoved 	= false;
		Boolean	bAFileWasOpen				= false;
		UInt32	ioCount							= sizeof( FSSpec );
		
		LArray	openFilesList( sizeof(FSSpec) );
		
		FSSpec	theCurItem					= { 0, 0, "\p" };

		LArrayIterator 	itemIterator( mItemList );
				
		while ( itemIterator.Next(&theCurItem, ioCount) )
			{

				Assert_( ioCount == sizeof(FSSpec) );
				ioCount = sizeof( FSSpec );
				
				Boolean			thisFileOpen = false;
				
				if ( !IsItemQualified(theCurItem, &thisFileOpen) )
					{
					
						if (!bHaveAsked)
							{
/*
								TString<Str255> 	theProblem( STRx_ProjectActions, str_ActionPromptProblem );
								TString<Str255> 	theQuestion( STRx_ProjectActions, str_ActionPromptContinue );
								TString <Str255> 	theDescription, theDummy;
								
								GetDescription( theDescription, theDummy );
								
								TString<Str255> 	theMessage = theProblem + theDescription + '.' + '\r' + '\r' + theQuestion;
								
								bShouldContinue = UStdDialogs::AskOkCancel(theMessage, NULL, LCommander::GetTarget());
*/
								bShouldContinue = true;
								bHaveAsked = true;
							
							}
					
					if (bHaveAsked && bShouldContinue)
						{
						
							bSomethingRemoved = true;
							mItemList.RemoveItemsAt( 1, itemIterator.GetCurrentIndex() );
						
						}
					else if (bHaveAsked && !bShouldContinue)
							
						throw (ExceptionCode)kProjectActionCancelled;
						
				}
			else		// is qualified
				{
					if ( thisFileOpen )
						openFilesList.AddItem(&theCurItem);
				
					bAFileWasOpen |= thisFileOpen;
				}
		}
		
		if ( inWarnOfOpenFiles && bAFileWasOpen )
			{
				StDialogHandler			fileListDialog(kOpenFilesListDialogID, mOwnerDoc);
	
				LDialogBox 		*dialogBox = dynamic_cast<LDialogBox *>( fileListDialog.GetDialog() );	
				ThrowIfNil_( dialogBox );

				LListBox			*filesList = dynamic_cast<LListBox *>( dialogBox->FindPaneByID(kOpenFilesListTableID) );
				ThrowIfNil_( filesList );
				
				LPushButton		*okButton = dynamic_cast<LPushButton *>( dialogBox->FindPaneByID(kOpenFilesListOKButtonID) );
				ThrowIfNil_( okButton );
				
				ListHandle		theList = filesList->GetMacListH();
				Cell					theCell = { 0, 0 };
				
				LAddColumn( 1, 0, theList );
				LAddRow( openFilesList.GetCount(), 0, theList );
				
				LArrayIterator 	openFilesIterator( openFilesList );
				
				FSSpec		thisOpenFile;
				ioCount = sizeof( FSSpec );
				
				while ( openFilesIterator.Next(&thisOpenFile, ioCount) )
					{
											
						LSetCell( &thisOpenFile.name[1], thisOpenFile.name[0], theCell, theList );
						theCell.v ++;
						
					}
					
				dialogBox->SetDefaultButton( kOpenFilesListOKButtonID );
				dialogBox->SetCancelButton( kOpenFilesListCancelButtonID );

				dialogBox->Show();
			
				Boolean done = false;
				Boolean	dialogActive = IsWindowHilited(dialogBox->GetMacPort());
				
				while ( !done )
					{
					
						switch ( fileListDialog.DoDialog() )
							{ 
							
								case msg_OK:
									// remove those files that are still open from the list of items
									// to avoid stupid errors later
									
									openFilesIterator.ResetTo(LArrayIterator::from_Start);
									ioCount = sizeof( FSSpec );

									while ( openFilesIterator.Next(&thisOpenFile, ioCount) )
										{
											mItemList.Remove(&thisOpenFile);
										}

									done = true;
									break;
									
								case msg_Cancel:
									bShouldContinue = false;
									done = true;
									break;
								
								default:
									// find out if the user switch back to MacCVS after closing some windows
									// in other apps. Yeah, I could subclass LDialoxBox and override
									// Suspend/Resume, but this is cheaper and cheesier.
									
									if ( !dialogActive && IsWindowHilited(dialogBox->GetMacPort()) )
										{
											Cell		testCell = { 0, 0 };
											
											itemIterator.ResetTo(LArrayIterator::from_Start);
											
											// clear the list of open files, prior to rebuilding it
											openFilesList.RemoveAllItemsAfter(0);
											
											// clear the list also
											LDelRow( (**theList).dataBounds.bottom, 0, theList);
											
											ioCount = sizeof( FSSpec );
											
											while ( itemIterator.Next(&thisOpenFile, ioCount) )
												{
													LStr255			listItem(thisOpenFile.name);
													Boolean			fileStillOpen = false;
													
													if ( FSpIsAFile( &thisOpenFile, &fileStillOpen ) == noErr )
														{
															
															if ( fileStillOpen )
																{
																	openFilesList.AddItem(&thisOpenFile);
																	
																	LAddRow(1, testCell.v, theList );
																	LSetCell( &thisOpenFile.name[1], thisOpenFile.name[0], testCell, theList );
																	
																	testCell.v ++;
																}
														}

												}		// while iterating
												
											filesList->Refresh();
											
											if ( testCell.v == 0)			// no open files
												{
													Str255			okString = "\p      [No files remain open]";
													
													LAddRow(1, testCell.v, theList );
													LSetCell( &okString[1], okString[0], testCell, theList );
												
												}
											
										}		// we we just became frontmost
										
									dialogActive = IsWindowHilited(dialogBox->GetMacPort());
									
									// disable the OK button if there are no files left to process
									if (mItemList.GetCount() > openFilesList.GetCount() )
											okButton->Enable();
									else
											okButton->Disable();

							}
					}
					
					if ( ! bShouldContinue )
						throw (ExceptionCode)kProjectActionCancelled;
			
			}
		
		return !bSomethingRemoved;
		
	}


void CProjectAction::UndoSelf(void)
	{		
		Assert_( false );  // project actions are not undoable
	}


#pragma mark -

// 
//		CProjectRevealAction
// JWW - Added Reveal In Finder action class
// 


CProjectRevealAction::
CProjectRevealAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
	}


Boolean 
CProjectRevealAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr || error == kCVSEngineIsFolder )
			{	
				bResult = true;	
			}
			
		return bResult;
		
	}


void 
CProjectRevealAction::RedoSelf(void)
	{
		
		UInt32	ioCount			= sizeof( FSSpec );
		FSSpec	theCurItem	= { 0, 0, "\p" };
		UInt32	delayTicks 	= 0;
		
		try 
			{

				CheckAllItemsQualify();
				
				LArrayIterator	itemIterator( mItemList );
				
				while( itemIterator.Next(&theCurItem, ioCount) )
					{
				
						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof( FSSpec );
						
						AliasHandle		alias 		= NULL;
						AppleEvent		event 		= { typeNull, NULL };
						AppleEvent		event2 		= { typeNull, NULL };

						try 
						{
							
							ProcessSerialNumber		finderPSN;
							OSErr					error;
							
							//	Get the serial number of the Finder... we will
							//	use this as the destination for our AppleEvent.		
							ProcessInfoRec 		info;
							
							finderPSN.highLongOfPSN = 0;
							finderPSN.lowLongOfPSN  = kNoProcess;
							
							while (1)
								{
									error = ::GetNextProcess(&finderPSN);
									if( error == noErr )
										{
											info.processInfoLength = sizeof(ProcessInfoRec);
											info.processName = NULL;
											info.processAppSpec = NULL;
											error = GetProcessInformation(&finderPSN, &info);
										}
									
									if ((error != noErr) ||
										(info.processSignature == 'MACS') ||
										(info.processType == 'FNDR'))
										break;
									
								}
							
							if ( error != noErr )
								throw( error );

							StAEDescriptor	editorAddress(	typeProcessSerialNumber, 
															&finderPSN, 
															sizeof(finderPSN));

							//	Create the AppleEvent.
							error = ::AECreateAppleEvent(	kAEMiscStandards, 
															kAEMakeObjectsVisible, 
															(const AEAddressDesc *) &editorAddress, 
															0, 
															0, 
															&event);
							
							if ( error != noErr )
								throw( error );
										
							//	Create an alias for this item								
							error = ::NewAlias(NULL, &theCurItem, &alias);

							if ( error != noErr )
								throw( error );

							//	Stuff the alias in the AppleEvent as the 
							//	direct parameter.
							::HLock((Handle)alias);
							StAEDescriptor	aliasObject(typeAlias, *alias, ::GetHandleSize((Handle)alias));
							::AEPutParamDesc(&event, keyDirectObject, aliasObject);
							::HUnlock((Handle)alias);

							//	Send the open AppleEvent
							UAppleEventsMgr::SendAppleEvent(event, true);
							
							//	Create the AppleEvent.
							error = ::AECreateAppleEvent(	kAEMiscStandards, 
															kAESelect, 
															(const AEAddressDesc *) &editorAddress, 
															0, 
															0, 
															&event2);
							
							if ( error != noErr )
								throw( error );
										
							::AEPutParamDesc(&event2, keyDirectObject, aliasObject);
							
							//	Send the open AppleEvent
							UAppleEventsMgr::SendAppleEvent(event2, true);
							
							::SetFrontProcess(&finderPSN);
						}	
						
						catch(...)
						{
							::SysBeep(1);	
						}
						
						if (alias != NULL)
							::DisposeHandle((Handle)alias);
							
						::AEDisposeDesc(&event);
						::AEDisposeDesc(&event2);
						
					}
					
			}
			
		catch ( ExceptionCode inErr )
			{
			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			
			}
			
	}

#pragma mark -

// 
//		CProjectCheckoutDefaultAction
// 


CProjectCheckoutDefaultAction::
CProjectCheckoutDefaultAction( CMacCVSDoc *inOwnerDoc ):

	CProjectAction( inOwnerDoc )
	
	{
		mCustomOptions.Init();
		mUseCustomOptions = false;				
	}


CProjectCheckoutDefaultAction::
CProjectCheckoutDefaultAction( CMacCVSDoc *inOwnerDoc, CCVSEngineCheckoutOptions *inOptions ):

	CProjectAction( inOwnerDoc )
	
	{
		ThrowIfNil_( inOptions );
		
		mCustomOptions = *inOptions;
		mUseCustomOptions = true;				
	}


CProjectCheckoutDefaultAction::
~CProjectCheckoutDefaultAction()
	{
	}


Boolean 
CProjectCheckoutDefaultAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{		
		return true;				
	}


void 
CProjectCheckoutDefaultAction::RedoSelf(void)
	{				
	
		Str255 pstrDefaultTag	= "\p";
		CCVSEngineCheckoutOptions sCheckoutOptions;
		
		try
			{
				sCheckoutOptions.Init();
						
				if ( !mUseCustomOptions )		//checking out default module
					{					
						mOwnerDoc->GetModulePStr( sCheckoutOptions.mModulePStr );					
						sCheckoutOptions.mMergePolicy = mOwnerDoc->GetMergePolicy();
						sCheckoutOptions.mLFTranslation = mOwnerDoc->GetTranslation();
						sCheckoutOptions.mFolderOptions.mFoldersGetListView = mOwnerDoc->GetNewFoldersUseListView();
						sCheckoutOptions.mFolderOptions.mFoldersInheritLabels = mOwnerDoc->GetNewFoldersInheritLabels();
						sCheckoutOptions.mCompressionLevel = mOwnerDoc->GetCompressionLevel();	
												
						if ( sCheckoutOptions.mCompressionLevel > 0 )
							{
								sCheckoutOptions.mUseCompression = true;
							}						
					
						mOwnerDoc->GetDefaultTagPStr( pstrDefaultTag );
						Assert_( pstrDefaultTag[0] != 1 );
						
						if ( pstrDefaultTag[0] > 1 )
							{							
								if ( pstrDefaultTag[1] == 'T' )
									{
										sCheckoutOptions.mRevisionPStr[0] = pstrDefaultTag[0] - 1;
										::BlockMoveData( &pstrDefaultTag[2], &sCheckoutOptions.mRevisionPStr[1], sCheckoutOptions.mRevisionPStr[0] );							
									}
								else if ( pstrDefaultTag[1] == 'D' )
									{								
										sCheckoutOptions.mDatePStr[0] = pstrDefaultTag[0] - 1;
										::BlockMoveData( &pstrDefaultTag[2], &sCheckoutOptions.mDatePStr[1], sCheckoutOptions.mDatePStr[0] );								
									}
								else Assert_( false );							
							}
							
					}
				else
					{
						sCheckoutOptions = mCustomOptions;
					}
					
				CMacCVSCheckoutTask *pTask = new CMacCVSCheckoutTask( &sCheckoutOptions, mOwnerDoc );
				ThrowIfNil_( pTask );
				
				pTask->SetCompletionEvent( mCurrentEvent, mReplyEvent );
				pTask->Resume();
			}
			
		catch ( ... )
			{	
			//  Should we delete pTask if we have an exception?
				throw;
			}
				
	}
	
#pragma mark -

// 
//		CProjectCheckoutModuleToAction
// 


CProjectCheckoutModuleToAction::
CProjectCheckoutModuleToAction( CMacCVSDoc *inOwnerDoc ):

	CProjectAction( inOwnerDoc )
	
	{				
	}


Boolean 
CProjectCheckoutModuleToAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{		
		return true;				
	}


void 
CProjectCheckoutModuleToAction::RedoSelf(void)
	{
				
		Boolean bUserCancelled 	= false;
		Str255	pstrTag					= "\p";
		
		CCVSEngineCheckoutOptions	sCheckoutOptions;
		
		try
			{		
				sCheckoutOptions.Init();
				
				if ( mOwnerDoc->IsProjectRootSpecified() )
					{
						bUserCancelled = GetDesiredModule( sCheckoutOptions.mModulePStr, pstrTag );
						
						if ( !bUserCancelled )
							{
								sCheckoutOptions.mMergePolicy = mOwnerDoc->GetMergePolicy();
								sCheckoutOptions.mLFTranslation = mOwnerDoc->GetTranslation();
								sCheckoutOptions.mCompressionLevel = mOwnerDoc->GetCompressionLevel();	
															
								if ( sCheckoutOptions.mCompressionLevel > 0 )
									{
										sCheckoutOptions.mUseCompression = true;
									}						
			
								if ( pstrTag[0] > 1 )
									{								
										if ( pstrTag[1] == 'T' )
											{
												sCheckoutOptions.mRevisionPStr[0] = pstrTag[0] - 1;
												::BlockMoveData( &pstrTag[2], &sCheckoutOptions.mRevisionPStr[1], sCheckoutOptions.mRevisionPStr[0] );								
											}
										else if ( pstrTag[1] == 'D' )
											{										
												sCheckoutOptions.mDatePStr[0] = pstrTag[0] - 1;
												::BlockMoveData( &pstrTag[2], &sCheckoutOptions.mDatePStr[1], sCheckoutOptions.mDatePStr[0] );									
											}
										else Assert_( false );										
									}
	
								sCheckoutOptions.mUpdateEverything = false;
						
								CMacCVSCheckoutTask *pTask = new CMacCVSCheckoutTask( &sCheckoutOptions, mOwnerDoc );
								ThrowIfNil_( pTask );
						
								pTask->Resume();
							}								
					}					
			}
		catch ( ... )
			{
				throw;			
			}
				
	}


Boolean
CProjectCheckoutModuleToAction::GetDesiredModule( Str255 pstrOutModule, Str255 pstrOutTag )
	{
			
		Boolean			bUserCancelled	= false;
		Boolean			bDone 					= false;
		Boolean			bTagsModified		= false;
													
		MessageT		intItemHit			= 0;
		LArray			*pCurTagList		= NULL;
		LArray			*pNewTagList		= NULL;
		
		Str255			pstrCurTag			= "\p";
		Str255			pstrModule			= "\p";
		Str255			pstrDefaultTag	= "\p";
		
		try
			{		
				StDialogHandler	stHandler( kCheckoutModuleToDialogPPobID, mOwnerDoc );						
				stHandler.AddAttachment( new CThreadedTaskYieldAttachment );
				
				LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
				ThrowIfNil_( pDialog );
				
				LPushButton *pRemoveButton = dynamic_cast<LPushButton *>( pDialog->FindPaneByID(kCheckoutModuleToRemoveButtonID) );
				ThrowIfNil_( pRemoveButton );
						
				LEditText *pModuleField = dynamic_cast<LEditText *>( pDialog->FindPaneByID(kCheckoutModuleToModuleFieldID) );
				ThrowIfNil_( pModuleField );
						
				CTagPopup *pTagPopup = dynamic_cast<CTagPopup *>( pDialog->FindPaneByID(kCheckoutModuleToTagPopupID) );
				ThrowIfNil_( pTagPopup );

//  Set up tag list in the popup. CurTagList is not owned by us
//  so we should never delete it.
				
				pCurTagList = mOwnerDoc->GetTagListPtr();
				Assert_( pCurTagList != NULL );
						
				if ( pCurTagList != NULL )
					{
						pTagPopup->SetTagList( *pCurTagList );
					}
											
				mOwnerDoc->GetDefaultTagPStr( pstrDefaultTag );
				pTagPopup->SetDefaultTag( pstrDefaultTag );

				pDialog->Show();
						
				while ( !bDone )
					{					
						intItemHit = stHandler.DoDialog();
						
						switch ( intItemHit )
							{						
								case msg_OK:
								case msg_Cancel:

									pModuleField->GetDescriptor( pstrModule );
									
									if ( (intItemHit == msg_OK) && (pstrModule[0] == 0) )
										{
										
//  Don't allow zero length module names.												
												
											UErrorDialog theErrorDialog( kNoZeroLengthModuleNameErr, mOwnerDoc );												
										}												
									else
										{										
											bDone = true;
										}									
								break;
								
								case kCheckoutModuleToRemoveButtonValueMessage:								
									pTagPopup->RemoveCurrentTag();

									if ( pTagPopup->CanRemoveCurrentTag() )
										{
											pRemoveButton->Enable();
										}
									else 
										{
											pRemoveButton->Disable();
										}
								break;
								
								case kCheckoutModuleToTagPopupValueMessage:	
									pTagPopup->DoMouseDownInTagPopup();
									
									if ( pTagPopup->CanRemoveCurrentTag() )
										{
											pRemoveButton->Enable();
										}
									else 
										{
											pRemoveButton->Disable();
										}
								break;
							}
					} /* while */
					
				pDialog->Hide();
							
				pNewTagList = pTagPopup->GetTagList();								
				Assert_( pNewTagList != NULL );
				
				if ( pNewTagList != NULL )
					{											
						bTagsModified = !mOwnerDoc->AreTagListsEqual( *pCurTagList, *pNewTagList );
						
						if ( bTagsModified )	
							{															
								mOwnerDoc->SetModified( true );
								mOwnerDoc->UseNewTagList( *pNewTagList );													
							}
							
						delete pNewTagList;
						pNewTagList = NULL;									
					}
					
				if ( intItemHit == msg_OK )
					{
						pTagPopup->GetCurrentTag( pstrOutTag );
						pModuleField->GetDescriptor( pstrOutModule );
					}
				else
					{	
						bUserCancelled = true;
					}
			}
			
		catch ( ... )
			{
				if ( pNewTagList != NULL )
					{
						delete pNewTagList;
						pNewTagList = NULL;
					}
					
				throw;		
			}
	
		return bUserCancelled;
		
	}


#pragma mark -

// 
//		CProjectAddAction
// 


CProjectAddAction::
CProjectAddAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, Boolean inAskOptions ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		mAskForEncoding = inAskOptions;				
	}


Boolean 
CProjectAddAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr			error					= noErr;
		Boolean		bResult				= false;
		FSSpec		fsParent			= { 0, 0, "\p" };
		Str255		pstrRevision 	= "\p";
		
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{			
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{			
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );
								
								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry(inTargetItem.name, pstrRevision, NULL, NULL, NULL);
						
						if ( error == kCVSEngineNoRevisionDataErr )
							{
								bResult = true;		
							}
						else if ( error == noErr )
							{
								if ( (pstrRevision[0] > 0) && (pstrRevision[1] == '-') )
									bResult = true;		
							}
					}
			}	
		else if ( error == kCVSEngineIsFolder )
			{		
				if ( !FSpIsRepositoryDir(&inTargetItem) )
					bResult = true;		
			}
	
		return bResult;
		
	}


/* -------------------------------------------------------------------

	Instead of spawning a new thread for each file or directory,
	we now collect all the data first, and then spawn a thread 
	for each directory that contains files. That reduces the locking
	slowdown, and is generally a Good Thing.

---------------------------------------------------------------------*/

void 
CProjectAddAction::RedoSelf(void)
	{
		
		OSErr											result						= noErr;
		Boolean										bUserCancelled		= false;
		Boolean										bSkipThisFile			= false;
		Boolean										bHasMapping				= false;
		Boolean										bIsFile						= false;
		Boolean										bMappingsChanged	= false;
		UInt32										intIndex 					= 0;
		UInt32										intMaxIndex 			= 0;
		UInt32										intSize						= 0;
		UInt32										intMappingsCount	= 0;
		
		Str255										pstrTemp					= "\p";
		LArray										*pItemList				= 0;
		LArray										*pNewMappingsList	= NULL;

		FSSpec										fsItem						= { 0, 0, "\p" };
		CCVSEngineFileMappingRec	*pMappingsList		= NULL;
		CCVSEngineFileMappingRec	sMapping;

		CCVSMappings							cMappings;
		
		CMacCVSAddTask::CMacCVSAddTaskItem sTaskItem;
		
		sMapping.Init();
		sTaskItem.Init();
					
		try 
			{			
				pItemList = new LArray( sizeof(sTaskItem) );
				ThrowIfNil_( pItemList );
						
				CheckAllItemsQualify( true );
								
				if ( mItemList.GetCount() > 0 )
					{
						pNewMappingsList = mOwnerDoc->GetFileMappingList();
					
						if ( pNewMappingsList == NULL )
							{
								pNewMappingsList = new LArray( sizeof(CCVSEngineFileMappingRec) );
								ThrowIfNil_( pNewMappingsList );
							}
							
						if ( pNewMappingsList->GetCount() > 0 )
							{
								result = CMacCVSDoc::BuildFileMappingList( pNewMappingsList, &pMappingsList, &intMappingsCount );
								ThrowIfOSErr_( result );
							
								if ( pMappingsList != NULL )
									{
										result = cMappings.SetMappings( pMappingsList, intMappingsCount );
										ThrowIfOSErr_( result );
									}						
							}
														
						intMaxIndex = mItemList.GetCount();
						
						for ( intIndex = 1; (intIndex <= intMaxIndex) && (!bUserCancelled); intIndex++ )
							{								
								bSkipThisFile = false;

								intSize = sizeof(FSSpec);
								mItemList.FetchItemAt( intIndex, &fsItem, intSize );
								Assert_( intSize == sizeof(FSSpec) );
								
								result = FSpIsAFile( &fsItem, NULL );
								
								if ( result == noErr )
									{
										bIsFile = true;
									}
								else if ( result == kCVSEngineIsFolder )
									{
										bIsFile = false;
									}
								else
									{
										ThrowOSErr_( result );
									}
									
								if ( bIsFile )
									{
										bHasMapping = cMappings.GetMapping( fsItem.name, sMapping, false );
										
										if ( bHasMapping && (sMapping.fFileNamePStr[0] > 0) )
											{			
												//  Bug fix, lowercase both items before
												//  calling PStrCompare.
												PStrLowercase( sMapping.fFileNamePStr	);
												PStrCopy( fsItem.name, pstrTemp, sizeof(pstrTemp) );
												PStrLowercase( pstrTemp );
																				
												if ( (sMapping.fFileNamePStr[1] == '.') || (PStrCompare(sMapping.fFileNamePStr, pstrTemp) == true) )
													{
														//  If the mapping starts with a '.', then it is most
														//  likely a file extension. So the mapping probably applies
														//  to the file we're adding. 
												
														//  If not, check to see if the mapping is the same as us.
														//  If it isn't, we should ask.
													}
												else
													{
														bUserCancelled = !GetEncodingPreference( fsItem, sMapping, bSkipThisFile );		
															
														if ( !bUserCancelled && !bSkipThisFile )
															{
																pNewMappingsList->InsertItemsAt( 1, LArray::index_Last, &sMapping, sizeof(sMapping) );
																bMappingsChanged = true;

																result = cMappings.AddMapping( &sMapping );
																ThrowIfOSErr_( result );
															}											
													}
											}
										else
											{
												//  If the mapping name is 0, then it was the
												//  default mapping. So we better ask.
												bUserCancelled = !GetEncodingPreference( fsItem, sMapping, bSkipThisFile );
															
												if ( !bUserCancelled && !bSkipThisFile )
													{
														pNewMappingsList->InsertItemsAt( 1, LArray::index_Last, &sMapping, sizeof(sMapping) );
														bMappingsChanged = true;
														
														result = cMappings.AddMapping( &sMapping );
														ThrowIfOSErr_( result );
													}											
											}
									}

								if ( !bUserCancelled && !bSkipThisFile )
									{
										sTaskItem.Init();
										sTaskItem.mFileSpec = fsItem;
										sTaskItem.mOptions.mAutomagicallyEncode = true;
										sTaskItem.mOptions.mEncodingOptions = 0;
								
										pItemList->InsertItemsAt( 1, 0x7FFFFFFF, &sTaskItem, sizeof(sTaskItem) );
									}
							}					
					}

				if ( !bUserCancelled )
					{
						if ( bMappingsChanged )
							{
								mOwnerDoc->SetFileMappingList( *pNewMappingsList );
							}							
							
						if ( pItemList->GetCount() > 0 )
							{
								CMacCVSAddTask *pTask = new CMacCVSAddTask( pItemList, mOwnerDoc );		
								ThrowIfNil_( pTask );
								
								pTask->Resume();
							}	
					}
					
				delete pItemList;
				pItemList = NULL;

				if ( pMappingsList != NULL )
					{
						::DisposePtr( (Ptr)pMappingsList );
						pMappingsList = NULL;
					}
				
				if ( pNewMappingsList != NULL )
					{
						delete pNewMappingsList;
						pNewMappingsList = NULL;
					}
			}	// try
			
		catch ( ExceptionCode inErr )
			{	
				if ( pItemList != NULL )
					{			
						delete pItemList;
						pItemList = NULL;
					}
				
				if ( pMappingsList != NULL )
					{
						::DisposePtr( (Ptr)pMappingsList );
						pMappingsList = NULL;
					}

				if ( pNewMappingsList != NULL )
					{
						delete pNewMappingsList;
						pNewMappingsList = NULL;
					}
						
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			}
								
	}


Boolean
CProjectAddAction::GetEncodingPreference( const FSSpec &fsInItem, CCVSEngineFileMappingRec &sOutMapping, Boolean &bOutSkipThisFile )
	{
		
		Boolean 				bResult							= false;
		Boolean					bDone								= false;
			
		MessageT				intItemHit					= 0;	

		LDialogBox			*pDialog						= NULL;
		LRadioButton		*pBinaryButton			= NULL;
		LRadioButton		*pAppleSingleButton	= NULL;
		LRadioButton		*pBinhexButton			= NULL;
		LRadioButton		*pMacBinaryButton		= NULL;
		LPushButton			*pSkipButton				= NULL;										
		LCheckBox				*pUseForAllCheckBox	= NULL;										
		LCaption				*pFileCaption				= NULL;
		LCaption				*pEncodingCaption		= NULL;
		LPopupGroupBox	*pAddFileAsPopup		= NULL;
		
		Str255					pstrBinaryCaption		= "\p";
		Str255					pstrEncodedCaption	= "\p";
		Str255					pstrTextCaption			= "\p";

		LStr255					pstrCaption;
		
		FInfo						sFileInfo;
		FSSpec					fsItem							= fsInItem;
		
		memset( &sFileInfo, 0, sizeof(sFileInfo) );
		
		if ( FSpIsValid(fsItem) )
			{
				ThrowIfError_( FSpGetFInfo(&fsItem, &sFileInfo) );
			}

		sOutMapping.Init();
		bOutSkipThisFile = false;
			
		StDialogHandler	stDialogHandler( kMacCVSAddDialogPPobID, mOwnerDoc );
		stDialogHandler.AddAttachment( new CThreadedTaskYieldAttachment );
		
		pDialog = dynamic_cast<LDialogBox *>( stDialogHandler.GetDialog() );
		ThrowIfNil_( pDialog );
																										
		pFileCaption = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kMacCVSAddDialogFileCaptionID) );
		ThrowIfNil_( pFileCaption );

		pEncodingCaption = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kMacCVSAddDialogEncodingCaptionID) );
		ThrowIfNil_( pEncodingCaption );
		
		pBinaryButton = dynamic_cast<LRadioButton *>( pDialog->FindPaneByID(kMacCVSAddDialogBinaryButtonID) );
		ThrowIfNil_( pBinaryButton );

		pAppleSingleButton = dynamic_cast<LRadioButton *>( pDialog->FindPaneByID(kMacCVSAddDialogAppleSingleButtonID) );
		ThrowIfNil_( pAppleSingleButton );

		pBinhexButton = dynamic_cast<LRadioButton *>( pDialog->FindPaneByID(kMacCVSAddDialogBinhexButtonID) );
		ThrowIfNil_( pBinhexButton );

		pMacBinaryButton = dynamic_cast<LRadioButton *>( pDialog->FindPaneByID(kMacCVSAddDialogMacBinaryButtonID) );
		ThrowIfNil_( pMacBinaryButton );

		pSkipButton =  dynamic_cast<LPushButton *>( pDialog->FindPaneByID(kMacCVSAddDialogCancelButtonID) );	// this is the skip button
		ThrowIfNil_( pSkipButton );
				
		pAddFileAsPopup = dynamic_cast<LPopupGroupBox *>( pDialog->FindPaneByID(kMacCVSAddDialogEncodingPopupID) );
		ThrowIfNil_( pAddFileAsPopup );
											
		::GetIndString( pstrTextCaption, STRx_AddHelpCaptions, str_NonBinaryText );
		::GetIndString( pstrBinaryCaption, STRx_AddHelpCaptions, str_BinaryText );
		::GetIndString( pstrEncodedCaption, STRx_AddHelpCaptions, str_AppleSingleText );
										
		pstrCaption[0] = 0;
		pstrCaption.Append( "\pYou are about to add the file " );
		pstrCaption.Append( fsItem.name );
		pstrCaption.Append( "\p to the repository. Please choose an encoding method below:" );
										
		pFileCaption->SetDescriptor( pstrCaption );

		pAddFileAsPopup->SetValue( 1 );
		pEncodingCaption->SetDescriptor( pstrTextCaption );

		pBinaryButton->Disable();
		pAppleSingleButton->Disable();
		pBinhexButton->Disable();
		pMacBinaryButton->Disable();
									
		if ( sFileInfo.fdType == 'TEXT' )
			{

			}
		else if ( sFileInfo.fdType == 'MMPr' || sFileInfo.fdType == 'SITD' )		// CW projects and Stuffit archives
			{
				pAddFileAsPopup->SetValue( 2 );
				pBinaryButton->Enable();
				pAppleSingleButton->Enable();
				pBinhexButton->Enable();
				pMacBinaryButton->Enable();

				pBinaryButton->SetValue( 1 );

				pEncodingCaption->SetDescriptor( pstrBinaryCaption );
			}
		else
			{
				pAddFileAsPopup->SetValue( 2 );
				pBinaryButton->Enable();
				pAppleSingleButton->Enable();
				pBinhexButton->Enable();
				pMacBinaryButton->Enable();

				pAppleSingleButton->SetValue( 1 );
				
				pEncodingCaption->SetDescriptor( pstrBinaryCaption );
			}
											
		pDialog->Show();
	
		while ( !bDone )
			{			
				intItemHit = stDialogHandler.DoDialog();
			
				if ( intItemHit > 0 )
					{
						switch ( intItemHit )
							{						
								case msg_OK:
									bResult = true;
									bDone = true;

									sOutMapping.fFileType = sFileInfo.fdType;
									sOutMapping.fFileCreator = sFileInfo.fdCreator;									
									sOutMapping.fTranslation = kCVSEngineCRTranslation;
									PStrCopy( fsItem.name, sOutMapping.fFileNamePStr, sizeof(sOutMapping.fFileNamePStr) );
									
									if ( pAddFileAsPopup->GetValue() == 1 )
										{
											sOutMapping.fEncoding = kCVSEngineTextEncoding;
										}
									else
										{
											if ( pBinaryButton->GetValue() > 0 )
												{
													sOutMapping.fEncoding = kCVSEngineBinaryEncoding;
												}
											else if ( pAppleSingleButton->GetValue() > 0 )
												{
													sOutMapping.fEncoding = kCVSEngineAppleSingleEncoding;
												}
											else if ( pBinhexButton->GetValue() > 0 )
												{
													sOutMapping.fEncoding = kCVSEngineBinhex4Encoding;
												}
											else if ( pMacBinaryButton->GetValue() > 0 )
												{
													sOutMapping.fEncoding = kCVSEngineMacBinaryEncoding;
												}
											else
												{
													Assert_( false );
												}
										}
									
								break;
								
								case msg_Cancel:			// skip								
									bDone = true;	
									bResult = true;	
									bOutSkipThisFile = true;							
								break;
							
								case msg_CancelAll:		// cancel								
									bDone = true;
								break;
									
								case kMacCVSAddDialogEncodingPopupID:
								
									if ( pAddFileAsPopup->GetValue() == 1 )
										{
											pBinaryButton->Disable();
											pAppleSingleButton->Disable();
											pBinhexButton->Disable();
											pMacBinaryButton->Disable();		
											
											pEncodingCaption->SetDescriptor( pstrTextCaption );							
										}
									else
										{
											pBinaryButton->Enable();
											pAppleSingleButton->Enable();
											pBinhexButton->Enable();
											pMacBinaryButton->Enable();																			

											if ( pBinaryButton->GetValue() > 0 )
												{
													pEncodingCaption->SetDescriptor( pstrBinaryCaption );
												}
											else if ( (pAppleSingleButton->GetValue() > 0) ||
													 (pBinhexButton->GetValue() > 0) || 
													 (pMacBinaryButton->GetValue() > 0) )
												{
													pEncodingCaption->SetDescriptor( pstrEncodedCaption );
												}
										}									
								break;
								
								case kMacCVSAddDialogBinaryButtonID:
								case kMacCVSAddDialogAppleSingleButtonID:
								case kMacCVSAddDialogBinhexButtonID:
								case kMacCVSAddDialogMacBinaryButtonID:
									
									if ( pBinaryButton->GetValue() > 0 )
										{
											pEncodingCaption->SetDescriptor( pstrBinaryCaption );
										}
									else if ( (pAppleSingleButton->GetValue() > 0) ||
											 (pBinhexButton->GetValue() > 0) || 
											 (pMacBinaryButton->GetValue() > 0) )
										{
											pEncodingCaption->SetDescriptor( pstrEncodedCaption );
										}								
								break;
									
								case msg_ControlClicked:

									if ( pBinaryButton->GetValue() > 0 )
										{
											pEncodingCaption->SetDescriptor( pstrBinaryCaption );
										}
									else if ( (pAppleSingleButton->GetValue() > 0) ||
											 (pBinhexButton->GetValue() > 0) || 
											 (pMacBinaryButton->GetValue() > 0) )
										{
											pEncodingCaption->SetDescriptor( pstrEncodedCaption );
										}								
								break;
							}
					}
			} // while (! bDone)
			
		return bResult;
		
	}	


#pragma mark -

// 
//		CProjectCommitAction
// 


CProjectCommitAction::
CProjectCommitAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean bInAsk ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{		
		mAskUser = bInAsk;
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );				
	}


Boolean 
CProjectCommitAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{	
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		SInt32							intModDate			= 0;
		UInt32							intLen					= 0;
		
		Str255							pstrRevision		= "\p";
		Str255							pstrTimeStamp		= "\p";
		Str255							pstrActualTime 	= "\p";
		
		CkidResourcePB			ckidInfo;
				
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				intModDate = FSpGetModDate( &inTargetItem );
		
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, pstrTimeStamp, NULL, NULL );
				
						if ( error == noErr )
							{
								if ( CUIsAddedOrRemovedRevision(pstrRevision) )
									bResult = true;
							}
							
						if ( (bResult == false) && (error == noErr) )
							{
								error = FSpReadCkidResource( &inTargetItem, &ckidInfo, false );
													
								if ( error == noErr )
									{															
										error = CUIsFileModified( &ckidInfo, pstrTimeStamp, intModDate );
									}	
								else
									{
										error = CUIsFileModified( NULL, pstrTimeStamp, intModDate );
									}
								
								if ( error == kCVSEngineFileLocallyModifiedErr )
									{
										bResult = true;
										error = noErr;
									}
											
								ThrowIfError_( error );
							}
					}
			}
						
		return bResult;		
	}


void 
CProjectCommitAction::RedoSelf(void)
	{	
	
		const			kAskForEveryFile						= 1;
		const			kAskForOnlyFilesThatNeed		= 2;
		const			kUseCommentForAll						= 3;

		OSErr			result											= noErr;					
								
		Boolean		bDone												= false;
		Boolean		bUserCancelled							= false;
		Boolean		bUserAborted								= false;
												
		FSSpec		fsCurItem										= { 0, 0, "\p" };
	
		UInt32		intReqAction								= kAskForEveryFile;
		UInt32		intSize											= sizeof( FSSpec );

		Handle		hText												= NULL;
		LArray		*pItemsList									= NULL;
		
		CMacCVSCommitTask::CMacCVSCommitTaskItem sTaskItem;
			
		try 
			{
				LStr255 pstrCaption("You are about to commit ");
				LStr255 pstrFileCount("0");
		
				sTaskItem.Init();
				
				pItemsList = new LArray( sizeof(sTaskItem) );
				ThrowIfNil_( pItemsList );
				
				CheckAllItemsQualify( true );

				if ( mAskUser == false )
					{
						LArrayIterator stIterator( mItemList );

						sTaskItem.Init();
						sTaskItem.mOptions.mUseFileRsrcComment = true;
						
						while ( stIterator.Next(&fsCurItem) )
							{
								sTaskItem.mFileSpec = fsCurItem;
								sTaskItem.mOptions.mCommentText = NULL;
		
								pItemsList->AddItem(&sTaskItem, sizeof(sTaskItem));	
							}
					}
				else
					{
						//  If we're committing more than 1 file, ask some options...

						if ( mItemList.GetCount() > 1 ) 
							{				
								StDialogHandler commitDialogHandler( kMacCVSCommitWindowID, mOwnerDoc );
								commitDialogHandler.AddAttachment( new CThreadedTaskYieldAttachment );
								
								LDialogBox *commitDialogBox = dynamic_cast<LDialogBox *>( commitDialogHandler.GetDialog() );
								ThrowIfNil_( commitDialogBox );
								
								LCaption *commitCaption = dynamic_cast<LCaption *>( commitDialogBox->FindPaneByID(kMacCVSCommitCaptionPaneID) );
								ThrowIfNil_( commitCaption );
								
								LTextEditView	*textField = dynamic_cast<LTextEditView *>( commitDialogBox->FindPaneByID(kMacCVSCommitCommentTEPaneID) );	
								ThrowIfNil_( textField );

								Int16 curTextLength = GetHandleSize( textField->GetTextHandle() );
								
								LRadioButton *askForEachButton = dynamic_cast<LRadioButton *>( commitDialogBox->FindPaneByID(kMacCVSCommitAskForEachRadioPaneID) );
								ThrowIfNil_( askForEachButton );

								LRadioButton *askOnlyButton = dynamic_cast<LRadioButton *>( commitDialogBox->FindPaneByID(kMacCVSCommitAskOnlyRadioPaneID) );
								ThrowIfNil_( askOnlyButton );

								LRadioButton *useForAllButton = dynamic_cast<LRadioButton *>( commitDialogBox->FindPaneByID(kMacCVSCommitUseForAllRadioPaneID) );
								ThrowIfNil_( useForAllButton );
								
								UInt32ToStr255( mItemList.GetCount(), pstrFileCount );
								
								pstrCaption = pstrCaption + pstrFileCount + "\p files to the repository";
														
								commitCaption->SetDescriptor( pstrCaption );
								commitDialogBox->SetDefaultButton( kMacCVSCommitOkButtonPaneID );
								commitDialogBox->SetCancelButton( kMacCVSCommitCancelButtonPaneID );
								commitDialogBox->SetLatentSub( textField );
				
								commitDialogBox->Show();
						
								while ( !bDone )
									{								
										try
											{								
												switch ( commitDialogHandler.DoDialog() )
													{ 									
														case kMacCVSCommitOkButtonValueMessage:
															bDone = true;
														break;
						
														case kMacCVSCommitCancelButtonValueMessage:
															bDone = true;
															bUserCancelled = true;
														break;
																										
														default:
															Int16			newTextLength = GetHandleSize(textField->GetTextHandle());
															
															if ( newTextLength != curTextLength && useForAllButton->GetValue() != 1)
																{
																	useForAllButton->SetValue(1);		// radio group turns off others
																	curTextLength = newTextLength;
																}
																
														break;																
													}											
											}
										catch ( ExceptionCode inErr )
											{									
												if ( inErr != err_32kLimit )
													throw inErr;									
											}
									}		// while not done
									
								if ( !bUserCancelled )
									{								
										if ( askForEachButton->GetValue() > 0 )
											{									
												intReqAction = kAskForEveryFile;									
											}
										else if ( askOnlyButton->GetValue() > 0 )
											{									
												intReqAction = kAskForOnlyFilesThatNeed;									
											}
										else if ( useForAllButton->GetValue() > 0 )
											{

		//  If the user chose to use a comment for all, save the handle.
		//  This handle is owned by us.

												intReqAction = kUseCommentForAll;
										
												hText = textField->GetTextHandle();
												
												result = ::HandToHand( &hText );										
												ThrowIfOSErr_( result );										
											}
										else
											{

		//  We shouldn't ever get here.
								
												Assert_( false );										
												intReqAction = kAskForEveryFile;											
											}								
									}	//  if not cancelled
							
							}  //  if > 1 file 
									
						if ( !bUserCancelled )
							{
								bUserCancelled = false;
			
								LArrayIterator	itemIterator( mItemList );
								
								while ( itemIterator.Next(&fsCurItem, intSize) )
									{						
										Assert_( intSize == sizeof(FSSpec) );
										intSize = sizeof( FSSpec );
			
										sTaskItem.mOptions.mCommentText = NULL;
										sTaskItem.mOptions.mUseFileRsrcComment = true;
										sTaskItem.mFileSpec = fsCurItem;
										
										switch ( intReqAction )
											{									
												case kAskForEveryFile:																					
													bUserCancelled = !EditCheckoutComment( &fsCurItem, mOwnerDoc, true, true, bUserAborted );												
												break;
												
												case kAskForOnlyFilesThatNeed:									
													bUserCancelled = !EditCheckoutComment( &fsCurItem, mOwnerDoc, false, true, bUserAborted );									
												break;
												
												case kUseCommentForAll:

		//  Duplicate our comment handle. The duplicated handle is owned by the CMacCVSCommitTask.
		//  If we fail to create the commit task, we will dispose of this handle in our catch clause.

													sTaskItem.mOptions.mUseFileRsrcComment = false;
													sTaskItem.mOptions.mCommentText = hText;
													
													result = ::HandToHand( &sTaskItem.mOptions.mCommentText );
													
													if ( result != noErr )
														{										
															sTaskItem.mOptions.mCommentText = NULL;													
															ThrowOSErr_( result );													
														}												
												break;																			
											}
										
										if ( bUserAborted )
											break;
											
										if ( !bUserCancelled )			
											{
												pItemsList->AddItem(&sTaskItem, sizeof(sTaskItem));	
												
												//  Set it to NULL so we don't double dispose
												sTaskItem.mOptions.mCommentText = NULL;
											}			
									}	//  while()
							}	//  if ( !bUserCancelled )	
					}
					
				// now spawn the threads
				if ( pItemsList->GetCount() > 0 )
					{
  					CMacCVSCommitTask *pTask = new CMacCVSCommitTask( pItemsList, mOwnerDoc );
						ThrowIfNil_( pTask );

						pTask->SetCompletionEvent( mCurrentEvent, mReplyEvent );
						pTask->Resume();
					}

				//  Remove all the items because the data is now
				//  owned by the Commit task.
				pItemsList->RemoveItemsAt( pItemsList->GetCount(), LArray::index_First );
				
				if ( hText != NULL )
					{				
						::DisposeHandle( hText );
						hText = NULL;					
					}
				
				delete pItemsList;
				pItemsList = NULL;
			
			}
		catch ( ExceptionCode inErr )
			{			
				if ( hText != NULL )
					{					
						::DisposeHandle( hText );
						hText = NULL;					
					}
					
				if ( sTaskItem.mOptions.mCommentText != NULL )
					{					
						::DisposeHandle( sTaskItem.mOptions.mCommentText );
						sTaskItem.mOptions.mCommentText = NULL;					
					}
				
				if ( pItemsList != NULL )
					{
						LArrayIterator stIterator( *pItemsList );
						
						while ( stIterator.Next(&sTaskItem) )
							{
								if ( sTaskItem.mOptions.mCommentText != NULL )
									{					
										::DisposeHandle( sTaskItem.mOptions.mCommentText );
										sTaskItem.mOptions.mCommentText = NULL;					
									}						
							}
							
						delete pItemsList;
						pItemsList = NULL;
					}
					
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			}
			
		catch ( ... )
			{
				if ( hText != NULL )
					{					
						::DisposeHandle( hText );
						hText = NULL;					
					}
					
				if ( sTaskItem.mOptions.mCommentText != NULL )
					{					
						::DisposeHandle( sTaskItem.mOptions.mCommentText );
						sTaskItem.mOptions.mCommentText = NULL;					
					}
				
				if ( pItemsList != NULL )
					{
						LArrayIterator stIterator( *pItemsList );
						
						while ( stIterator.Next(&sTaskItem) )
							{
								if ( sTaskItem.mOptions.mCommentText != NULL )
									{					
										::DisposeHandle( sTaskItem.mOptions.mCommentText );
										sTaskItem.mOptions.mCommentText = NULL;					
									}						
							}
							
						delete pItemsList;
						pItemsList = NULL;
					}
	
				throw;
			}
			
	}

#pragma mark -

// 
//		CProjectCompareAction
// 


CProjectCompareAction::CProjectCompareAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
	}


Boolean 
CProjectCompareAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{

		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		SInt32							intModDate			= 0;
		UInt32							intLen					= 0;
		
		Str255							pstrRevision		= "\p";
		Str255							pstrTimeStamp		= "\p";
		Str255							pstrActualTime 	= "\p";
		Str255							pstrOptions			= "\p";
		
		CkidResourcePB			ckidInfo				= { 0 };
				
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				intModDate = FSpGetModDate( &inTargetItem );
		
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, pstrTimeStamp, pstrOptions, NULL );
				
						if ( error == noErr )
							{
								if ( !CUIsAddedOrRemovedRevision(pstrRevision) && !CUIsCVSBinaryOption(pstrOptions) )
									{
										error = FSpReadCkidResource( &inTargetItem, &ckidInfo, false );
								
										if ( error == noErr )
											{															
												error = CUIsFileModified( &ckidInfo, pstrTimeStamp, intModDate );
											}	
										else
											{
												error = CUIsFileModified( NULL, pstrTimeStamp, intModDate );
											}
										
										if ( error == kCVSEngineFileLocallyModifiedErr )
											{
												bResult = true;
												error = noErr;
											}
											
										ThrowIfError_( error );
									}
							}
					}		
			}
			
		return bResult;
		
	}


void 
CProjectCompareAction::RedoSelf(void)
	{
		
		UInt32	ioCount			= sizeof( FSSpec );
		FSSpec	theCurItem	= { 0, 0, "\p" };
		UInt32	delayTicks	= 0;
		
		CCVSEngineDiffOptions	sCompareOptions;
		
		sCompareOptions.Init();
		mOwnerDoc->GetDefaultsForDiffOrCompare( sCompareOptions, true );

		try 
			{
				CheckAllItemsQualify();
				
				LArrayIterator itemIterator( mItemList );
				
				while( itemIterator.Next(&theCurItem, ioCount) )
					{
						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof( FSSpec );
	
						CMacCVSCompareWindow *pCompareWindow = dynamic_cast<CMacCVSCompareWindow *>( LWindow::CreateWindow(CMacCVSCompareWindow::kMacCVSCompareWindowID, mOwnerDoc) );
						ThrowIfNil_( pCompareWindow );
						
						mOwnerDoc->AddToWindowList( pCompareWindow );
						pCompareWindow->SetOwnerDoc( mOwnerDoc );
						pCompareWindow->SetWindowPosition();
						
						pCompareWindow->Compare( &theCurItem, mOwnerDoc, (const void *)&sCompareOptions, delayTicks );
						delayTicks += kThreadSpawnIntervalTicks;
					}				
			}
		catch ( ExceptionCode inErr )
			{
				if ( inErr != CProjectAction::kProjectActionCancelled )
					{
						throw( inErr );
					}
			}
							
	}

#pragma mark -

// 
//		CProjectDiffAction
// 


CProjectDiffAction::CProjectDiffAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, Boolean inShowAdvancedDialog ):

	CProjectAction( inOwnerDoc, inItemList ),
	mShowAdvancedDialog(inShowAdvancedDialog)
	
	{
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
	}


Boolean 
CProjectDiffAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{

		OSErr								error						= noErr;
		Boolean							bResult					= false;
		Boolean							bHasCkid				= false;
		
		FSSpec							fsParent				= { 0, 0, "\p" };
		SInt32							intModDate			= 0;
		UInt32							intLen					= 0;
		
		Str255							pstrRevision		= "\p";
		Str255							pstrTimeStamp		= "\p";
		Str255							pstrActualTime 	= "\p";
		
		CkidResourcePB			ckidInfo;
				
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				intModDate = FSpGetModDate( &inTargetItem );
		
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, pstrTimeStamp, NULL, NULL );
				
						if ( error == noErr )
							{				
								if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
									{
										error = FSpReadCkidResource( &inTargetItem, &ckidInfo, false );
								
										if ( error == noErr )
											{															
												error = CUIsFileModified( &ckidInfo, pstrTimeStamp, intModDate );
											}	
										else
											{
												error = CUIsFileModified( NULL, pstrTimeStamp, intModDate );
											}
										
										if ( error == kCVSEngineFileLocallyModifiedErr )
											{
												bResult = true;
												error = noErr;
											}
											
										ThrowIfError_( error );
									}
							}
					}	
			}
			
		return bResult;

	}

// static
Boolean
CProjectDiffAction::GetDiffOptions( CMacCVSDoc *inOwnerDoc, CCVSEngineDiffOptions &ioDiffOptions, Boolean inUseDefaults, Boolean inLockRevisions )
	{
			static 	CCVSEngineDiffOptions		sLocalOptions;
			static 	Boolean									sFirstTime = true;

			if ( sFirstTime ) 	
				{
					sLocalOptions.Init();
					sFirstTime = false;
				}

			CCVSEngineDiffOptions					tempDiffOptions;

			if ( inUseDefaults )
				tempDiffOptions = sLocalOptions;
			else
				tempDiffOptions = ioDiffOptions;

			StDialogHandler			diffOptionsDialog(kDiffOptionsDialogID, nil);

			LDialogBox 		*dialogBox = dynamic_cast<LDialogBox *>( diffOptionsDialog.GetDialog() );	
			ThrowIfNil_( dialogBox );

			// set up the whitespace radio group
			LRadioGroupView		*whiteSpaceGroup = dynamic_cast<LRadioGroupView *>( dialogBox->FindPaneByID(kDiffOptionsWhiteSpaceRadioGroupID) );	
			ThrowIfNil_( whiteSpaceGroup );
			
			PaneIDT		theCurRadio;
			
			// I don't know if these should really be mutually exclusive, but it makes the
			// dialog much tidier. I do know that it's OK to send more than one to the server.
			if ( tempDiffOptions.mIgnoreWhiteSpace )
				theCurRadio = kDiffOptionsIgnoreWhiteSpaceRadio;
			else if ( tempDiffOptions.mIgnoreSpaceChange )
				theCurRadio = kDiffOptionsIgnoreSpaceChangeRadio;
			else if ( tempDiffOptions.mIgnoreBlankLines )
				theCurRadio = kDiffOptionsIgnoreBlankLinesRadio;
			else
				theCurRadio = kDiffOptionsDiffWhiteSpaceRadio;
				
			whiteSpaceGroup->SetCurrentRadioID( theCurRadio );
			
			// set up the other options
			LCheckBox		*ignoreCaseCheckbox = dynamic_cast<LCheckBox *>( dialogBox->FindPaneByID(kDiffOptionsIgnoreCaseCheckboxID) );
			ThrowIfNil_( ignoreCaseCheckbox );
			ignoreCaseCheckbox->SetValue( (Int32)tempDiffOptions.mIgnoreCase );

			LCheckBox		*showFunctionCheckbox = dynamic_cast<LCheckBox *>( dialogBox->FindPaneByID(kDiffOptionsShowFunctionCheckboxID) );
			ThrowIfNil_( showFunctionCheckbox );
			showFunctionCheckbox->SetValue( (Int32)tempDiffOptions.mShowCFunction );
			
			LCheckBox		*useForCompareCheckbox = dynamic_cast<LCheckBox *>( dialogBox->FindPaneByID(kDiffOptionsUseForCompareCheckboxID) );
			ThrowIfNil_( useForCompareCheckbox );
			useForCompareCheckbox->SetValue( 0 );			// initialize to off

			LCheckBox		*useForDiffCheckbox = dynamic_cast<LCheckBox *>( dialogBox->FindPaneByID(kDiffOptionsUseForDiffCheckboxID) );
			ThrowIfNil_( useForDiffCheckbox );
			useForDiffCheckbox->SetValue( 0 );			// initialize to off

			// set up the context diff options
			LRadioButton		*regularDiffRadio = dynamic_cast<LRadioButton *>( dialogBox->FindPaneByID(kDiffOptionsRegularDiffRadioID) );
			ThrowIfNil_( regularDiffRadio );
			regularDiffRadio->SetValue( 0 );

			LRadioButton		*rcsDiffRadio = dynamic_cast<LRadioButton *>( dialogBox->FindPaneByID(kDiffOptionsRCSDiffRadioID) );
			ThrowIfNil_( rcsDiffRadio );
			rcsDiffRadio->SetValue( (Int32)tempDiffOptions.mRCSDiff );

			LRadioButton		*contextDiffRadio = dynamic_cast<LRadioButton *>( dialogBox->FindPaneByID(kDiffOptionsContextDiffRadioID) );
			ThrowIfNil_( contextDiffRadio );
			contextDiffRadio->SetValue( 0 );
			
			LEditText		*contextDiffLines = dynamic_cast<LEditText *>( dialogBox->FindPaneByID(kDiffOptionsContextLinesEditTextID) );
			ThrowIfNil_( contextDiffLines );
			contextDiffLines->SetValue( tempDiffOptions.mContextLines );
							
			// set up the unified diff options
			LRadioButton		*unifDiffRadio = dynamic_cast<LRadioButton *>( dialogBox->FindPaneByID(kDiffOptionsUnifiedDiffRadioID) );
			ThrowIfNil_( unifDiffRadio );
			unifDiffRadio->SetValue( 0 );
			
			// set up the radio buttons
			if ( tempDiffOptions.mRCSDiff )
				rcsDiffRadio->SetValue( 1 );
			else if ( tempDiffOptions.mContextDiff )
				contextDiffRadio->SetValue( 1 );
			else if ( tempDiffOptions.mUnifiedDiff )
				unifDiffRadio->SetValue( 1 );
			else
				regularDiffRadio->SetValue( 1 );
				
			// disable the text fields
			if ( !tempDiffOptions.mContextDiff && !tempDiffOptions.mUnifiedDiff)
				contextDiffLines->Disable();

			// set up the revisions info
			LCheckBoxGroupBox		*revisionsGroup = dynamic_cast<LCheckBoxGroupBox *>( dialogBox->FindPaneByID(kDiffOptionsUseRevsCheckboxGroupID) );
			ThrowIfNil_( revisionsGroup );
			revisionsGroup->SetValue( tempDiffOptions.mUseRevisions);
			
			LEditText		*fromRevText = dynamic_cast<LEditText *>( dialogBox->FindPaneByID(kDiffOptionsFromRevEditTextID) );
			LEditText		*toRevText = dynamic_cast<LEditText *>( dialogBox->FindPaneByID(kDiffOptionsToRevEditTextID) );
			ThrowIfNil_( fromRevText );
			ThrowIfNil_( toRevText );
			fromRevText->SetDescriptor(tempDiffOptions.mOrigRevisionPStr);
			toRevText->SetDescriptor(tempDiffOptions.mDiffRevisionPStr);
			
			if (inLockRevisions)
				{
					revisionsGroup->Disable();
				}
				
			LCaption		*commandLine = dynamic_cast<LCaption *>( dialogBox->FindPaneByID(kDiffOptionsCommandLineLabel) );
			ThrowIfNil_( commandLine );

			Str255			commandLineLabel;
			CUMakeDiffCommandLineOptionsString( &tempDiffOptions, commandLineLabel, true);
			commandLine->SetDescriptor(commandLineLabel);
			
			dialogBox->SetDefaultButton( kOpenFilesListOKButtonID );
			dialogBox->SetCancelButton( kOpenFilesListCancelButtonID );

			dialogBox->Show();
		
			Boolean done = false;
			Boolean	confirmed = false;
			
			while ( !done )
				{
					MessageT			theMessage = diffOptionsDialog.DoDialog();
					
					if ( theMessage != cmd_Nothing )
						{
							tempDiffOptions.Init();
							
							switch ( whiteSpaceGroup->GetCurrentRadioID() )
								{
									case kDiffOptionsDiffWhiteSpaceRadio:
										// nothing
										break;
										
									case kDiffOptionsIgnoreWhiteSpaceRadio:
										tempDiffOptions.mIgnoreWhiteSpace = true;
										break;
									
									case kDiffOptionsIgnoreSpaceChangeRadio:
										tempDiffOptions.mIgnoreSpaceChange = true;
										break;
										
									case kDiffOptionsIgnoreBlankLinesRadio:
										tempDiffOptions.mIgnoreBlankLines = true;
										break;
								
									default:
										Assert_(false);
								
								}
						
							tempDiffOptions.mIgnoreCase 		= ignoreCaseCheckbox->GetValue() == Button_On;
							tempDiffOptions.mShowCFunction	= showFunctionCheckbox->GetValue() == Button_On;

							tempDiffOptions.mRCSDiff		 		= rcsDiffRadio->GetValue() == Button_On;
							tempDiffOptions.mContextDiff		= contextDiffRadio->GetValue() == Button_On;
							tempDiffOptions.mUnifiedDiff		= unifDiffRadio->GetValue() == Button_On;
							tempDiffOptions.mContextLines		= (Int16)contextDiffLines->GetValue();

							tempDiffOptions.mUseRevisions		= revisionsGroup->GetValue() == Button_On;
							fromRevText->GetDescriptor(tempDiffOptions.mOrigRevisionPStr);
							toRevText->GetDescriptor(tempDiffOptions.mDiffRevisionPStr);
							
							CUMakeDiffCommandLineOptionsString( &tempDiffOptions, commandLineLabel, true );
							commandLine->SetDescriptor(commandLineLabel);
							
							// Ensure that 'Use for compare checkbox can only be on for valid options
							if (	regularDiffRadio->GetValue() == Button_On &&
										revisionsGroup->GetValue() == Button_Off &&
										showFunctionCheckbox->GetValue() == Button_Off)
								{
									useForCompareCheckbox->Enable();
								}
							else
								{
									useForCompareCheckbox->Disable();
								}
							
							// Only allow default diff if not using revisions
							if (	revisionsGroup->GetValue() == Button_Off )
								{
									useForDiffCheckbox->Enable();
								}
							else
								{
									useForDiffCheckbox->Disable();
								}
							
						}
					
					switch ( theMessage )
						{ 
						
							case msg_OK:
							
								ioDiffOptions = tempDiffOptions;

								if ( inUseDefaults )
									sLocalOptions = tempDiffOptions;
								
								if ( useForCompareCheckbox->IsEnabled() && useForCompareCheckbox->GetValue() == Button_On )
									{
									
										inOwnerDoc->SetDefaultsForDiffOrCompare(tempDiffOptions, true);
									
									}
								
								if ( useForDiffCheckbox->IsEnabled() && useForDiffCheckbox->GetValue() == Button_On )
									{
									
										inOwnerDoc->SetDefaultsForDiffOrCompare(tempDiffOptions, false);
									
									}
								
								done = true;
								confirmed = true;
								break;
							
							case msg_Cancel:

								if ( inUseDefaults )
									sLocalOptions = tempDiffOptions;

								done = true;
								break;
						
							case kDiffOptionsRegularDiffRadioID:
							case kDiffOptionsContextDiffRadioID:
							case kDiffOptionsUnifiedDiffRadioID:

								if ( contextDiffRadio->GetValue() == Button_On || unifDiffRadio->GetValue() == Button_On)
									contextDiffLines->Enable();
								else
									contextDiffLines->Disable();
														
								break;
							
						}
						
				}

		return confirmed;
	}

void 
CProjectDiffAction::RedoSelf(void)
	{
		
		UInt32	ioCount			= sizeof( FSSpec );
		FSSpec	theCurItem	= { 0, 0, "\p" };
		UInt32	delayTicks	= 0;
		
		CCVSEngineDiffOptions			diffOptions;
		
		diffOptions.Init();
		mOwnerDoc->GetDefaultsForDiffOrCompare(diffOptions, false);
		
		try 
			{

				CheckAllItemsQualify();
				
				if ( mShowAdvancedDialog )
					{
						if ( ! GetDiffOptions( mOwnerDoc, diffOptions ) )
							throw (ExceptionCode)(CProjectAction::kProjectActionCancelled);
					}
					
				LArrayIterator	itemIterator( mItemList );
				
				while( itemIterator.Next(&theCurItem, ioCount) )
					{

						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof( FSSpec );
	
						CMacCVSStatusWindow	*statusWindow = dynamic_cast<CMacCVSStatusWindow *>( LWindow::CreateWindow(CMacCVSStatusWindow::kMacCVSStatusWindowID, mOwnerDoc) );
						ThrowIfNil_( statusWindow );
						
						mOwnerDoc->AddToWindowList( statusWindow );
						statusWindow->SetOwnerDoc(mOwnerDoc);
						statusWindow->SetWindowPosition();

						statusWindow->Status( &theCurItem, mOwnerDoc, true, &diffOptions, delayTicks );
						delayTicks += kThreadSpawnIntervalTicks;
					}
					
			}
			
		catch ( ExceptionCode inErr )
			{
			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			
			}
							
	}


	
#pragma mark -


// 
//		CProjectEditAction
// 


CProjectEditAction::
CProjectEditAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean	inEdit ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		mEdit = inEdit;		
	}


Boolean 
CProjectEditAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		SInt32							intModDate			= 0;
		UInt32							intLen					= 0;
		
		Str255							pstrRevision		= "\p";
		Str255							pstrTimeStamp		= "\p";
		Str255							pstrActualTime 	= "\p";
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				intModDate = FSpGetModDate( &inTargetItem );
		
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, pstrTimeStamp, NULL, NULL );
				
						if ( error == noErr )
							{
								if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
									{
										bResult = true;
									}
							}	
					}	
			}

		return bResult;
		
	}


void 
CProjectEditAction::RedoSelf(void)
	{
		
		OSErr			error						= noErr;
		Boolean		bUserCancelled 	= false;
				
		try 
			{		
				if ( !mEdit )
					{
						bUserCancelled = WarnUser();
					}
					
				if ( !bUserCancelled )
					{										
						CheckAllItemsQualify( true );

						if ( mItemList.GetCount() > 0 )
							{
								CMacCVSEditTask *pTask = new CMacCVSEditTask( &mItemList, mOwnerDoc, mEdit );
								ThrowIfNil_( pTask );
								
								pTask->Resume();
							}
					}				
			}			
		catch ( ExceptionCode inErr )
			{
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );		
			}		
		catch ( ... )
			{
				throw;
			}
				
	}


Boolean
CProjectEditAction::WarnUser( )
	{

		Boolean 	bUserCancelled 			= false;
		Boolean		bDone								= false;
		MessageT	intItemHit					= 0;
		Str255		pstrTruncated				= "\p";
		
		StDialogHandler	stHandler( kRevertDialogPPobID, mOwnerDoc );
		stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

		LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
		ThrowIfNil_( pDialog );
	
		LCaption *pMessage = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kRevertDialogMessageFieldID) );
		ThrowIfNil_( pMessage );

		LPushButton *pOKButton = dynamic_cast<LPushButton *>( pDialog->FindPaneByID(kRevertDialogOKButtonID) );
		ThrowIfNil_( pOKButton );
		
		pOKButton->SetDescriptor( "\pUn-Edit" );
		
		LStr255	pstrWarning;
		pstrWarning.Append( "\pAre you sure you want to un-edit these files? Your changes will be preserved in a backup file." );
		
		pMessage->SetDescriptor( pstrWarning );		
		
		pDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
		pDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
		
		::SysBeep( 10 );
		pDialog->Show();
									
		while ( !bDone )
			{		
				intItemHit = stHandler.DoDialog();
				
				switch ( intItemHit )
					{			
						case msg_OK:					
							bDone = true;							
						break;
						
						case msg_Cancel:						
							bDone = true;
							bUserCancelled = true;
						break;					
					}				
			}

		return bUserCancelled;										
	
	}
	

#pragma mark -

// 
//		CProjectGetFileInfoAction
// 


CProjectGetFileInfoAction::
CProjectGetFileInfoAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
	}


Boolean 
CProjectGetFileInfoAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );
				
				if ( FSpIsRepositoryDir(&fsParent) )
					{
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, NULL, NULL, NULL, NULL );
				
						if ( error == noErr )
							bResult = true;	
					}																		
			}
			
		return bResult;
		
	}


void 
CProjectGetFileInfoAction::RedoSelf(void)
	{
		
		UInt32	ioCount			= sizeof( FSSpec );
		FSSpec	theCurItem	= { 0, 0, "\p" };
		UInt32	delayTicks 	= 0;
		
		try 
			{

				CheckAllItemsQualify();
				
				LArrayIterator	itemIterator( mItemList );
				
				while( itemIterator.Next(&theCurItem, ioCount) )
					{
				
						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof( FSSpec );
						
						CMacCVSLogWindow	*logWindow = dynamic_cast<CMacCVSLogWindow *>( LWindow::CreateWindow(CMacCVSLogWindow::kMacCVSLogWindowID, mOwnerDoc) );
						
						ThrowIfNil_( logWindow );
						
						mOwnerDoc->AddToWindowList( logWindow );
						logWindow->SetOwnerDoc(mOwnerDoc);
						logWindow->SetWindowPosition();
						
						// delayTicks is nasty hack to try to reduce the effects of server directory locking
						logWindow->Log( &theCurItem, mOwnerDoc, delayTicks );
						delayTicks += kThreadSpawnIntervalTicks;
					}
					
			}
			
		catch ( ExceptionCode inErr )
			{
			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			
			}
			
	}

#pragma mark -

// 
//		CProjectGetStatusInfoAction
// 


CProjectGetFileStatusAction::
CProjectGetFileStatusAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
	}


Boolean 
CProjectGetFileStatusAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, NULL, NULL, NULL, NULL );
				
						if ( error == noErr )
							bResult = true;		
					}																	
			}
			
		return bResult;
		
	}


void 
CProjectGetFileStatusAction::RedoSelf(void)
	{
		
		UInt32	ioCount			= sizeof( FSSpec );
		FSSpec	theCurItem	= { 0, 0, "\p" };
		UInt32	delayTicks	= 0;
		
		try 
			{

				CheckAllItemsQualify();
				
				LArrayIterator	itemIterator( mItemList );
				
				while( itemIterator.Next(&theCurItem, ioCount) )
					{

						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof( FSSpec );
		
						CMacCVSStatusWindow	*statusWindow = dynamic_cast<CMacCVSStatusWindow *>( LWindow::CreateWindow(CMacCVSStatusWindow::kMacCVSStatusWindowID, mOwnerDoc) );
						ThrowIfNil_( statusWindow );
						
						mOwnerDoc->AddToWindowList( statusWindow );
						statusWindow->SetOwnerDoc(mOwnerDoc);
						statusWindow->SetWindowPosition();

						statusWindow->Status( &theCurItem, mOwnerDoc, false, NULL, delayTicks );
						delayTicks += kThreadSpawnIntervalTicks;
					}
					
			}
			
		catch ( ExceptionCode inErr )
			{
			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			
			}
							
	}


#pragma mark -


// 
//		CProjectShowEditorsAction
// 


CProjectShowEditorsWatchersAction::
CProjectShowEditorsWatchersAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean inShowEditors ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{		
		bShowEditors = inShowEditors;
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );		
	}


Boolean 
CProjectShowEditorsWatchersAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr			error			= noErr;
		Boolean		bResult		= false;
	
		FSSpec		fsParent	= { 0, 0, "\p" };
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, NULL, NULL, NULL, NULL );
				
						if ( error == noErr )
							bResult = true;		
					}																	
			}
		else if ( FSpIsRepositoryDir(&inTargetItem) )
			{
				bResult = true;
			}
			
		return bResult;
		
	}


void 
CProjectShowEditorsWatchersAction::RedoSelf( void )
	{
		
		UInt32	ioCount			= sizeof( FSSpec );
		FSSpec	theCurItem	= { 0, 0, "\p" };
		UInt32	delayTicks	= 0;
		
		try 
			{
				CheckAllItemsQualify();
				
				LArrayIterator itemIterator( mItemList );
				
				while( itemIterator.Next(&theCurItem, ioCount) )
					{

						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof( FSSpec );
		
						CMacCVSStatusWindow	*pOutputWindow = dynamic_cast<CMacCVSStatusWindow *>( LWindow::CreateWindow(CMacCVSStatusWindow::kMacCVSStatusWindowID, mOwnerDoc) );
						ThrowIfNil_( pOutputWindow );
						
						mOwnerDoc->AddToWindowList( pOutputWindow );
						pOutputWindow->SetOwnerDoc( mOwnerDoc );
						pOutputWindow->SetWindowPosition();

						if ( bShowEditors )
							{
								pOutputWindow->ShowEditors( &theCurItem, mOwnerDoc, delayTicks );
							}
						else
							{
								pOutputWindow->ShowWatchers( &theCurItem, mOwnerDoc, delayTicks );
							}
							
						delayTicks += kThreadSpawnIntervalTicks;
					}				
			}			
		catch ( ExceptionCode inErr )
			{			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					{
						throw inErr;	
					}		
			}
							
	}


#pragma mark -


// 
//		CProjectImportAction
// 


CProjectImportAction::
CProjectImportAction( CMacCVSDoc *inOwnerDoc ):

	CProjectAction( inOwnerDoc )	
	{				
	}


Boolean 
CProjectImportAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{							
		return true;	
	}


void 
CProjectImportAction::RedoSelf(void)
	{
	
		Boolean		bUserCancelled 	= false;
		Boolean		bDone						= false;
		MessageT	intItemHit			= 0;
		
		CMacCVSImportTask::CMacCVSImportTaskItem sTaskItem;
		
		try
			{
				sTaskItem.Init();

				LArray itemList( sizeof(sTaskItem) );

				bUserCancelled = GetImportOptions( sTaskItem.mFileSpec, sTaskItem.mOptions.fRepositoryPStr, sTaskItem.mOptions.fVendorTagPStr, sTaskItem.mOptions.fReleaseTagPStr, sTaskItem.mOptions.fBranchTagPStr );
				
				if ( !bUserCancelled )
					{
						StDialogHandler	stHandler( kAskOKCancelDialogPPobID, mOwnerDoc );
						stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

						LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
						ThrowIfNil_( pDialog );
					
						LCaption *pMessage = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kAskOKCancelDialogCaptionID) );
						ThrowIfNil_( pMessage );

						LStr255	pstrWarning;

						pstrWarning.Append( "\pAre you absolutely sure you want to import the directory " );
						pstrWarning.Append( sTaskItem.mFileSpec.name );
						pstrWarning.Append( "\p to the repository? This is your last warning! Don't do this if you don't know what you're doing!" );
						
						pMessage->SetDescriptor( pstrWarning );
						pDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
						pDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
										
						::SysBeep( 10 );
						
						pDialog->Show();
											
						while ( !bDone )
							{			
								intItemHit = stHandler.DoDialog();
							
								switch ( intItemHit )
									{			
										case msg_OK:					
											bDone = true;						
										break;
										
										case msg_Cancel:						
											bDone = true;
											bUserCancelled = true;						
										break;					
									}					
							}
					}
				
				if ( !bUserCancelled )
					{		
						itemList.InsertItemsAt( 1, LArray::index_First, &sTaskItem, sizeof(sTaskItem) );
						
						CMacCVSImportTask *pTask = new CMacCVSImportTask( &itemList, mOwnerDoc );
						ThrowIfNil_( pTask );
						
						pTask->Resume();
					}					
			}
			
		catch ( ... )
			{
				throw;		
			}
				
	}


Boolean
CProjectImportAction::GetImportOptions( FSSpec &fsOutSpec, Str255 pstrOutRepository, Str255 pstrOutVendorTag, Str255 pstrOutReleaseTag, Str255 pstrOutBranch )
	{
	
		Boolean		bUserCancelled			= false;
		Boolean		bGotValidDirectory 	= false;
		FSSpec		fsCurItem						= { 0, 0, "\p" };
		StandardFileReply	theReply;

		{
			Boolean	bDone				= false;
			UInt32	intItemHit 	= 0;
									
			StDialogHandler	stWarning( kAskOKCancelDialogPPobID, mOwnerDoc );
			stWarning.AddAttachment( new CThreadedTaskYieldAttachment );
					
			LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stWarning.GetDialog() );
			ThrowIfNil_( pDialog );
						
			LCaption *pMessage = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kAskOKCancelDialogCaptionID) );
			ThrowIfNil_( pMessage );

			LStr255	pstrWarning;
												
			pstrWarning.Append( "\pAre you sure you want to do a CVS import?" );
			pstrWarning.Append( "\p Improper use of this feature could destroy data in the repository!" );
							
			pMessage->SetDescriptor( pstrWarning );
			pDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
			pDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
							
			::SysBeep( 10 );
							
			pDialog->Show();
														
			while ( !bDone )
				{				
					intItemHit = stWarning.DoDialog();
					
					switch ( intItemHit )
						{				
							case msg_OK:						
								bDone = true;								
							break;
							
							case msg_Cancel:							
								bDone = true;
								bUserCancelled = true;
							break;						
						}						
				}
		}
		
		if ( bUserCancelled == false )
			{
				// JWW - Use Navigation Services if possible
//				UStandardFile::GetDirectory( theReply );
				SInt32 theDirID;
				theReply.sfGood = UConditionalDialogs::AskChooseFolder(theReply.sfFile, theDirID);
						
				if ( theReply.sfGood )
					{			
						fsCurItem = theReply.sfFile;
						
						if ( FSpIsRepositoryDir(&fsCurItem) )
							{
								UErrorDialog theError( kCantImportCVSDirectoryErr, mOwnerDoc );
							}
						else if ( fsCurItem.parID == fsRtParID )
							{
								UErrorDialog theError( kCantImportRootDirErr, mOwnerDoc );
							}
						else 
							{
								bGotValidDirectory = true;
							}
					}
			}
		
		if ( !bGotValidDirectory )
			{
				bUserCancelled = true;
			}
		else
			{
				Boolean 	bDone 					= false;
				UInt32		intLen					= 0;
				UInt32		intItemHit			= 0;
				Str255		pstrFolderPath	= "\p";
										
				StDialogHandler	stHandler( kImportDialogPPobID, mOwnerDoc );
				stHandler.AddAttachment( new CThreadedTaskYieldAttachment );
				
				LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
				ThrowIfNil_( pDialog );

				LCaption *pPathCaption = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kImportPathCaptionID) );
				ThrowIfNil_( pPathCaption );
				
				LEditText *pRepository = dynamic_cast<LEditText *>( pDialog->FindPaneByID(kImportRepositoryEditFieldID) );
				ThrowIfNil_( pRepository );

				LEditText *pVendor = dynamic_cast<LEditText *>( pDialog->FindPaneByID(kImportVendorTagEditFieldID) );
				ThrowIfNil_( pVendor );

				LEditText *pRelease = dynamic_cast<LEditText *>( pDialog->FindPaneByID(kImportReleaseTagEditFieldID) );
				ThrowIfNil_( pRelease );

				LEditText *pBranch = dynamic_cast<LEditText *>( pDialog->FindPaneByID(kImportBranchTagEditFieldID) );
				ThrowIfNil_( pBranch );

				intLen = 255;		
				ThrowIfOSErr_( FSpGetFullPathStr( &fsCurItem, &intLen, (char *)&pstrFolderPath[1] ) );
				pstrFolderPath[0] = intLen;
						
//  Truncate the string if we need to.
		
				Rect sPathCaptionRect = { 0, 0, 0, 0 };
				pPathCaption->CalcLocalFrameRect( sPathCaptionRect );
								
				StTextState	theTextState;
				UTextTraits::SetPortTextTraits( kImportPathCaptionTextTraitsID );
						
				::TruncString( sPathCaptionRect.right - sPathCaptionRect.left - 2, pstrFolderPath, truncMiddle );
				pPathCaption->SetDescriptor( pstrFolderPath );
												
				pDialog->Show();
								
				while ( !bDone )
					{						
						intItemHit = stHandler.DoDialog();
						
						switch ( intItemHit )
							{								
								case msg_OK:									
																		
									pRepository->GetDescriptor( pstrOutRepository );
									pVendor->GetDescriptor( pstrOutVendorTag );
									pRelease->GetDescriptor( pstrOutReleaseTag );
									pBranch->GetDescriptor( pstrOutBranch );

									if ( pstrOutRepository[0] == 0 || pstrOutVendorTag[0] == 0 || pstrOutReleaseTag[0] == 0 )
										{
											UErrorDialog theError( kImportFieldsRequiredErr, mOwnerDoc );
										}
									else
										{
											bUserCancelled = false;
											bDone = true;
										}
																																		
								break;
								
								case msg_Cancel:
									
									bUserCancelled = true;
									bDone = true;
											
								break;									
							}						
					}											
			}
		
		if ( !bUserCancelled )
			{
				fsOutSpec = fsCurItem;
			}
				
		return bUserCancelled;
		
	}
				
																											
#pragma mark -


// 
//		CProjectRemoveAction
// 


CProjectRemoveAction::
CProjectRemoveAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
				
	}


Boolean 
CProjectRemoveAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		Str255							pstrRevision		= "\p";
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );
				
				if ( FSpIsRepositoryDir(&fsParent) )
					{
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, NULL, NULL, NULL );
				
						if ( error == noErr )
							if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
								bResult = true;	
					}																		
			}
		else if ( error == kCVSEngineIsFolder )
			{			
				bResult = false;			
			}
			
		return bResult;
		
	}


void 
CProjectRemoveAction::RedoSelf(void)
	{
	
		OSErr				result							= noErr;					
		Boolean			bUserCancelled			= false;
														
		try 
			{			
				CheckAllItemsQualify( true );

//  Better ask the user before we do the deed...

				if ( mItemList.GetCount() > 0 ) 
					{
						bUserCancelled = WarnUser();
						
						if ( !bUserCancelled )
							{
								CMacCVSRemoveTask *pTask = new CMacCVSRemoveTask( &mItemList, mOwnerDoc );	
								ThrowIfNil_( pTask );
								
								pTask->Resume();	
							}						
					}										
			}	
		catch ( ExceptionCode inErr )
			{							
				if ( inErr != CProjectAction::kProjectActionCancelled )
					{					
						throw( inErr );					
					}	
			}
		
	}


Boolean
CProjectRemoveAction::WarnUser()
	{

		Boolean			bDone								= false;
		Boolean			bUserCancelled			= false;
		UInt32			intItemHit					= 0;
				
		StDialogHandler	stHandler( kAskOKCancelDialogPPobID, mOwnerDoc );
		stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

		LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
		ThrowIfNil_( pDialog );
	
		LCaption *pMessage = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kAskOKCancelDialogCaptionID) );
		ThrowIfNil_( pMessage );
						
		LStr255	pstrWarning;

		pstrWarning.Append( "\pAre you sure you want to remove these files from the repository?" );
		pstrWarning.Append( "\p This could have serious consequences, so youd better know what you are doing." );
					
		pMessage->SetDescriptor( pstrWarning );
		pDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
		pDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
		
		::SysBeep( 10 );
		
		pDialog->Show();
									
		while ( !bDone )
			{		
				intItemHit = stHandler.DoDialog();
				
				switch ( intItemHit )
					{			
						case msg_OK:						
							bDone = true;							
						break;
						
						case msg_Cancel:
						
							bDone = true;
							bUserCancelled = true;

						break;						
					}					
			}

		pDialog->Hide();
		
		return bUserCancelled;
		
	}
	
	
#pragma mark -


// 
//		CProjectRevertToAction
// 


CProjectRevertToAction::
CProjectRevertToAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean	inAskUser ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		Assert_( inAskUser == false );

/*
**  We have disabled the "Revert To" option because you can run
**  into weird problems with files being a different version
**  than their tag revision.
*/
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		mAskUser = false;
		
	}


Boolean 
CProjectRevertToAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		SInt32							intModDate			= 0;
		UInt32							intLen					= 0;
		
		Str255							pstrRevision		= "\p";
		Str255							pstrTimeStamp		= "\p";
		Str255							pstrActualTime 	= "\p";
		
		CkidResourcePB			ckidInfo;
				
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				intModDate = FSpGetModDate( &inTargetItem );
		
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, pstrTimeStamp, NULL, NULL );
				
						if ( error == noErr )
							{
								if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
									{
										error = FSpReadCkidResource( &inTargetItem, &ckidInfo, false );
								
										if ( error == noErr )
											{															
												error = CUIsFileModified( &ckidInfo, pstrTimeStamp, intModDate );
											}	
										else
											{
												error = CUIsFileModified( NULL, pstrTimeStamp, intModDate );
											}
										
										if ( error == kCVSEngineFileLocallyModifiedErr )
											{
												bResult = true;
												error = noErr;
											}
											
										ThrowIfError_( error );
									}
							}	
					}	
			}

		return bResult;
		
	}


void 
CProjectRevertToAction::RedoSelf(void)
	{
		
		OSErr			error						= noErr;
		Boolean		bUserCancelled 	= false;
		UInt32		intMaxIndex			= 0;
		FSSpec		fsCurItem				= {	0, 0, "\p" };
		FSSpec		fsParent				= { 0, 0, "\p" };
	
		LArray		*pItemsList			= NULL;
		
		CMacCVSUpdateTask::CMacCVSUpdateTaskItem sTaskItem;
		
		try 
			{		
				bUserCancelled = WarnUser();
				
				if ( !bUserCancelled )
					{		
						pItemsList = new LArray( sizeof(sTaskItem) );
						ThrowIfNil_( pItemsList );
								
						CheckAllItemsQualify( true );

						LArrayIterator	stItemIterator( mItemList );

						while ( stItemIterator.Next(&fsCurItem) )
							{
								sTaskItem.Init();
							
								sTaskItem.mFileSpec = fsCurItem;
								sTaskItem.mOptions.mRevert = true;

								error = FSpGetParentFSSpec( &fsCurItem, &fsParent );			
								ThrowIfError_( error );
												
								if ( !mEntries.AreYouThisDirectory(&fsParent) )
									{
										error = mEntries.CloseEntriesFile();
										ThrowIfError_( error );

										error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
										ThrowIfError_( error );
									}
									
								error = mEntries.GetEntry( fsCurItem.name, sTaskItem.mOptions.mRevisionPStr, NULL, NULL, NULL );
								ThrowIfOSErr_( error );
																
								pItemsList->InsertItemsAt( 1, LArray::index_First, &sTaskItem, sizeof(sTaskItem) );
							}	
							
						if ( pItemsList->GetCount() > 0 )
							{
								CMacCVSUpdateTask *pTask = new CMacCVSUpdateTask( pItemsList, mOwnerDoc );
								ThrowIfNil_( pTask );
								
								pTask->Resume();
							}
							
						if ( pItemsList != NULL )
							{
								delete pItemsList;
								pItemsList = NULL;
							}	
					}				
			}			
		catch ( ExceptionCode inErr )
			{
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}					

				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );		
			}		
		catch ( ... )
			{
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}					
				throw;
			}
				
	}


Boolean
CProjectRevertToAction::WarnUser( )
	{

		Boolean 	bUserCancelled 			= false;
		Boolean		bDone								= false;
		MessageT	intItemHit					= 0;
		Str255		pstrTruncated				= "\p";
		
		StDialogHandler	stHandler( kRevertDialogPPobID, mOwnerDoc );
		stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

		LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
		ThrowIfNil_( pDialog );
	
		LCaption *pMessage = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kRevertDialogMessageFieldID) );
		ThrowIfNil_( pMessage );

		LStr255	pstrWarning;
/*				
		PStrCopy( pstrInName, pstrWarning, sizeof(pstrWarning) );
				
		if ( pstrWarning[0] > 39 )
			{		
				pstrWarning[0] = 40;
				pstrWarning[pstrWarning[0]] = '';			
			}
*/			
		pstrWarning.Append( "\pAre you sure you want to rollback these files? Your changes will be preserved in a backup file." );

//		pstrWarning.Append( pstrTruncated );
//		pstrWarning.Append( "\p?" );
		
		pMessage->SetDescriptor( pstrWarning );
		
		//Kathy wants the defaults reversed
		
		//pDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
		//pDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
		
		::SysBeep( 10 );
		pDialog->Show();
									
		while ( !bDone )
			{		
				intItemHit = stHandler.DoDialog();
				
				switch ( intItemHit )
					{			
						case msg_OK:					
							bDone = true;							
						break;
						
						case msg_Cancel:						
							bDone = true;
							bUserCancelled = true;
						break;					
					}				
			}

		return bUserCancelled;										
	
	}
	
	
#pragma mark -


// 
//		CProjectTagAction
// 


CProjectTagAction::
CProjectTagAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList ), mHaveData(false)
	
	{

		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
	}


CProjectTagAction::
CProjectTagAction( CMacCVSDoc	*inOwnerDoc, LArray	*inItemList, Str255	inTagString, Boolean	inTagIsStatic, Boolean inRemoveExistingTags ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		::BlockMoveData(inTagString, mTagPStr, inTagString[0] + 1);
		mTagIsStatic				= inTagIsStatic;
		mRemoveExistingTag	= inRemoveExistingTags;
		mHaveData						= true;
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
	}


Boolean 
CProjectTagAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		Str255							pstrRevision		= "\p";
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );
				
				if ( FSpIsRepositoryDir(&fsParent) )
					{
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, NULL, NULL, NULL );
				
						if ( error == noErr )
							if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
								bResult = true;												
					}														
			}
		else if ( error == kCVSEngineIsFolder )
			{			
				if ( FSpIsRepositoryDir(&inTargetItem) )
					bResult = true;			
			}
			
		return bResult;
		
	}


void 
CProjectTagAction::RedoSelf(void)
	{
		
		OSErr			error							= noErr;		
		Boolean 	bUserCancelled		= false;
		Boolean		bStaticTag				= false;
		UInt32		intMaxIndex				= 0;
		LArray		*pItemsList				= NULL;
		FSSpec		fsCurSpec					= { 0, 0, "\p" };
		
		CMacCVSTagTask::CMacCVSTagTaskItem	sTaskItem;
		
		try 
			{
				pItemsList = new LArray( sizeof(sTaskItem) );
				ThrowIfNil_( pItemsList );
								
				CheckAllItemsQualify();
								
				if ( mItemList.GetCount() > 0 )
					{
						sTaskItem.Init();
						
						if (mHaveData)
							{	
								PStrCopy( mTagPStr, sTaskItem.mOptions.fTagPStr, sizeof(sTaskItem.mOptions.fTagPStr) );
								
								bStaticTag									= mTagIsStatic;
								sTaskItem.mOptions.fRemove	= mRemoveExistingTag;																	
							}
						else
							{					
								bUserCancelled = GetDesiredTag( sTaskItem.mOptions.fTagPStr, sTaskItem.mOptions.fRemove, bStaticTag );
							}							
						
						if ( !bUserCancelled )
							{
								if ( sTaskItem.mOptions.fRemove && !mHaveData)
									{
										bUserCancelled = WarnUserOfImpendingDoom( sTaskItem.mOptions.fTagPStr );
									}
							}
							
						if ( !bUserCancelled )
							{				
								if ( !bStaticTag )
									{
										sTaskItem.mOptions.fBranch = true;
									}
								else
									{
										sTaskItem.mOptions.fBranch = false;
									}
			
								sTaskItem.mOptions.fTagDeep = true;
								sTaskItem.mOptions.fDontTagMROFiles = false;
																					
								intMaxIndex = mItemList.GetCount();
								
								if ( intMaxIndex > 0 )
									{
										for ( UInt32 intIndex = 1; intIndex <= intMaxIndex; intIndex++ )
											{
												if ( mItemList.FetchItemAt(intIndex, &fsCurSpec) )
													{
														sTaskItem.mFileSpec = fsCurSpec;

														pItemsList->InsertItemsAt( 1, 1, &sTaskItem, sizeof(sTaskItem) );
													}
											}
									}
					
								if ( pItemsList->GetCount() > 0 )
									{
										CMacCVSTagTask *pTask = new CMacCVSTagTask( pItemsList, mOwnerDoc );
										ThrowIfNil_( pTask );
		
										pTask->SetCompletionEvent( mCurrentEvent, mReplyEvent );
										pTask->Resume();
									}
														
								delete pItemsList;
								pItemsList = NULL;
							}						
					}															
			}
			
		catch ( ExceptionCode inErr )
			{										
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}

				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );			
			}
			
		catch ( ... )
			{
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
		
				throw;			
			}
			
	}


Boolean		
CProjectTagAction::GetDesiredTag( Str255 pstrOutTag, Boolean &bOutRemove, Boolean &bOutStaticTag )
	{
		
		Boolean bUserCancelled 	= false;
		Boolean	bDone					 	= false;
		Boolean	bStaticTag			= false;
		Boolean	bRemoveTag			= false;
		UInt32	intItemHit			= 0;
		Str255	pstrTag					= "\p";
									
		StDialogHandler	stHandler( kTagDialogPPobID, mOwnerDoc );
		stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

		LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
		ThrowIfNil_( pDialog );
						
		LEditText *pTagField = dynamic_cast<LEditText *>( pDialog->FindPaneByID(kTagDialogTagEditFieldID) );
		ThrowIfNil_( pTagField );
		
		LRadioButton *pStaticButton = dynamic_cast<LRadioButton *>( pDialog->FindPaneByID(kTagDialogStaticRadioButtonID) );
		ThrowIfNil_( pStaticButton );

		LRadioButton *pBranchButton = dynamic_cast<LRadioButton *>( pDialog->FindPaneByID(kTagDialogBranchRadioButtonID) );
		ThrowIfNil_( pBranchButton );
			
		LCheckBox *pRemoveBox = dynamic_cast<LCheckBox *>( pDialog->FindPaneByID(kTagDialogRemoveCheckBoxID) );
		ThrowIfNil_( pRemoveBox );
					
		pDialog->SetDefaultButton( kTagDialogOKButtonID );
		pDialog->SetCancelButton( kTagDialogCancelButtonID );
		pDialog->Show();
								
		while ( !bDone )
			{		
				intItemHit = stHandler.DoDialog();
				
				switch ( intItemHit )
					{				
						case msg_OK:
						
							pTagField->GetDescriptor( pstrTag );
							
							if ( pstrTag[0] == 0 )
								{								
									UErrorDialog	theError( CTagPopup::kInvalidTagEnteredErr, mOwnerDoc );							
								}							
							else
								{									
									bDone = true;
												
									if ( pStaticButton->GetValue() > 0 )
										{										
											bStaticTag = true;						
										}
									else
										{
											bStaticTag = false;
										}

									if ( pRemoveBox->GetValue() > 0 )
										{
											bRemoveTag = true;
										}
									else
										{
											bRemoveTag = false;
										}
								}
								
						break;
						
						case msg_Cancel:																		
							bDone = true;	
							bUserCancelled = true;									
						break;																		
					
						case kTagDialogRemoveCheckBoxValueMessage:
							
							if ( pRemoveBox->GetValue() > 0 )
								{
									pStaticButton->Disable();
									pBranchButton->Disable();
								}
							else
								{
									pStaticButton->Enable();
									pBranchButton->Enable();
								}
								
						break;
					}				
			}				
		
		pDialog->Hide();
		
		if ( !bUserCancelled )
			{
				PStrCopy( pstrTag, pstrOutTag, sizeof(Str255) );
				
				bOutStaticTag = bStaticTag;
				bOutRemove = bRemoveTag;
			}
				
		return bUserCancelled;
		
	}
	

Boolean
CProjectTagAction::WarnUserOfImpendingDoom( ConstStringPtr pstrInTag )
	{
	
		OSErr				result							= noErr;					
								
		Boolean			bDone								= false;
		Boolean			bUserCancelled			= false;
		UInt32			intItemHit					= 0;
		Str255			pstrTruncated				= "\p";
				
		StDialogHandler	stHandler( kAskOKCancelDialogPPobID, mOwnerDoc );
		stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

		LDialogBox *pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
		ThrowIfNil_( pDialog );
	
		LCaption *pMessage = dynamic_cast<LCaption *>( pDialog->FindPaneByID(kAskOKCancelDialogCaptionID) );
		ThrowIfNil_( pMessage );
						
		PStrCopy( pstrInTag, pstrTruncated, sizeof(pstrTruncated) );
				
		if ( pstrTruncated[0] > 39 )
			{			
				pstrTruncated[0] = 40;
				pstrTruncated[pstrTruncated[0]] = '';				
			}

		LStr255	pstrWarning;
			
		pstrWarning.Append( "\pAre you sure you want to remove the tag " );
		pstrWarning.Append( pstrTruncated );
		pstrWarning.Append( "\p? Your build engineer will be *really* mad if you delete a tag without first consulting them!" );
		
		pMessage->SetDescriptor( pstrWarning );
		pDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
		pDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
		
		::SysBeep( 10 );
		
		pDialog->Show();
									
		while ( !bDone )
			{		
				intItemHit = stHandler.DoDialog();
				
				switch ( intItemHit )
					{			
						case msg_OK:						
							bDone = true;							
						break;
						
						case msg_Cancel:
						
							bDone = true;
							bUserCancelled = true;

						break;						
					}					
			}

		pDialog->Hide();
		
		return bUserCancelled;
			
	}


#pragma mark -


// 
//		CProjectUpdateAction
// 


CProjectUpdateAction::
CProjectUpdateAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean	inIsCheckout ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
		mCustomOptions.Init();
		mUseCustomOptions = false;

		mIsCheckout = inIsCheckout;
		
	}


CProjectUpdateAction::
CProjectUpdateAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean	inIsCheckout, const CCVSEngineUpdateOptions *inOptions ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		ThrowIfNil_( inOptions );
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
		mCustomOptions = *inOptions;
		mUseCustomOptions = true;

		mIsCheckout = inIsCheckout;
		
	}


CProjectUpdateAction::
~CProjectUpdateAction()
	{
	}
	
	
Boolean 
CProjectUpdateAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
					
		OSErr			error						= noErr;
		Boolean		bResult					= false;
	
		FSSpec		fsParent				= { 0, 0, "\p" };
		Str255		pstrRevision		= "\p";
				
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );
				
				if ( FSpIsRepositoryDir(&fsParent) )
					{
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, NULL, NULL, NULL );
				
						if ( error == noErr )
							{
								if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
									{
										bResult = true;	
									}
							}		
					}																
			}
		else if ( error == kCVSEngineIsFolder )
			{
				if ( FSpIsRepositoryDir(&inTargetItem) )
					{
						bResult = true;
					}
			}
			
		return bResult;
		
	}


void 
CProjectUpdateAction::RedoSelf(void)
	{
		
		FSSpec		fsCurSpec			= { 0, 0, "\p" };
		UInt32		intMaxIndex		= 0;
		UInt32		intIOCount		= 0;
		LArray		*pItemsList		= NULL;
	
		CMacCVSUpdateTask::CMacCVSUpdateTaskItem	sTaskItem;
			
		try 
			{
				pItemsList = new LArray( sizeof(sTaskItem) );
				ThrowIfNil_( pItemsList );
								
				CheckAllItemsQualify();

				intMaxIndex = mItemList.GetCount();
				
				if ( intMaxIndex > 0 )
					{
						for ( UInt32 intIndex = 1; intIndex <= intMaxIndex; intIndex++ )
							{
								if ( mItemList.FetchItemAt(intIndex, &fsCurSpec) )
									{
										sTaskItem.Init();
										
										if ( !mUseCustomOptions )
											{										
												sTaskItem.mOptions.mGetNewDirectories = mIsCheckout;

												sTaskItem.mOptions.mMergePolicy = mOwnerDoc->GetMergePolicy(); 
												sTaskItem.mOptions.mLFTranslation = mOwnerDoc->GetTranslation();		
												sTaskItem.mOptions.mCompressionLevel = mOwnerDoc->GetCompressionLevel();	
												
												if ( sTaskItem.mOptions.mCompressionLevel > 0 )
													{
														sTaskItem.mOptions.mUseCompression = true;
													}						
											}
										else 
											{
												sTaskItem.mOptions = mCustomOptions;
											}
										
										sTaskItem.mFileSpec = fsCurSpec;

										pItemsList->InsertItemsAt( 1, 1, &sTaskItem, sizeof(sTaskItem) );
									}
							}
					}
					
				if ( pItemsList->GetCount() > 0 )
					{
						CMacCVSUpdateTask *pTask = new CMacCVSUpdateTask( pItemsList, mOwnerDoc );
						ThrowIfNil_( pTask );
						
						if ( pTask != NULL )
							{	
								pTask->SetCompletionEvent( mCurrentEvent, mReplyEvent );
							}
							
						pTask->Resume();
					}
						
				delete pItemsList;
				pItemsList = NULL;						
			}
			
		catch ( ExceptionCode inErr )
			{				
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
					
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );			
			}
			
		catch ( ... )
			{
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
		
				throw;			
			}
			
	}


#pragma mark -


// 
//		CProjectUpdateToAction
// 


CProjectUpdateToAction::
CProjectUpdateToAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const Boolean	inIsCheckout ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
		mIsCheckout = inIsCheckout;
		
	}


CProjectUpdateToAction::
~CProjectUpdateToAction()
	{		
	}
	
	
Boolean 
CProjectUpdateToAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		Str255							pstrRevision		= "\p";
					
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );

				if ( FSpIsRepositoryDir(&fsParent) )
					{				
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory(&fsParent, fsRdPerm);
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, NULL, NULL, NULL );
				
						if ( error == noErr )
							if ( !CUIsAddedOrRemovedRevision(pstrRevision) )
								bResult = true;
					}																			
			}
		else if ( error == kCVSEngineIsFolder )
			{			
				if ( FSpIsRepositoryDir(&inTargetItem) )
					bResult = true;			
			}
			
		return bResult;
		
	}


void 
CProjectUpdateToAction::RedoSelf(void)
	{
		
		FSSpec		fsCurItem				= { 0, 0, "\p" };
		Boolean		bUserCancelled  = false;		
		UInt32		intMaxIndex			= 0;
		UInt32		intIOCount			= 0;
		Str255		pstrTag					= "\p";
		LArray		*pItemsList			= NULL;
															
		CMacCVSUpdateTask::CMacCVSUpdateTaskItem sTaskItem;
		
		try 
			{
				bUserCancelled = GetDesiredRevision( mIsCheckout, (mItemList.GetCount() > 1), pstrTag );
				
				if ( !bUserCancelled )
					{					
						pItemsList = new LArray( sizeof(sTaskItem) );
						ThrowIfNil_( pItemsList );
						
						CheckAllItemsQualify();

						LArrayIterator	itemIterator( mItemList );

						intIOCount = sizeof( FSSpec );		
						while( itemIterator.Next(&fsCurItem, intIOCount) )
							{
								Assert_( intIOCount == sizeof(FSSpec) );
								intIOCount = sizeof( FSSpec );
								
								sTaskItem.Init();
								sTaskItem.mFileSpec = fsCurItem;
								
								if ( pstrTag[0] > 1 )
									{						
										if ( pstrTag[1] == 'T' )
											{
												sTaskItem.mOptions.mRevisionPStr[0] = pstrTag[0] - 1;
												::BlockMoveData( &pstrTag[2], &sTaskItem.mOptions.mRevisionPStr[1], sTaskItem.mOptions.mRevisionPStr[0] );									
											}
										else if ( pstrTag[1] == 'D' )
											{										
												sTaskItem.mOptions.mDatePStr[0] = pstrTag[0] - 1;
												::BlockMoveData( &pstrTag[2], &sTaskItem.mOptions.mDatePStr[1], sTaskItem.mOptions.mDatePStr[0] );										
											}
										else Assert_( false );										
									}
	
								sTaskItem.mOptions.mGetNewDirectories = mIsCheckout;
								sTaskItem.mOptions.mResetStickyTags = true;					
								sTaskItem.mOptions.mMergePolicy = mOwnerDoc->GetMergePolicy();
								sTaskItem.mOptions.mLFTranslation = mOwnerDoc->GetTranslation();
									
								pItemsList->InsertItemsAt( 1, LArray::index_First, &sTaskItem, sizeof(sTaskItem) );			
							}			

						if ( pItemsList->GetCount() > 0 )
							{
								CMacCVSUpdateTask *pTask = new CMacCVSUpdateTask( pItemsList, mOwnerDoc );
								ThrowIfNil_( pTask );
								
								pTask->Resume();
							}
					}																		
				
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
			}
						
		catch ( ExceptionCode inErr )
			{			
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
					
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			}		
			
		catch ( ... )
			{
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
					
				throw;
			}
	}


Boolean
CProjectUpdateToAction::GetDesiredRevision( const Boolean inIsCheckout, const Boolean inMultipleItems, Str255 pstrOutTag )
	{
		Boolean					bUserCancelled 	= false;
		Boolean					bDone 					= false;
		Boolean					bTagsModified		= false;
		Boolean					bHeededEvilWarning	= false;
		
		LDialogBox			*pDialog				= NULL;
		LPushButton			*pRemoveButton	= NULL;
		CTagPopup				*pTagPopup			= NULL;
		
		ResIDT					intDialogID			= 0;
		MessageT				intItemHit			= 0;
		Str255					pstrCurTag			= "\p";
		LArray					*pCurTagList		= NULL;
		LArray					*pNewTagList		= NULL;
					
		if ( inIsCheckout )
			{
				intDialogID = kCheckoutToDialogPPobID;
			}
		else 
			{
				intDialogID = kUpdateToDialogPPobID;
			}
						
		StDialogHandler	stHandler( intDialogID, mOwnerDoc );
		stHandler.AddAttachment( new CThreadedTaskYieldAttachment );

		pDialog = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
		ThrowIfNil_( pDialog );
						
		pRemoveButton = dynamic_cast<LPushButton *>( pDialog->FindPaneByID(kUpdateToRemoveButtonID) );
		ThrowIfNil_( pRemoveButton );
						
		pTagPopup = dynamic_cast<CTagPopup *>( pDialog->FindPaneByID(kUpdateToTagPopupID) );
		ThrowIfNil_( pTagPopup );

//  Set up tag list in the popup. CurTagList is not owned by us
//  so we should never delete it.
				
		pCurTagList = mOwnerDoc->GetTagListPtr();
		Assert_( pCurTagList != NULL );
						
		if ( pCurTagList != NULL )
			{
				pTagPopup->SetTagList( *pCurTagList );
			}
					
		Str255 defaultTagString = "\p";
						
		mOwnerDoc->GetDefaultTagPStr( defaultTagString );
		pTagPopup->SetDefaultTag( defaultTagString );
						
		pDialog->Show();
						
		while ( !bDone )
			{			
				intItemHit = stHandler.DoDialog();
				
				switch ( intItemHit )
					{				
						case msg_OK:

							pstrCurTag[0] = 0;
							pTagPopup->GetCurrentTag( pstrCurTag );									
						
							bHeededEvilWarning = false;

							if ( (inMultipleItems) && (pstrCurTag[0] > 0) )
								{
									Boolean bDoneWithWarning = false;
									UInt32 intWarningItem = 0;
									
									StDialogHandler	stWarning( kAskOKCancelDialogPPobID, mOwnerDoc );
									stWarning.AddAttachment( new CThreadedTaskYieldAttachment );
					
									LDialogBox *pWarningDialog = dynamic_cast<LDialogBox *>( stWarning.GetDialog() );
									ThrowIfNil_( pWarningDialog );
						
									LCaption *pMessage = dynamic_cast<LCaption *>( pWarningDialog->FindPaneByID(kAskOKCancelDialogCaptionID) );
									ThrowIfNil_( pMessage );

									LStr255	pstrWarning;
												
									pstrWarning.Append( "\pCareful! Are you really sure you want to update all the selected items" );
									pstrWarning.Append( "\p to a particular tag? CVS will obsolete the files in of your selection" );
									pstrWarning.Append( "\p that cannot be updated to that tag!" );
							
									pMessage->SetDescriptor( pstrWarning );
									pWarningDialog->SetDefaultButton( kAskOKCancelDialogCancelButtonID );
									pWarningDialog->SetCancelButton( kAskOKCancelDialogCancelButtonID );
							
									::SysBeep( 10 );
							
									pWarningDialog->Show();
														
									while ( !bDoneWithWarning )
										{									
											intWarningItem = stWarning.DoDialog();
											
											switch ( intWarningItem )
												{									
													case msg_OK:										
														bDoneWithWarning = true;														
													break;
													
													case msg_Cancel:													
														bDoneWithWarning = true;
														bHeededEvilWarning = true;
													break;												
												}											
										}													
								}
								
							if ( bHeededEvilWarning == false )
								{
									pDialog->Hide();
									bDone = true;
												
									PStrCopy( pstrCurTag, pstrOutTag, sizeof(Str255) );						
								}
																		
						break;
							
						case msg_Cancel:
										
							pDialog->Hide();
							bDone = true;
							bUserCancelled = true;
							
						break;
									
						case kUpdateToRemoveButtonValueMessage:
						
							pTagPopup->RemoveCurrentTag();

							if ( pTagPopup->CanRemoveCurrentTag() )
								{
									pRemoveButton->Enable();
								}
							else 
								{
									pRemoveButton->Disable();
								}
								
						break;
						
						case kUpdateToTagPopupValueMessage:
						
							pTagPopup->DoMouseDownInTagPopup();
							
							if ( pTagPopup->CanRemoveCurrentTag() )
								{
									pRemoveButton->Enable();
								}
							else 
								{
									pRemoveButton->Disable();
								}
								
						break;				
					}									
			} // while

		pNewTagList = pTagPopup->GetTagList();						
		Assert_( pNewTagList != NULL );
		
		if ( pNewTagList != NULL )
			{						
				bTagsModified = !mOwnerDoc->AreTagListsEqual( *pCurTagList, *pNewTagList );
				
				if ( bTagsModified )	
					{										
						mOwnerDoc->SetModified( true );										
						mOwnerDoc->UseNewTagList( *pNewTagList );										
					}
					
				delete pNewTagList;
				pNewTagList = NULL;										
			}
	

		return bUserCancelled;
	}
	

#pragma mark -


// 
//		CProjectWatchAction
// 


CProjectWatchAction::
CProjectWatchAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, const UInt32 intInAction, const CCVSEngineWatchOptions *inOptions ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{
		mOptions.Init();
		mAction = intInAction;
		
		if ( inOptions != NULL )
			{
				mOptions = *inOptions;
			}
	}


Boolean 
CProjectWatchAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error						= noErr;
		Boolean							bResult					= false;
	
		FSSpec							fsParent				= { 0, 0, "\p" };
		Str255							pstrRevision		= "\p";
						
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				error = FSpGetParentFSSpec( &inTargetItem, &fsParent );			
				ThrowIfError_( error );
				
				if ( FSpIsRepositoryDir(&fsParent) )
					{
						if ( !mEntries.AreYouThisDirectory(&fsParent) )
							{
								error = mEntries.CloseEntriesFile();
								ThrowIfError_( error );

								error = mEntries.LoadWorkingDirectory( &fsParent, fsRdPerm );
								ThrowIfError_( error );
							}
							
						error = mEntries.GetEntry( inTargetItem.name, pstrRevision, NULL, NULL, NULL );
				
						if ( error == noErr )
							{
								bResult = true;				
							}								
					}														
			}
		else if ( error == kCVSEngineIsFolder )
			{			
				if ( FSpIsRepositoryDir(&inTargetItem) )
					{
						bResult = true;	
					}		
			}
			
		return bResult;
		
	}


void 
CProjectWatchAction::RedoSelf(void)
	{
		
		FSSpec		fsCurSpec			= { 0, 0, "\p" };
		UInt32		intMaxIndex		= 0;
		UInt32		intIOCount		= 0;
		LArray		*pItemsList		= NULL;
	
		CMacCVSWatchTask::CMacCVSWatchTaskItem sTaskItem;
			
		try 
			{
				pItemsList = new LArray( sizeof(sTaskItem) );
				ThrowIfNil_( pItemsList );
								
				CheckAllItemsQualify();

				intMaxIndex = mItemList.GetCount();
				
				if ( intMaxIndex > 0 )
					{
						for ( UInt32 intIndex = 1; intIndex <= intMaxIndex; intIndex++ )
							{
								if ( mItemList.FetchItemAt(intIndex, &fsCurSpec) )
									{
										sTaskItem.Init();
										sTaskItem.mFileSpec = fsCurSpec;
										sTaskItem.mOperation = mAction;
										sTaskItem.mOptions = mOptions;
										pItemsList->InsertItemsAt( 1, 1, &sTaskItem, sizeof(sTaskItem) );
									}
							}
					}
					
				if ( pItemsList->GetCount() > 0 )
					{
						CMacCVSWatchTask *pTask = new CMacCVSWatchTask( pItemsList, mOwnerDoc );
						ThrowIfNil_( pTask );
						
						pTask->Resume();
					}
						
				delete pItemsList;
				pItemsList = NULL;
			}			
		catch ( ExceptionCode inErr )
			{				
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
					
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );			
			}			
		catch ( ... )
			{
				if ( pItemsList != NULL )
					{
						delete pItemsList;
						pItemsList = NULL;
					}
		
				throw;			
			}		
				
	}


#pragma mark -


// 
//	 CProjectEditCommentAction
// 


CProjectEditCommentAction::
CProjectEditCommentAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
				
	}


Boolean 
CProjectEditCommentAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error				= noErr;
		Boolean							bResult			= false;
		
		CkidResourcePB			ckidInfo;
				
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	

				error = FSpReadCkidResource( &inTargetItem, &ckidInfo, false );
		
				if ( error == noErr )
					{															
						if ( error == noErr )
							if ( ckidInfo.fModifyReadOnly != 0 ) 
								bResult = true;																	
					}	

			}
						
		return bResult;
		
	}


void 
CProjectEditCommentAction::RedoSelf(void)
	{
		
		OSErr									error						= noErr;
		Boolean								bUserAborted	 	= false;
		FSSpec								theCurItem			= { 0, 0, "\p" };
		UInt32								ioCount					= sizeof( FSSpec );
		SInt32								oldModDate			= 0x00000000;
				
		try 
			{

				CheckAllItemsQualify(true);
				
				LArrayIterator	itemIterator( mItemList );
				
				while ( itemIterator.Next(&theCurItem, ioCount) )
					{

						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof(FSSpec);
								
						oldModDate = FSpGetModDate( &theCurItem );
				
						if ( EditCheckoutComment(&theCurItem, mOwnerDoc, true, false, bUserAborted) )
							FSpSetModDate( &theCurItem, oldModDate );
						else
							if (bUserAborted) break;
					}
					
			}
			
		catch ( ExceptionCode inErr )
			{
			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			
			}
			
	}


#pragma mark -

// 
//	 CProjectOrphanFilesAction
// 


CProjectOrphanFilesAction::
CProjectOrphanFilesAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList, CMacCVSOrphanFilesQueryRec *inQuery ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		ThrowIfNil_( inQuery );
		
		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
		
		mQueryRec = *inQuery;
				
	}


Boolean 
CProjectOrphanFilesAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error					= noErr;
		Boolean							bResult				= false;
		CkidMcvsResourcePB	ckidMcvsInfo;

		ckidMcvsInfo.mcvs.fRemotePathStr = NULL;
		
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{
				bResult = true;					
			}		
		else if ( error == kCVSEngineIsFolder )
			{		
				if ( FSpIsRepositoryDir(&inTargetItem) )
					bResult = true;		
			}
			
		return bResult;
		
	}


void 
CProjectOrphanFilesAction::RedoSelf(void)
	{
		
		OSErr									error						= noErr;
		Boolean								bUserAborted	 	= false;
				
		try 
			{

				CheckAllItemsQualify();
				
				CMacCVSOrphanFilesTask	*wipeInfoTask = NULL;
							
				wipeInfoTask = new CMacCVSOrphanFilesTask( mOwnerDoc, &mQueryRec, &mItemList );
				ThrowIfNil_( wipeInfoTask );
				
				wipeInfoTask->SetCompletionEvent( mCurrentEvent, mReplyEvent );
				wipeInfoTask->Resume();
					
			}
			
		catch ( ExceptionCode inErr )
			{
			
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw( inErr );
			
			}
			
	}


#pragma mark -

// 
//	 CProjectMROAction
// 


CProjectMROAction::
CProjectMROAction( CMacCVSDoc *inOwnerDoc, LArray *inItemList ):

	CProjectAction( inOwnerDoc, inItemList )
	
	{

		mOwnerDoc->GetProjectRootFileSpec( mLocalRoot );
				
	}


Boolean 
CProjectMROAction::IsItemQualified( const FSSpec &inTargetItem, Boolean *outFileWasOpen )
	{
		
		OSErr								error				= noErr;
		Boolean							bResult			= false;
		
		error = FSpIsAFile( &inTargetItem, outFileWasOpen );
		
		if ( error == noErr )
			{	
				bResult = true;
			}
							
		return bResult;
		
	}


void 
CProjectMROAction::RedoSelf(void)
	{
	
		OSErr									result					= noErr;
		FSSpec								theCurItem			= { 0, 0, "\p" };
		UInt32								ioCount					= sizeof( FSSpec );
		CkidResourcePB				ckidInfo;
		CInfoPBRec						pb;
		
		ckidInfo.fCommentHndl = NULL;					
		
		pb.hFileInfo.ioCompletion = NULL;
		pb.hFileInfo.ioFDirIndex 	= 0;
		pb.hFileInfo.ioNamePtr 		= NULL;
		pb.hFileInfo.ioVRefNum 		= 0;
		pb.hFileInfo.ioDirID 			= 0;
		
		try 
			{

				CheckAllItemsQualify(true);
				
				LArrayIterator	itemIterator( mItemList );
				
				while ( itemIterator.Next(&theCurItem, ioCount) )
					{

						Assert_( ioCount == sizeof(FSSpec) );
						ioCount = sizeof(FSSpec);
						
						ckidInfo.fCommentHndl = NULL;					
						
						pb.hFileInfo.ioNamePtr = theCurItem.name;
						pb.hFileInfo.ioVRefNum = theCurItem.vRefNum;
						pb.hFileInfo.ioDirID = theCurItem.parID;
						
						result = ::PBGetCatInfoSync( &pb );
						ThrowIfOSErr_( result );

/*
**  Here we make *sure* the file we are MROing isn't already open.
**  If we don't do this and the user tries to MRO the running app file,
**  FSpWriteCkidMcvsResource will close the app resource fork. Bad idea.
*/
										
						if ( 	(pb.hFileInfo.ioFlAttrib & kResForkOpen) || 
									(pb.hFileInfo.ioFlAttrib & kDataForkOpen) || 
									(pb.hFileInfo.ioFlAttrib & kBothForksOpen) )
							{
							
								result = opWrErr;
								ThrowOSErr_( result );
								
							}
						
						result = ::FSpRstFLock( &theCurItem );	
						ThrowIfOSErr_( result );
													
						result = FSpReadCkidResource( &theCurItem, &ckidInfo, true );
						
						if ( result == noErr )
							{
								ckidInfo.fModifyReadOnly = true;
						
								result = FSpWriteCkidResource( &theCurItem, &ckidInfo );
								ThrowIfOSErr_( result );
							}
						else
							{
								result = noErr;
							
								UInt32 intTime = 0;
								
								::GetDateTime( &intTime );
								
								result = FSpSetModDate( &theCurItem, intTime );
								ThrowIfOSErr_( result );							
							}
							
						mOwnerDoc->UpdateItemInProjectViews( &theCurItem, true );
													
						if ( ckidInfo.fCommentHndl != NULL )
							{							
								::DisposeHandle( ckidInfo.fCommentHndl );							
								ckidInfo.fCommentHndl = NULL;							
							}
							
					}
					
			}
			
		catch ( ExceptionCode inErr )
			{
				if ( ckidInfo.fCommentHndl != NULL )
					{							
						::DisposeHandle( ckidInfo.fCommentHndl );							
						ckidInfo.fCommentHndl = NULL;							
					}
					
				if ( inErr != CProjectAction::kProjectActionCancelled )
					throw inErr;			
			}
			
	}
	
	
#pragma mark -


Boolean EditCheckoutComment( const FSSpec *inTargetItem, CMacCVSDoc *inOwnerDoc, const Boolean inAskAlways, const Boolean inIsCommit, Boolean &cancelAll )
	{

		OSErr			intError			= noErr;
		SInt32		intModDate		= 0;
	
		Boolean		bResult				= false;
		Boolean		bDone 				= false;
		Boolean		bHasCkid			= false;
		
		CkidResourcePB sCkidInfo;
	
		memset( &sCkidInfo, 0, sizeof(CkidResourcePB) );

		ThrowIfNil_( inTargetItem );
		ThrowIfNil_( inOwnerDoc );
	
		LStr255	pstrTitle("Comment for ");
		
		try
			{					
				intModDate = FSpGetModDate( inTargetItem );			
				intError = FSpReadCkidResource( inTargetItem, &sCkidInfo, true );			
				
				if ( intError == noErr )
					{
						bHasCkid = true;
					}
				else
					{
						//  Uh oh, the file doesn't have a ckid. So set
						//  up a bogus one.
						
						memset( &sCkidInfo, 0, sizeof(CkidResourcePB) );
						
						inOwnerDoc->GetRemoteUserPStr( sCkidInfo.fUserNamePStr );
						PStrCopy( inTargetItem->name, sCkidInfo.fFileNamePStr, sizeof(sCkidInfo.fFileNamePStr) );
						
						bHasCkid = false;
					}
					
				if ( (sCkidInfo.fCommentLen > 0) && (!inAskAlways) )
					{				
//  Don't ask anything, file has a comment.
						bResult = true;						
					}					
				else
					{					 
						StDialogHandler stHandler( kMacCVSCommentDialogPPobID, inOwnerDoc );
						stHandler.AddAttachment( new CThreadedTaskYieldAttachment );
										
						LDialogBox *pDialogBox = dynamic_cast<LDialogBox *>( stHandler.GetDialog() );
						ThrowIfNil_( pDialogBox );
																
						LTextEditView	*pTextField = dynamic_cast<LTextEditView *>( pDialogBox->FindPaneByID(kMacCVSCommentDialogTEID) );	
						ThrowIfNil_( pTextField );
						
						LPushButton		*pSaveButton = dynamic_cast<LPushButton *>( pDialogBox->FindPaneByID(kMacCVSCommentDialogOKButtonID) );
						ThrowIfNil_( pSaveButton );
						
						if ( !inIsCommit )
							{
								pSaveButton->SetDescriptor("\pSave");
							}
							
						pDialogBox->SetDefaultButton( kMacCVSCommentDialogOKButtonID );
						pDialogBox->SetCancelButton( kMacCVSCommentDialogCancelButtonID );
						
						pDialogBox->SetLatentSub( pTextField );
											
						if ( sCkidInfo.fCommentHndl != NULL )
							{
								pTextField->SetTextHandle( sCkidInfo.fCommentHndl );
							}
							
						pstrTitle.Append( inTargetItem->name );
						
						pDialogBox->SetDescriptor( pstrTitle );
						pDialogBox->Show();
						
						while ( !bDone )
							{						
								try
									{								
										switch ( stHandler.DoDialog() )
											{ 										
												case msg_OK:
													bDone = true;
													bResult = true;
												break;

												case msg_Cancel:
													bDone = true;
												break;
												
												case msg_CancelAll:
													bDone = true;
													cancelAll = true;
												break;	
											}											
									}									
								catch ( ExceptionCode inErr )
									{									
										if ( inErr != err_32kLimit )
											{
												throw inErr;			
											}						
									}									
							}

//  If the user pressed ok, save the comment.

						if ( bResult )
							{						
								if ( sCkidInfo.fCommentHndl != NULL )
									{									
										::DisposeHandle( sCkidInfo.fCommentHndl );
										sCkidInfo.fCommentHndl = NULL;
										sCkidInfo.fCommentLen = 0;										
									}
									
								sCkidInfo.fCommentHndl = pTextField->GetTextHandle();

								intError = ::HandToHand( &sCkidInfo.fCommentHndl );
								
								if ( intError != noErr )
									{									
										sCkidInfo.fCommentHndl = NULL;										
										ThrowOSErr_( intError );									
									}
								
								sCkidInfo.fCommentLen = ::GetHandleSize( sCkidInfo.fCommentHndl );
										
								{									
									CInfoPBRec	pb;
									FSSpec			fsTarget = *inTargetItem;
									
									pb.hFileInfo.ioCompletion = NULL;
									pb.hFileInfo.ioFDirIndex 	= 0;
									pb.hFileInfo.ioNamePtr 		= fsTarget.name;
									pb.hFileInfo.ioVRefNum 		= fsTarget.vRefNum;
									pb.hFileInfo.ioDirID 			= fsTarget.parID;
						
									intError = ::PBGetCatInfoSync( &pb );
									ThrowIfOSErr_( intError );
/*
**  Here we make *sure* the file we are MROing isn't already open.
**  If we don't do this and the user tries to MRO the running app file,
**  FSpWriteCkidMcvsResource will close the app resource fork. Bad idea.
*/										
									if ( 	(pb.hFileInfo.ioFlAttrib & kResForkOpen) || 
												(pb.hFileInfo.ioFlAttrib & kDataForkOpen) || 
												(pb.hFileInfo.ioFlAttrib & kBothForksOpen) )
										{										
											intError = opWrErr;
											ThrowOSErr_( intError );										
										}										
								}
							
								Boolean bFileLocked = FSpIsLocked( inTargetItem );
								
								if ( bFileLocked )
									{
										::FSpRstFLock( inTargetItem );
									}
									
								intError = FSpWriteCkidResource( inTargetItem, &sCkidInfo );
								ThrowIfOSErr_( intError );
									
								if ( bFileLocked )
									{
										::FSpSetFLock( inTargetItem );
									}
									
								FSpSetModDate( inTargetItem, intModDate );										
							}						
					}
					
				if ( sCkidInfo.fCommentHndl != NULL )
					{					
						::DisposeHandle( sCkidInfo.fCommentHndl );				
						sCkidInfo.fCommentHndl = NULL;						
					}				
			}
					
		catch ( ... )
			{								
				if ( sCkidInfo.fCommentHndl != NULL )
					{					
						::DisposeHandle( sCkidInfo.fCommentHndl );				
						sCkidInfo.fCommentHndl = NULL;						
					}					
				throw;				
			}
					
		return bResult;
			
	}
	
